%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% Path-sensitive interprocedural alias analysis

% the strategy of the alias analysis is to generally assume non-aliasing
% on input data and side effects of functions, and identify only those places
% where this assumption is violated. the results of the analysis are a set
% of points-to graphs for possibly aliased values on globals, composite types,
% and function entry and exit points.

import "../memory/intignore.clp".

analyze session_name("cil_body").
analyze analysis_name("aliasing").
analyze bottomup().

using map.

% PREDICATES

% ignore spoints on the given traces. these are either covered by annotated
% invariants, or deal with low level system bits we aren't modelling
predicate drop_spoints(in SUM:sum,in POS:point_pos,in S:t_trace,in T:t_trace).

% spoints(S,T) should not be considered for a composite type invariant on C/STR
predicate omit_comp_spoints(in C:string,in SUM:sum,in STR:t_trace,
                            in S:t_trace,in T:t_trace).

% spoints(S,T) should not be considered for a global invariant on G
predicate omit_glob_spoints(in G:string,in S:t_trace,in T:t_trace).

% don't use entry_points() to identify aliased traces on callees
predicate no_entry_trace_alias().

% don't generate ANY composite or global invariants
predicate no_comp_glob_points().

% supporting information for each generated spoints edge
type t_points_asn ::=
    asn_exit{P:pp,ASN:g_guard_asn}
  | asn_entry{P:pp,T:t_trace,CT0:t_trace,CT1:t_trace,ASN:g_guard_asn}
  | asn_none{SUM:sum}.

% at entry/exit to SUM, S may point to T according to A
predicate asn_points(SUM:sum,POS:point_pos,S:t_trace,T:t_trace,A:t_points_asn).
predicate try_asn_points(SUM:sum,POS:point_pos,S:t_trace,T:t_trace,A:t_points_asn).

% info about aliases used to extend rel_inst_trace. CT maps to T at I due to Y
type ext_why ::=
    rel_direct
  | rel_alias{XT:t_trace,PT:t_trace,DPT:t_trace}.
predicate ext_rel_inst_trace(I:sum,P:pp,CT:t_trace,T:t_trace,G:g_guard,
                             Y:ext_why).

% RULES

% we split and fixpoint over all loops
+split_loops().

% we only use the entry aliasing to get additional entry aliasing on callees
+no_entry_aliasing().

% only generate inst_trace for arguments included in prototypes (we want to
% be able to type everything). extra args to varargs will be smashed anyways
+matching_arg_inst_trace().

% interested in asm statements for usemod info
+cfg_translate_asm().

% also compute usemod and UI information
import "usemod.clp".
import "aliasingui.clp".

% propagate generated exit/entry aliasing up/down
try_asn_points(SUM,fn_exit,S,T,_), +inst_exit_points(SUM,S,T).
try_asn_points(SUM,fn_entry,S,T,_), +inst_entry_points(SUM,S,T).

% generate exit aliasing

predicate scalar_trace(in SUM:sum,in S:t_trace).
?scalar_trace(SUM,S), trace_type(SUM,S,y_int{_}), +scalar_trace(SUM,S).
?scalar_trace(SUM,S), trace_type(SUM,S,y_flt{_}), +scalar_trace(SUM,S).
?scalar_trace(SUM,S), trace_type(SUM,S,y_enum{_}), +scalar_trace(SUM,S).

% mark any updated traces as as live at func/loop exit,
% then get their targets to construct the exit points-to graph, updating
% inst_exit_points for use by memory.clp
access(P,S,write), sum_body(SUM,P),
    ~scalar_trace(SUM,S), sum_bound(SUM,_,PE),
    val(PE,S,trace{T},VG), trace_visible(SUM,T),
    guard(PE,GG), #and(GG,VG,G), #satasn(G,ASN),
    +try_asn_points(SUM,fn_exit,S,T,asn_exit{PE,ASN}).

% for path-sensitive treatment of loops we have to disable inst_exit_points
% and fixpoint manually to avoid invalidated negation/collection
iloop(P,_,PH), sum_body(s_loop{PH},P), +no_exit_aliasing(s_loop{PH},P),
    inst_exit_points(s_loop{PH},CS,CT),
    inst_trace(s_loop{PH},P,CS,trace{S},SG), ~scalar_trace(s_loop{PH},S),
    inst_trace(s_loop{PH},P,CT,trace{T},TG), trace_visible(s_loop{PH},T),
    guard(P,GG), and3(GG,SG,TG,G), #satasn(G,ASN),
    +try_asn_points(s_loop{PH},fn_exit,S,T,asn_exit{P,ASN}).

% existing aliased traces at entry

% trace T could be aliased with NT due to entry points-to relationship PT->DPT
predicate trace_alias(in SUM:sum,in T:t_trace,AT:t_trace,
                      PT:t_trace,DPT:t_trace).
~no_entry_trace_alias(), ?trace_alias(SUM,T,_,_,_),
    trace_sub(T,drf{ST},RT), entry_points(SUM,ST,DST), ~trace_bad_cast(SUM,DST),
    trace_compose(SUM,DST,RT,AT), ~trace_bad_cast(SUM,AT),
    +trace_alias(SUM,T,AT,ST,DST).

% fill in glob_trace_rename() if we are not computing global points-to rels
no_comp_glob_points(), +glob_trace_rename(T,T).

% base rel_inst_trace to extend according to aliases
rel_inst_trace(I,P,CT,T,G), +ext_rel_inst_trace(I,P,CT,T,G,rel_direct).

% extend rel_inst_trace according to any aliased traces on entry to SUM
% only get aliasing for non-global callee roots, pure global-global aliasing
% will be handled within the callee
ext_rel_inst_trace(I,P,CT,T,G,_), CT\=rfld{_,_,_},
    (~trace_root(CT,glob{_}); no_comp_glob_points()),
    sum_body(SUM,P), trace_alias(SUM,T,NT,PT,DPT),
    +ext_rel_inst_trace(I,P,CT,NT,G,rel_alias{T,PT,DPT}),
    ~trace_root(NT,arg{_}), inst_may_access(I,NT,_),
    +ext_rel_inst_trace(I,P,NT,NT,G,rel_direct).

% extend rel_inst_trace according to side effects on aliased traces on entry
% to SUM. if S is written with DS and S might alias NS, NS may also point to DS
access(AP,S,write), sum_body(SUM,AP), trace_alias(SUM,S,NS,PT,DPT),
    ext_rel_inst_trace(I,P,CT,drf{NS},G,_), val(P,S,trace{DS},_),
    +ext_rel_inst_trace(I,P,CT,DS,G,rel_alias{drf{NS},PT,DPT}).

% have to use a td_inst_trace for aliasing that is not as precise with
% guards as we'd like. due to fixpointing entry aliasing over loops
% that makes figuring out the relevant traces impossible without introducing
% cycles. can probably fix this only by analyzing cil_sum_body() instead
predicate alias_td_inst_trace(in I:sum,in P:pp,in T:t_trace,CT:t_trace,G:g_guard).
?alias_td_inst_trace(I,P,T,_,_), trace_sub(T,ST,RT),
    ext_rel_inst_trace(I,P,CST,ST,G,_),
    trace_compose(I,CST,RT,CT), +alias_td_inst_trace(I,P,T,CT,G).
?alias_td_inst_trace(I,P,T,_,_), (~trace_root(T,arg{_}); I=s_loop{_}),
    inst_may_access(I,T,_),
    inst_trace(I,P,T,trace{T},G),
    +alias_td_inst_trace(I,P,T,T,G).

% generate entry aliasing on loops and callees

% to identify aliased traces on invoked calls and loops, we need take all
% traces the callee could access, map them back to traces in the current
% function, and see if there is an intersection where a caller trace could
% be accessed through two different callee traces

% if T0 and T1 are aliased, SUM may access aliased data through them.
% use coarse may_access unless per-trace usemod information is available
predicate may_access_alias(in SUM:sum,in T0:t_trace,in T1:t_trace).
?may_access_alias(SUM,T0,T1),
    inst_may_access(SUM,T0,Y0), inst_may_access(SUM,T1,Y1),
    (Y0=write; Y0=deepwrite; Y1=write; Y1=deepwrite),
    +may_access_alias(SUM,T0,T1).

predicate omit_access_alias(in T0:t_trace,in T1:t_trace).
?omit_access_alias(T0,T1), trace_root(T0,R), trace_root(T1,R), +omit_access_alias(T0,T1).
?omit_access_alias(T0,T1), trace_root(T0,glob{_}), trace_root(T1,glob{_}), +omit_access_alias(T0,T1).

% hack for dealing with function pointers where the same c_instr may
% reference multiple different targets. TODO: fixme
predicate actual_usemod(in FN:string,in T:t_trace).
?actual_usemod(FN,T), trace_sub(T,ST,_),
    sum_usemod(FN,s_func)->stuse(ST,_), +actual_usemod(FN,T).

% get aliased traces on each call/loop
ext_rel_inst_trace(I,P,drf{CPT},T,G0,_), alias_td_inst_trace(I,P,T,CT,G1),
    may_access_alias(I,drf{CPT},CT), ~omit_access_alias(drf{CPT},CT),
    guard(P,GG), and3(GG,G0,G1,G), #satasn(G,ASN),
    +try_asn_points(I,fn_entry,CPT,CT,asn_entry{P,T,drf{CPT},CT,ASN}).

% fill in the second half of each alias relationship, S->*T => T->*T as well.
% need this for renaming later
try_asn_points(I,fn_entry,_,drf{T},_),
    +try_asn_points(I,fn_entry,T,drf{T},asn_none{I}).

% merging assignments together

predicate pri_points_asn(in A0:t_points_asn,in A1:t_points_asn).
+pri_points_asn(_,asn_none{_}).

predicate merge_points_asns(in A0:t_points_asn,in A1:t_points_asn,
                            A:t_points_asn).
?merge_points_asns(A0,A1,_), lessthan(A0,A1), pri_points_asn(A0,A1),
    +merge_points_asns(A0,A1,A0).
?merge_points_asns(A0,A1,_), lessthan(A0,A1), ~pri_points_asn(A0,A1),
    +merge_points_asns(A0,A1,A1).
?merge_points_asns(A0,A1,_), lessthan(A1,A0),
    merge_points_asns(A1,A0,A), +merge_points_asns(A0,A1,A).
+merge_points_asns(A,A,A).

predicate merge_points_asns_list(in BA:t_points_asn,in ALIST:list[t_points_asn],A:t_points_asn).
+merge_points_asns_list(A,[],A).
?merge_points_asns_list(BA,[NA|TAIL],_),
    merge_points_asns(BA,NA,NBA), merge_points_asns_list(NBA,TAIL,A),
    +merge_points_asns_list(BA,[NA|TAIL],A).

try_asn_points(SUM,POS,S,T,_),
    \/try_asn_points(SUM,POS,S,T,XA):list_all(XA,[BA|ALIST]),
    merge_points_asns_list(BA,ALIST,ASN),
    +asn_points(SUM,POS,S,T,ASN).

% rename traces in function/loop entry/exit aliasing

% at entry and exit to each function and loop, we want to rename traces in the
% points-to graphs in a fashion that will simplify them while still preserving
% the structure of the original graph.
%
% at entry to functions/loops, we are concerned with eliminating redundant
% points-to relationships between complex aliased structures. if S->drf{T}
% is in the entry graph, drf{S}->drf{drf{T}} will as well, but does not
% provide any new information. rename drf{T} and all subtraces to drf{S} to
% collapse the two together, giving the final relationship T->drf{S}
%
% at exit to functions/loops, we filter out fresh traces, those heap cells
% allocated during execution of the function/loop (or any callees), by renaming
% them according to how they are used in the exit graph. if T->fresh{X} in
% the exit graph, rename fresh{X} to drf{T}, allowing us to omit the edge from
% the exit graph. if another S->fresh{X} as well, S->drf{T} will be in the
% exit graph

% T is filtered out of the graph for SUM by renaming to NT
predicate filter_rename(SUM:sum,POS:point_pos,T:t_trace,NT:t_trace,ASN:t_points_asn).

% NT is a renaming for T but is pre-empted by another
predicate weak_filter_rename(SUM:sum,POS:point_pos,T:t_trace,NT:t_trace).

% S is a source for filtering out T from the graph for SUM
% all traces *MAY* have T replaced with drf{S}
predicate filter_source(SUM:sum,POS:point_pos,S:t_trace,T:t_trace,ASN:t_points_asn).

predicate filter_node(SUM:sum,POS:point_pos,T:t_trace).
predicate filter_trace_sub(SUM:sum,POS:point_pos,
                           T0:t_trace,T1:t_trace,TR:t_trace).
filter_node(SUM,POS,T0), trace_sub(T0,T1,TR),
    +filter_trace_sub(SUM,POS,T0,T1,TR).

% in the entry graph we are renaming the target of any pointer
% aliased with another
asn_points(SUM,fn_entry,S,T,ASN),
    +filter_node(SUM,fn_entry,S), +filter_node(SUM,fn_entry,T),
    ~trace_root(T,glob{_}), T=drf{_},
    +filter_source(SUM,fn_entry,S,T,ASN).

% in the exit graph we are renaming fresh traces reachable at exit only
% (these are identified as temp{_}, i.e. function return values), so add
% filter sources for any non-fresh traces that point to them
asn_points(SUM,fn_exit,S,T,ASN),
    +filter_node(SUM,fn_exit,S), +filter_node(SUM,fn_exit,T),
    ~trace_root(S,temp{_,_}), trace_visible(SUM,S), trace_root(T,temp{_,_}),
    ~drop_spoints(SUM,fn_exit,S,drf{S}),
    ~(trace_sub(S,fld{STR,_,C},_), comp_to_relative(STR,S,RS), comp_poly(C,_,RS,_)),
    +filter_source(SUM,fn_exit,S,T,ASN).

% we need additional fresh sources if there are multiple levels of fresh values
asn_points(SUM,fn_exit,S,T,ASN),
    trace_root(S,temp{_,_}), trace_root(T,temp{_,_}),
    ~trace_simplify(SUM,drf{S},T),
    filter_rename(SUM,fn_exit,S,NS,_), +filter_source(SUM,fn_exit,NS,T,ASN).

% for every source S->T, replace T with drf{S} in the renaming
filter_source(SUM,POS,S,T,ASN), filter_trace_sub(SUM,POS,T1,T,SUBT),
    trace_simplify(SUM,drf{S},DS),
    trace_compose(SUM,DS,SUBT,NT1), +filter_rename(SUM,POS,T1,NT1,ASN).

% special case for function exits. if the only pointer to a fresh trace is to
% one of its fields, reverse the operation through rfld to get a renaming for
% effects on reachable data
filter_source(SUM,fn_exit,S,fld{T,F,C},ASN),
    filter_trace_sub(SUM,fn_exit,T1,T,SUB),
    trace_simplify(SUM,drf{S},DS), trace_compose(SUM,rfld{DS,F,C},SUB,NT1),
    +filter_rename(SUM,fn_exit,T1,NT1,ASN).

% if there are multiple renamings for the same value, pick one and mark all
% others as weak. favor global over local values, no rfld over rfld, and
% shallower over deeper (in deref counts). otherwise just pick one canonically
predicate pri_rename(in T0:t_trace,in T1:t_trace,PRI:int).
pri_rename(T0,T1,2) :- trace_root(T0,glob{_}), ~trace_root(T1,glob{_}).
pri_rename(T0,T1,3) :- ~trace_sub(T0,rfld{_,_,_},_), trace_sub(T1,rfld{_,_,_},_).
pri_rename(T0,T1,4) :-  trace_derefs(T0,N0), trace_derefs(T1,N1), int_lt(N0,N1).
pri_rename(T0,T1,5) :- ~lessthan(T0,T1).
+pri_rename(_,_,6).
filter_rename(SUM,POS,T,T0,_), filter_rename(SUM,POS,T,T1,_), T0\=T1,
    \/pri_rename(T0,T1,P0):int_min_all(P0,PRI0),
    \/pri_rename(T1,T0,P1):int_min_all(P1,PRI1),
    (int_lt(PRI0,PRI1), WEAK=T1; int_lt(PRI1,PRI0), WEAK=T0),
    +weak_filter_rename(SUM,POS,T,WEAK).

% the filtered entry/exit graphs, S->T should be summarized on loop/call SUM
% or as a global or composite invariant
predicate add_filt_points(SUM:sum,POS:point_pos,S:t_trace,T:t_trace,ASN:t_points_asn).
predicate try_filt_points(SUM:sum,POS:point_pos,S:t_trace,T:t_trace,ASN:t_points_asn).

% T should be renamed to NT in the filtered graph for SUM
predicate trace_rename(in SUM:sum,in POS:point_pos,in T:t_trace,NT:t_trace,ASN:t_points_asn).
?trace_rename(SUM,POS,T,_,_), filter_rename(SUM,POS,T,NT,ASN),
    ~weak_filter_rename(SUM,POS,T,NT), +trace_rename(SUM,POS,T,NT,ASN).
?trace_rename(SUM,POS,T,_,_), ~filter_rename(SUM,POS,T,_,_),
    glob_trace_rename(T,NT), +trace_rename(SUM,POS,T,NT,asn_none{SUM}).

% rename the entry/exit graphs and add each NS->NT to the filtered graphs
asn_points(SUM,POS,S,T,PASN),
    (POS=fn_entry; ~trace_simplify(SUM,drf{S},T)),
    trace_rename(SUM,POS,S,NS,_), trace_rename(SUM,POS,T,NT,RASN),
    (POS=fn_entry, ~trace_simplify(SUM,drf{NS},NT);
     POS=fn_exit, ~trace_simplify(SUM,drf{NS},T)),
    merge_points_asns(PASN,RASN,ASN), +try_filt_points(SUM,POS,NS,NT,ASN).

try_filt_points(SUM,POS,S,T,_),
    \/try_filt_points(SUM,POS,S,T,XA):list_all(XA,[BA|ALIST]),
    merge_points_asns_list(BA,ALIST,ASN),
    +add_filt_points(SUM,POS,S,T,ASN).

% add_filt_points(SUM,POS,S,T) are behaviors that should not be considered
% for per-function summarization
predicate omit_points(SUM:sum,POS:point_pos,S:t_trace,T:t_trace,omit_reason).

% check that a summary trace has been fully simplified
predicate check_simplify(SUM:sum,T:t_trace).
check_simplify(SUM,T), trace_sub(T,SUB,_), trace_simplify(SUM,SUB,NSUB),
    SUB\=NSUB, +warning("Summary trace not fully simplified:",T).

% dropping generated points-to edges

% drop edges based on baked-in analysis behavior
add_filt_points(SUM,POS,S,T,_), drop_spoints(SUM,POS,S,T),
    +omit_points(SUM,POS,S,T,omit_dropped).

% drop edges pointing to constant strings, which are immutable
add_filt_points(SUM,POS,S,T,_), trace_root(S,cstr{_}),
    +omit_points(SUM,POS,S,T,omit_dropped).
add_filt_points(SUM,POS,S,T,_), trace_root(T,cstr{_}),
    +omit_points(SUM,POS,S,T,omit_dropped).

% drop edges pointing to a direct subtrace - these are recursive or identity
add_filt_points(SUM,POS,S,T,_), trace_sub(drf{S},T,_),
    +omit_points(SUM,POS,S,T,omit_dropped).

% drop edges which are redundant given an edge between their subtraces
add_filt_points(SUM,POS,S,drf{T},_), trace_sub(S,drf{SS},R),
    (trace_visible(SUM,SS); POS=fn_entry),
    add_filt_points(SUM,POS,SS,TT,_), trace_compose(SUM,TT,R,T),
    +omit_points(SUM,POS,S,drf{T},omit_dropped).

% drop edges if the traces involved cannot be typed as pointers or composites
add_filt_points(SUM,POS,S,T,_), trace_type(SUM,S,SY),
    (SY=y_ptr{Y0}, trace_type(SUM,T,Y1);
     Y0=SY, T=drf{PT}, trace_type(SUM,PT,Y1)),
    Y0\=y_array{_,_}, Y1\=y_array{_,_}, ~check_cast(Y0,Y1),
    +omit_points(SUM,POS,S,T,omit_dropped).

% drop edges containing casts between incompatible types, which casting.clp
% checks cannot occur
add_filt_points(SUM,POS,S,T,_), trace_bad_cast(SUM,S),
    +omit_points(SUM,POS,S,T,omit_dropped).
add_filt_points(SUM,POS,S,T,_), trace_bad_cast(SUM,T),
    +omit_points(SUM,POS,S,T,omit_dropped).

% find polymorphic composite type invariants

% FIXME

%add_filt_points(SUM,fn_exit,T,DT,_), trace_root(DT,glob{_}),
%    trace_sub(T,STR,TR), trace_type(SUM,STR,y_comp{C}), comp_poly(C,_,TR,_),
%    ~(comp_poly(C,XFTR,TR,_), trace_compose(SUM,STR,XFTR,XFT),
%      ~add_filt_points(SUM,fn_exit,XFT,root{glob{_}},_)),
%    comp_poly(C,FTR,TR,AT), trace_compose(SUM,STR,FTR,FT),
%    add_filt_points(SUM,fn_exit,FT,root{glob{FN}},_),
%    +omit_points(SUM,fn_exit,FT,root{glob{FN}},omit_poly_inv),
%    +add_spoints(td_comp_points{C,STR},FT,root{glob{FN}}),
%    actual_usemod(FN,DT), +add_spoints(td_points{FN,s_func,fn_entry},AT,DT).

%add_filt_points(SUM,fn_exit,T,DT,_), ~trace_root(DT,glob{_}),
%    trace_sub(T,STR,TR), trace_type(SUM,STR,y_comp{C}), comp_poly(C,_,TR,_),
%    cil_curfn(FN), +add_spoints(td_points{FN,SUM,fn_exit},T,DT).
%add_filt_points(SUM,fn_exit,T,DT,_), trace_root(DT,glob{_}),
%    trace_sub(T,STR,TR), trace_type(SUM,STR,y_comp{C}),
%    comp_poly(C,XFTR,TR,_), trace_compose(SUM,STR,XFTR,XFT),
%    ~add_filt_points(SUM,fn_exit,XFT,root{glob{_}},_),
%    cil_curfn(FN), +add_spoints(td_points{FN,SUM,fn_exit},T,DT).
%add_filt_points(SUM,fn_exit,S,T,_),
%    trace_sub(S,STR,RS), trace_type(SUM,STR,y_comp{C}), comp_poly(C,RS,_,_),
%    +omit_points(SUM,fn_exit,S,T,omit_poly_inv),
%    cil_curfn(FN), +add_spoints(td_points{FN,SUM,fn_exit},S,T).

add_filt_points(SUM,POS,S,T,_),
    trace_sub(S,fld{STR,_,C},_), comp_to_relative(STR,S,RS), comp_poly(C,_,RS,_),
    +omit_points(SUM,POS,S,T,omit_poly_inv).

% find composite type invariants

% points-to relationships between traces that have a common prefix generally
% reflect a composite type invariant rather than behavior particular to the
% function in question. identify these, add them to the composite type summary
% and prevent them from entering other summaries

% also valid are points-to relationships where the source is in a structure
% and the target is global data. these are not as clear-cut but are usually
% quite compact and sufficiently precise

% S->T *MAY* be summarized on the comp C given by STR
predicate add_filt_comp_points(C:string,STR:t_trace,S:t_trace,T:t_trace,ASN:t_points_asn).

% get the potential composite type points-to relationships
add_filt_points(SUM,POS,S,T,ASN),
    ~omit_points(SUM,POS,S,T,omit_dropped),
    ~omit_points(SUM,POS,S,T,omit_poly_inv),
    trace_sub(S,fld{STR,_,C},_), comp_to_relative(STR,T,_),
    ~omit_comp_spoints(C,SUM,STR,S,T),
    +omit_points(SUM,POS,S,T,omit_comp_inv),
    POS=fn_exit, +add_filt_comp_points(C,STR,S,T,ASN).

% omit invariants pre-empted by other invariants on internal structs
predicate weak_comp_spoints(C:string,STR:t_trace,S:t_trace,T:t_trace).
add_filt_comp_points(_,STR,S,T,_), trace_sub(STR,SSTR,_), STR\=SSTR,
    add_filt_comp_points(SC,SSTR,S,T,_),
    +weak_comp_spoints(SC,SSTR,S,T).

% add all composite type invariants, omitting the corresponding points-to
% relations for consideration in other summaries
add_filt_comp_points(C,STR,S,T,ASN), ~weak_comp_spoints(C,STR,S,T),
    +check_simplify(s_func,S), +check_simplify(s_func,T),
    +add_spoints_display(td_comp_points{C,STR},S,T,ASN).

%?omit_comp_spoints(C,SUM,STR,S,T),
%    trace_root(S,glob{_}), trace_root(T,glob{_}), ~trace_sub(T,STR,_),
%    +omit_comp_spoints(C,SUM,STR,S,T).

no_comp_glob_points(), +omit_comp_spoints(_,_,_,_,_).

% find global invariants

% points-to relationships between traces that are both global generally reflect
% a global invariant rather than behavior particular to the function in
% question. identify these, add them to the global summary and prevent them
% from entering the function entry/exit summaries
add_filt_points(SUM,POS,S,T,ASN),
    ~omit_points(SUM,POS,S,T,omit_dropped),
    ~omit_points(SUM,POS,S,T,omit_poly_inv),
    ~omit_points(SUM,POS,S,T,omit_comp_inv),
    trace_root(S,glob{G}), trace_root(T,glob{_}), ~omit_glob_spoints(G,S,T),
    +check_simplify(s_func,S), +check_simplify(s_func,T),
    +omit_points(SUM,POS,S,T,omit_glob_inv),
    POS=fn_exit, +add_spoints_display(td_glob_points{G},S,T,ASN).

no_comp_glob_points(), +omit_glob_spoints(_,_,_).

% function entry/exit summaries

% add to the entry/exit summaries any relationships not covered by the
% composite and global invariants
add_filt_points(SUM,POS,S,T,ASN), ~omit_points(SUM,POS,S,T,_),
    sum_bound(SUM,_,_), cil_curfn(FN),
    +check_simplify(SUM,S), +check_simplify(SUM,T),
    ~trace_root(S,temp{_,_}), trace_visible(SUM,S), trace_visible(SUM,T),
    +add_spoints_display(td_points{FN,SUM,POS},S,T,ASN).
add_filt_points(s_call{I},POS,S,T,ASN), ~omit_points(s_call{I},POS,S,T,_),
    +check_simplify(s_call{I},S), +check_simplify(s_call{I},T),
    dircall(I,FN), +add_spoints_display(td_points{FN,s_func,POS},S,T,ASN).
add_filt_points(s_call{I},POS,S,T,ASN), ~omit_points(s_call{I},POS,S,T,_),
    +check_simplify(s_call{I},S), +check_simplify(s_call{I},T),
    anycall(I,FN,indirect), actual_usemod(FN,drf{S}), actual_usemod(FN,T),
    +add_spoints_display(td_points{FN,s_func,POS},S,T,ASN).

% add sunify for any accessed traces in each summary that were inside
% a malloc and renamed at exit
filter_rename(SUM,fn_exit,T,NT,_), ~weak_filter_rename(SUM,fn_exit,T,NT),
    cil_curfn(F), +sum_exit(F,SUM)->sunify(T,NT).

% baked in analysis behavior

import "aliasingbaked.clp".

% checks

%?lval(P,LV,_,_), ~lval(P,LV,_,_),
%    cil_curfn(FN), point_location(P,_,LINE),
%    lval_string(LV,S), +warning("Could not evaluate LV",FN,LINE,P,LV,S).
%?eval(P,E,_,_), ~eval(P,E,_,_),
%    cil_curfn(FN), point_location(P,_,LINE),
%    exp_string(E,S), +warning("Could not evaluate EXP",FN,LINE,S).
%?beval(P,E,_,_), ~beval(P,E,_,_),
%    cil_curfn(FN), point_location(P,_,LINE),
%    exp_string(E,S), +warning("Could not evaluate BEXP",FN,LINE,S).
