%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% Loop handling in CFGs.

import "cilcfg.clp".

% PREDICATES

% split func/loop SUM has entry and exit points PIN and POUT
predicate sum_bound(SUM:sum,PIN:pp,POUT:pp).

% P is in the split body of func/loop SUM, i.e. not contained in any inner loop
predicate sum_body(SUM:sum,P:pp).

% get per-func/loop ordering information DEPTH, the depth of SUM within any
% loop nestings. Using DEPTH in $order will prioritize evaluation of
% summaries that are nested deeper
predicate sum_order(SUM:sum,DEPTH:int).

% get per-program-point ordering information DEPTH and DISTANCE. DEPTH is the
% depth of the func/loop containing P as in sum_order, and DISTANCE is the
% distance of P into the CFG for its containing func/loop. Using DEPTH/DISTANCE
% with $order will prioritize evaluation of points that are nested in
% deeper loops and are earlier in the CFG
predicate pp_order(in P:pp,DEPTH:int,DISTANCE:int).

% P0 will be processed before P1 under the pp_order stratification
predicate pp_order_prec(in P0:pp,in P1:pp).

% P1 is reachable from P0 (ignoring back edges)
predicate pp_reach(in P0:pp,P1:pp).

% P1 is dominated by P0 (ignoring back edges)
predicate pp_dominate(in P0:pp,P1:pp).

% LOOP CONSTRUCTION

% default behavior for loops is to just break the back edges.
% - preserve_loops() preserves the back edges so the final CFG may be cyclic
% - split_loops() splits each loop off into a separate CFG
predicate preserve_loops().
predicate split_loops().

% the default for loops that are split off is to pick points for the loop body
% that minimize the size of the split body. this includes all points which are
% dominated by the loop head and can also reach one of the loop's back edges.
% these can be specified *in addition* to split_loops() to change this behavior
% - nesting_split_loops() picks a body that best approximates the actual
% statement/block nesting of the source code. thus 'return' or other statements
% which cannot reach a back edge will still be in the loop body if they are
% nested within the source program's while/for loop.
% - maximize_split_loops() picks a body of maximum size, including all points
% dominated by the loop head regardless of whether they can reach a back edge
predicate nesting_split_loops().
predicate maximize_split_loops().

% RULES

% nodes and edges during loop splitting
predicate cfgn_loop(P:pp,N:cfg_node).
predicate cfge_loop(P0:pp,P1:pp,E:cfg_edge).

% loop entry points

% drop duplicate loop heads sharing the same entry point (which we might get
% after denop'ing) and pick a canonical one to use
predicate dupe_loop(P:pp,PH:loop).
cfgn(P,nlooph{PH0}), cfgn(P,nlooph{PH1}), lessthan(PH0,PH1), +dupe_loop(P,PH1).

% P is the entry point of potential loop PH
predicate pos_lentry(PH:loop,P:pp).
cfgn(P,nlooph{PH}), ~dupe_loop(P,PH), +pos_lentry(PH,P).

% loop dominators

% potential loop PH does NOT dominate P
predicate nodom(PH:loop,P:pp).

% no loops dominate the entry point
cfgn(P,nentry), cfgn(_,nlooph{PH}), +nodom(PH,P).

% propagate nodom through the CFG, stopping once the loop head is reached
cfge(P0,P1,_), nodom(PH,P0), ~pos_lentry(PH,P0), +nodom(PH,P1).

% loop body points

% P is in the body of loop PH
predicate lbody(PH:loop,P:pp).

% the edge (P0,P1,E) is a back edge on (real) loop PH
predicate backedge(PH:loop,P0:pp,P1:pp,E:cfg_edge).

% identify back edges, those going to a loop head whose source is dominated by
% that loop head. include the degenerate case with self edges on the head
% (in which case nodom is generated for the head)
pos_lentry(PH,P1), cfge(P0,P1,E), ~nodom(PH,P0), +backedge(PH,P0,P1,E).
pos_lentry(PH,P), cfge(P,P,E), +backedge(PH,P,P,E).

% make sure we got all the back edges. any others are due to irreducible loops
% which we don't handle yet.
predicate lreach(PH:loop,P:pp).
predicate irreducible().
cfge(P0,P1,_), ~backedge(_,P0,P1,_), lreach(PH,P0), +lreach(PH,P1).
pos_lentry(PH,P), +lreach(PH,P), cfge(P0,P,_),
    lreach(PH,P0), ~backedge(PH,P0,P,_),
    +warning("Irreducible loop detected"), +irreducible().

% P is the entry point of real loop PH which has a back edge. loops without
% back edges may result from some degenerate cases (do {...} while (0)) or
% 'goto' statements which don't create a loop. also update loophead()
% now that we know where the real loops are
predicate lentry(PH:loop,P:pp).
pos_lentry(PH,P), backedge(PH,PB,_,_),
    +loophead(PH), +lentry(PH,P), +lbody(PH,PB).
pos_lentry(PH,P), degenerate_infinite_loop(P),
    +loophead(PH), +lentry(PH,P).

% propagate lbody information back to the loop entry point
lentry(PH,P), lbody(PH,P1), P1\=P, cfge(P0,P1,_), +lbody(PH,P0).

% add lbody information for special next/trec/sum/term points
lentry(PH,P), +lbody(PH,P), lbody(OPH,P),
    +lbody(OPH,p_lnext{PH}), +lbody(OPH,p_ltrec{PH}),
    OPH\=PH, +lbody(OPH,p_lsum{PH}).

% if we are matching the split loops to statement nestings,
% add any potential loop body points that are nested inside the loop itself.
% note that this will only work for while/for loops, not goto-created loops
predicate edge_nest(in E:cfg_edge,PH:loop).
edge_nest(easm{I},PH) :- instr_nest(I,S), stmt_nest_stmt(S,PH).
edge_nest(eset{I},PH) :- instr_nest(I,S), stmt_nest_stmt(S,PH).
edge_nest(ecall{I},PH) :- instr_nest(I,S), stmt_nest_stmt(S,PH).
edge_nest(ebtrue{E},PH) :- branch_nest(E,S), stmt_nest_stmt(S,PH).
edge_nest(ebfalse{E},PH) :- branch_nest(E,S), stmt_nest_stmt(S,PH).
nesting_split_loops(), lreach(PH,P), ~nodom(PH,P), ~cfgn(P,nexit),
    cfge(P,_,E), edge_nest(E,PH), +lbody(PH,P).

% if we are maximizing the size of split loops, add lbody for any point
% dominated by the loop entry (except the function exit)
maximize_split_loops(), lreach(PH,P), ~nodom(PH,P), ~cfgn(P,nexit),
    +lbody(PH,P).

% loop nodes/edges

% the loop CFG is that remaining after dropping back edges
cfgn(P,N), N\=nlooph{_}, +cfgn_loop(P,N).
cfgn(P,nlooph{PH}), lentry(PH,P), +cfgn_loop(P,nlooph{PH}).
cfge(P0,P1,E), ~backedge(_,P0,P1,E), +cfge_loop(P0,P1,E).
cfge(P0,P1,E), backedge(PH,P0,P1,E), +cfge_loop(P0,p_lnext{PH},E).

% remove cfg_loop nodes/edges going to the loop head, replace with summary
predicate omit_cfgn_splitsum(P:pp,N:cfg_node).
predicate omit_cfge_splitsum(P0:pp,P1:pp,E:cfg_edge).

% remove edges exiting each loop, replace with p_lterm edges
predicate cfge_xloop(P0:pp,P1:pp,E:cfg_edge).
predicate omit_cfge_splitterm(P0:pp,P1:pp,E:cfg_edge).
predicate omit_cfge_splitnest(P0:pp,P1:pp,E:cfg_edge).

% loop preservation

% if we're generating a loop-containing final CFG, finalize the original
% nodes/edges and mark the function as irreducible so we don't try to
% do any splitting or back edge breaking
preserve_loops(), cfgn(P,N), +cfgn_final(P,N).
preserve_loops(), cfge(P0,P1,E), +cfge_final(P0,P1,E).
preserve_loops(), +irreducible().

% loop splitting

% add the tail recursive invocation for each split loop
split_loops(), lentry(PH,P), ~degenerate_infinite_loop(P),
    +cfge_xloop(p_lnext{PH},p_ltrec{PH},eloop{PH}).
split_loops(), lentry(PH,P), degenerate_infinite_loop(P),
    +cfge_xloop(P,p_ltrec{PH},eloop{PH}).

% to split loops off from the rest of the function body, the loop head PH
% is first replaced with a summary point p_lsum{PH}. for each edge exiting
% the loop to P, an iloop is added from p_lsum{PH} to P, and the target is
% then replaced with the termination point p_lterm{PH,P}, isolating the loop
% body from the rest of the function

% replace the loop head with its summary point
split_loops(), lentry(PH,P), cfgn_loop(P,N), N\=nlooph{PH},
    +omit_cfgn_splitsum(P,N), +cfgn_loop(p_lsum{PH},N).
split_loops(), lentry(PH,P), cfge_loop(P0,P,E),
    +omit_cfge_splitsum(P0,P,E), +cfge_loop(P0,p_lsum{PH},E).
cfge_loop(P0,P1,E), ~omit_cfge_splitsum(P0,P1,E), +cfge_xloop(P0,P1,E).

% P is an lterm point for PH or a loop within PH
predicate inner_lterm(in PH:loop,in P:pp).
inner_lterm(PH,p_lterm{OPH,P}) :- lentry(OPH,EP), lbody(PH,EP).

% replace exit edges with edges to p_lterm, and add summary edges to the target
split_loops(), lentry(PH,_), lbody(PH,P0), cfge_xloop(P0,P1,E),
    ~lbody(PH,P1), ~inner_lterm(PH,P1),
    +omit_cfge_splitterm(P0,P1,E),
    +cfge_xloop(p_lsum{PH},P1,eloop{PH}), +cfge_xloop(P0,p_lterm{PH,P1},E).

% for p_lterm edges exiting from within nested loops, we generate p_lterm edges
% for each level of nesting. get rid of all except for the innermost nesting
cfge_xloop(P0,p_lterm{PH,P1},E), cfge_xloop(P0,p_lterm{PH,p_lterm{_,P1}},E),
    +omit_cfge_splitnest(P0,p_lterm{PH,P1},E).

% finalizing

~irreducible(), cfgn_loop(P,N), ~omit_cfgn_splitsum(P,N), +cfgn_final(P,N).
~irreducible(), cfge_xloop(P0,P1,E), ~omit_cfge_splitterm(P0,P1,E),
    ~omit_cfge_splitnest(P0,P1,E), +cfge_final(P0,P1,E).

% add iloops aborting the program if a loop has no exit edges (is infinite)
~irreducible(), split_loops(), lentry(PH,_), ~cfge_xloop(_,p_lterm{PH,_},_),
    +cfge_final(p_lsum{PH},p_abort,eloop{PH}).

% get a sum_bound() for the entire function do this after finalization because
% the entry point may also be a loop head and can change during splitting
entry(PIN), exit(POUT), +sum_bound(s_func,PIN,POUT).

% get a sum_bound() for the tail recursive loop exit
~irreducible(), split_loops(), lentry(PH,P),
    +sum_bound(s_loop{PH},P,p_ltrec{PH}),

% get a sum_bound() for loop exit points within nestings
~irreducible(), split_loops(), lentry(PH,P), cfge_final(_,p_lterm{PH,P1},_),
    +sum_bound(s_loop{PH},P,p_lterm{PH,P1}).

% compute sum_body by following edges forward from func/loop entry points
sum_bound(SUM,P,_), +sum_body(SUM,P).
cfge_final(P0,P1,_), sum_body(SUM,P0), +sum_body(SUM,P1).

% get locations for loop end points
sum_bound(_,_,p_ltrec{PH}), cil_stmt_x_location(PH,FILE,LINE),
    +point_location(p_ltrec{PH},FILE,LINE).
sum_bound(_,_,p_lterm{PH,P}), point_location(P,FILE,LINE),
    +point_location(p_lterm{PH,P},FILE,LINE).

% point processing order

using intops.

% Depth for each summary when iterating over cil_body. The summaries form a 
% tree so just crawl it.
+sum_order(s_func,0).
iloop(P,_,PH), sum_body(SUM,P), SUM\=s_loop{PH},
    sum_order(SUM,N), int_sub(N,1,M), +sum_order(s_loop{PH},M).

% If we are iterating over cil_sum_body sessions there is only one visible
% loop/function body, so assign it depth 0.
cil_cursum(Sum), +sum_order(Sum, 0).

% point distance from func/loop entry. the CFG is a DAG so there may be many
% distances added for the same point. we will find the maximum, ensuring that
% each point has a distance less than all its successors
predicate pp_distance(pp,int).
~irreducible(), sum_bound(_,P,_), +pp_distance(P,0).
cfge_final(P0,P1,_), pp_distance(P0,N), int_add(N,1,M), +pp_distance(P1,M).

% depth and distance for each point
sum_body(SUM,P), sum_order(SUM,DEP), \/pp_distance(P,D):int_max_all(D,DIST),
    +pp_order(P,DEP,DIST).

% compare two program points based on DEP/DIST
pp_order_prec(P0,P1) :- pp_order(P0,DEP0,DIST0), pp_order(P1,DEP1,DIST1),
    (int_lt(DEP0,DEP1); DEP0=DEP1, int_lt(DIST0,DIST1)).

% point reachability

% compute point reachability
+pp_reach(P,P).
pp_reach(P0,P1) :- pp_reach(P0,MP), cfge_final(MP,P1,_).

% compute point dominators by taking the negation of pp_nodom
predicate pp_nodom(in P0:pp,P1:pp).
pp_nodom(P0,P1) :- sum_body(SUM,P0), sum_bound(SUM,P1,_), P0\=P1.
pp_nodom(P0,P1) :- pp_nodom(P0,MP), cfge_final(MP,P1,_), P0\=P1.
pp_dominate(P0,P1) :- sum_body(SUM,P0), sum_body(SUM,P1), ~pp_nodom(P0,P1).
