%%
% Authors:
%   Peter Hawkins  <hawkinsp@cs.stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

%-----------------------------------------------------------------------------%
% sumbody.clp
%
% Create a new session cil_sum_body(FN,SUM) that contains a single loop-free
% CFG for either a function's outer body or one of its inner loops.
%
% Since all intermediate predicates are deleted when fixpointing over separate
% sessions, analyses that handle cil_sum_body rather than cil_body do not need
% to worry about invalidated data when propagating both top-down and bottom-up,
% or when handling tail recursive loop invocations.
%
% cil_sum_body() contains only the CFG for a particular loop or outer function
% body. A special predicate cil_cursum(Sum) is defined to indicate which
% summary is currently being analyzed.
%
% The CFG construction code understands when it is being run over a 
% cil_sum_body session and modifies its behaviour accordingly.
%
%-----------------------------------------------------------------------------%

import "loops.clp".

analyze session_name("cil_body").

+split_loops().

    % Copy the CFG into each session, together with summary bounds for
    % just one summary.
cil_curfn(FN), sum_bound(SUM,PIN,POUT),
    +cil_sum_body(FN,SUM)->cil_curfn(FN),
    +cil_sum_body(FN,SUM)->cil_cursum(SUM),
    +cil_sum_body(FN,SUM)->sum_bound(SUM,PIN,POUT).
cil_curfn(FN), sum_body(SUM,P), cfgn_final(P,N),
    +cil_sum_body(FN,SUM)->cfgn_final(P,N).
cil_curfn(FN), sum_body(SUM,P0), cfge_final(P0,P1,E),
    +cil_sum_body(FN,SUM)->cfge_final(P0,P1,E).

    % branch_loc and loophead are needed to produce program point location
    % information.
cil_curfn(FN), sum_body(SUM,P), branch(P,_,_,E), branch_loc(E,File,Line),
    +cil_sum_body(FN,SUM)->branch_loc(E,File,Line).
cil_curfn(FN), sum_bound(SUM,_,_), loophead(PH),
    +cil_sum_body(FN,SUM)->loophead(PH).

%-----------------------------------------------------------------------------%

    % Copy CIL predicates that were created by the CFG construction process
    % into the cil_sum_body session.
    % XXX: It seems the stratification analysis is buggy, so we copy all of
    % the predicates without looking to see whether they are already present
    % in cil_body.
cil_type_int(A,B), cil_curfn(F), ~cil_body(F)->cil_type_int(A,B),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_type_int(A,B).
cil_type_x_bytes(A,B), cil_curfn(F), ~cil_body(F)->cil_type_x_bytes(A,B),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_type_x_bytes(A,B).

cil_var_name(A,B,C), cil_curfn(F), ~cil_body(F)->cil_var_name(A,B,C),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_var_name(A,B,C).
cil_var_return(A), cil_curfn(F), ~cil_body(F)->cil_var_return(A),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_var_return(A).
cil_var_temporary(A), cil_curfn(F), ~cil_body(F)->cil_var_temporary(A),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_var_temporary(A).

cil_lval_var(A,B,C), cil_curfn(F), ~cil_body(F)->cil_lval_var(A,B,C),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_lval_var(A,B,C).

cil_exp_lval(A,B), cil_curfn(F), ~cil_body(F)->cil_exp_lval(A,B),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_exp_lval(A,B).
cil_exp_binop(A,B,C,D,E), cil_curfn(F), ~cil_body(F)->cil_exp_binop(A,B,C,D,E),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_exp_binop(A,B,C,D,E).

cil_off_none(A), cil_curfn(F), ~cil_body(F)->cil_off_none(A),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_off_none(A).

cil_instr_set(A,B,C), cil_curfn(F), ~cil_body(F)->cil_instr_set(A,B,C),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_instr_set(A,B,C).
cil_instr_call(A,B), cil_curfn(F), ~cil_body(F)->cil_instr_call(A,B),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_instr_call(A,B).
cil_instr_call_ret(A,B), cil_curfn(F), ~cil_body(F)->cil_instr_call_ret(A,B),
    sum_bound(Sum,_,_), +cil_sum_body(F,Sum)->cil_instr_call_ret(A,B).
cil_instr_call_arg(A,B,C), cil_curfn(F), sum_bound(Sum,_,_), 
    ~cil_body(F)->cil_instr_call_arg(A,B,C), 
    +cil_sum_body(F,Sum)->cil_instr_call_arg(A,B,C).
cil_instr_x_location(A,B,C), cil_curfn(F), sum_bound(Sum,_,_), 
    ~cil_body(F)->cil_instr_x_location(A,B,C),
    +cil_sum_body(F,Sum)->cil_instr_x_location(A,B,C).

%-----------------------------------------------------------------------------%

    % Add process ordering dependencies from the current function to the
    % set of calls and loops that are performed.
icall(P0,_,I), cil_curfn(Fn), sum_body(Sum, P0), anycall(I, Callee, _),
    +cil_sum_body(Fn, Sum)->process_dependency("cil_sum_body", Callee, s_func).
iloop(P0,_,PH), cil_curfn(Fn), sum_body(Sum, P0),
    +cil_sum_body(Fn, Sum)->process_dependency("cil_sum_body", Fn, s_loop{PH}).

%-----------------------------------------------------------------------------%
