import "../return_correlate.clp".
+split_loops().
+print_correlations().
analyze session_name("cil_body").


% ----------------------------------------------------------------------------
% This run file uses some of the functionality of null.clp to test correlate.clp
% and return_correlate.clp.
%
% correlate.clp functionality tested: For every dereferenced pointer that would be included in
% the null function summary, see if there is a necessary condition for the dereference to occur.
% For example:
% if(flag && foo()) *p = 8;
% if(flag && bar()) *p = 6;
% should find that flag is a necessary condition for the deref of p.
%
% return_correlate.clp functionality tested: 
% Higher-level property for null: Is an argument set to a null or non-null value on function return.
% Check if there is a one-to-one correlation between a parameter being initialized to a non-null value
% and the class of possible return values. For example, does a positive return value 
% imply that a pointer p is non-null and does p being non-null imply a positive value.

% -------------------------------------------------------------------------
% This is just the functionality already existing in null.clp.
% If you are trying to understand  correlate and return_correlate.clp, this part 
% is not very important.
% If you are interested in using return_correlate.clp, read the part called
% "communication with return_correlate interface" below.


% guarded_loc_type from null.clp
type guarded_loc_type ::= guarded_loc{TRACE: t_trace_val, GUARD: g_guard}.


% Trace T that occurs in expression E under guard G is not checked
% for being null at program point OP.
predicate unchecked_loc(E:c_exp, T:t_trace_val, G:g_guard, OP: pp).


% Externally visible trace V should be included in the function summary. 
predicate funsum_var(V:t_trace, TRACE_GUARD:g_guard, STMT_GUARD:g_guard, CALL_STACK:list[string]).

% Final version of funsum var to be included in the summary. If there are multiple dereferences
% of the same variable under different guards, we collapse them into one entry in the function summary
% by or'ing all the guards under which the dereference is possible.
predicate filtered_funsum(V:t_trace, TRACE_GUARD:g_guard, STMT_GUARD:g_guard, CALL_STACK:list[string]).

% At program point P, there is a potential null dereference of trace T.
% Potential_deref here means that there is no statement guard that prevents
% this trace from evaluating to null; it does *not* mean that we know it has been
% set to null...
predicate potential_deref(P:pp, T:t_trace, TRACE_GUARD:g_guard, STMT_GUARD:g_guard).


% Don't mark a trace as potential null deref if the statement guard
% implies a return condition which in turn guarantees the pointer can never be null.
predicate return_checks(in P:pp, in T:t_trace).
return_checks(P, T) :- return_implication(P, T, "non_null", CG), guard(P, SG), 
	#sat(SG), #not(CG, NCG), #and(SG, NCG, TESTG), ~guard_sat(TESTG). 

cil_lval_mem(_, E, _),  exp_point(E,P), guard(P,G), 
 eval(P, E, trace{drf{T}}, TG), 
 ~return_checks(P, T),
 evals(P, E, IE), 
 g_eqz(IE, ISZ), g_nez(IE, NEZ),
 #and(G,NEZ,TESTG1), guard_sat(TESTG1), %not a certain dereference
 #and(G,ISZ,TESTG), guard_sat(TESTG),
 #and(TG, G,  RG), guard_sat(RG),
 #simplify(TG, STG),
 +unchecked_loc(E, trace{T}, STG, P),
 +print("UNCHECKED LOCATION: ", T),
 +potential_deref(P, T, STG, G).


potential_deref(_, L, TG, SG), 
   trace_root(L, arg{_}),
   +funsum_var(L, TG,SG, nil).


potential_deref(_, L, TG, SG), 
   trace_root(L, glob{_}),
   +funsum_var(L, TG,SG, nil).

funsum_var(X, TG,_, CALL_STACK), \/funsum_var(X, TG,G, CALL_STACK):#or_all(G, GR), +filtered_funsum(X, TG, GR, CALL_STACK).

% Are there necessary or sufficient conditions for the dereference of trace T?
% Use necessary_cond and sufficient_cond predicates from correlate.clp.
filtered_funsum(T, _, G, _), necessary_cond(G, s_func, CG), 
	+print(CG, "IS A NECESSARY CONDITION FOR THE DEREFERENCE OF", T).
filtered_funsum(T, _, G, _), sufficient_cond(G, s_func, CG), 
	+print(CG, "IS A SUFFICIENT CONDITION FOR THE DEREFERENCE OF", T).


%------------------------------------------------------------------


% Predicates get_null_guard and get_non_null-guard.
% ---------------------------------------------------
% get_null_guard: retrieve guard under which T evaluates to NULL.
predicate get_null_guard(in P:pp, in T:t_trace, NG:g_guard).

% get_non_null_guard: Retrieve guard under which T evaluates to a non-null value.
predicate get_non_null_guard(in P:pp, in T:t_trace, NG:g_guard).





% T is visible outside of SUM, but it's root is not the return value for this function. 
predicate visible_non_return(in SUM: sum, in T:t_trace).
visible_non_return(SUM, T) :- ~trace_root(T, return), trace_visible(SUM, T). 

predicate get_null_guard(in P:pp, in T:t_trace, NG:g_guard).
predicate get_non_null_guard(in P:pp, in T:t_trace, NG:g_guard).

% ---get_null_guard: Retrieves the guard under which T evaluates to NULL. ------

% Case 1: points-to trace involves a drf.
get_null_guard(P, T, NG) :- val(P, T, trace{drf{S}}, TG), vals(P, S, IE), g_eqz(IE, EQZ),#and(EQZ, TG, NG).

% Case 2: T is set to the address of another trace such that its val does not have a deref.
get_null_guard(P, T, NG) :- val(P, T, trace{S}, _), ~equal(S, drf{_}), #bool_g(false, NG).

% Case 3: T evaluates NULL.
get_null_guard(P, T, NG) :- val(P, T, nrep{BV}, NG),  g_nez(BV, NEZ), ~guard_sat(NEZ).

% -- get_non_null_guard -----

% Case 1: points-to trace involves a drf.
get_non_null_guard(P, T, NG) :- val(P, T, trace{drf{S}}, TG), vals(P, S, IE), g_nez(IE, NEZ),#and(NEZ, TG, NG).

% Case 2: T is set to the address of another trace such that its val does not have a deref.
get_non_null_guard(P, T, NG) :- val(P, T, trace{S}, NG), ~equal(S, drf{_}).

% Case 3: T evaluates NULL.
get_non_null_guard(P, T, NG) :- val(P, T, nrep{BV}, _), g_nez(BV, NEZ), ~guard_sat(NEZ), #bool_g(false, NG).

% ----------------------------------------------------------------------------------------------------

predicate unified_null_guard(T:t_trace, NG:g_guard).
predicate unified_non_null_guard(T:t_trace, NNG:g_guard).

predicate null_guard(T:t_trace, NG:g_guard).
predicate non_null_guard(T: t_trace, NNG: g_guard).

% Property of interest: Does a certain class of return values correlate with a
% parameter or global being definitely null or non-null as a result of a write?
% "correlate" here means: nullness of p implies a certain kind of return value 
% and return value implies that p is null. (same property for non-null).
% For this property, we are only interested in looking for correlations 
% between traces that have write access to them and return values.
% The "state space" for searching for return correlations is therefore only 
% traces that have a write access to them.

exit(P), access(_, T, write),   visible_non_return(s_func, T),
	get_null_guard(P, T, NG), +null_guard(T, NG),
	get_non_null_guard(P, T, NNG), +non_null_guard(T, NNG).


% unify null_guards and non_null guards for a certain trace.
null_guard(T, _), \/null_guard(T, NG):#or_all(NG, G), +unified_null_guard(T, G).
non_null_guard(T, _), \/non_null_guard(T, NG):#or_all(NG, G), +unified_non_null_guard(T, G).


% --------------------------------------------------------------------------------
% COMMUNICATION WITH "return_correlate" interface:
% The guard provided by query_return_correlation comes from correlate.clp.
% If a predicate query_return_correlation(RG) occurs, return_corrleate.clp 
% identified a class of return values that will only be returned if RG is 
% satisifed.
% If there is a correlation between this return guard RG and the initialization 
% of a pointer to a non-null value (or the failure of initialization),
% add the predicate return_correlation(RG, T, PROPERTY) so that 
% return_correlate.clp can include this correlation in its function summary. 
% return_correlation(RG, T, PROPERTY) means: Under the return condition guard RG, 
% there is a restriction on T as can be interpreted by the analysis according to
% the string "PROPERTY". In this case, properties "null" and "non-null" on a certain 
% trace T impose the restriction that T is null/non-null if the return condition
% is satisfied.

% TRACE T==NULL IMPLIES RET COND IS MET (T==NULL is sufficient for the return condition)
% IF THE RETURN CONDITION IS MET, T==NULL HOLDS. (T==NULL is necessary for the return condition)
query_return_correlation(TYPE, RG), unified_null_guard(T, NG), 
	#not(RG, NRG), #and(NRG, NG, SUFFICIENT), ~guard_sat(SUFFICIENT), %sufficient
	#not(NG, NNG), #and(NNG, RG, NECESSARY), ~guard_sat(NECESSARY), %necessary
	+return_correlation(TYPE, T, "null").
	


query_return_correlation(TYPE, RG), unified_non_null_guard(T, NG),
	%+print("HERE: (T, RG, NG)", T, "***", RG, "***", NG), 
	#not(RG, NRG), #and(NRG, NG, SUFFICIENT), ~guard_sat(SUFFICIENT),%sufficient
	#not(NG, NNG), #and(NNG, RG, NECESSARY), ~guard_sat(NECESSARY), %necessary
	+return_correlation(TYPE, T, "non_null").	
		
	
% ----------------------------------------------------------------------
% Visible trace T evaluates to NULL under guard EQZ.
predicate visible_trace_eqz(T:t_trace, EQZ: g_guard).

% Property: Is return value NULL if one input is NULL? 
branch(P, _, _, E), evals(P, E, IE), scalar_trace_used(IE, T), 
	visible_non_return(s_func, T), visible_non_heap_trace(s_func, T), 
	trace_type(s_func, T, TYPE), type_scalar_kind(TYPE, KIND), 
	traceval_scalar(P, no, trace{T},  KIND, SCALAR), 
	g_eqz(SCALAR, EQZ), 
	+visible_trace_eqz(T, EQZ).
	

% Does EQZ imply null return? (i.e is an input being NULL a sufficient condition
% for return value being NULL?)
query_return_correlation(zero, ZRG), visible_trace_eqz(T, EQZ), 
	#not(ZRG, NZRG), #and(EQZ, NZRG, TESTG), ~guard_sat(TESTG),
	equal(T, drf{S}), %the traces used involve one too many derefs, strip it.
	+return_correlation(zero, S, "null_input_null_return").
