%%
% Authors:
%   Isil Dillig    <isil@stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% Utility predicates to help with scalar and guard manipulations.

import "../memory/scalar.clp".




% given an integer expression OrigIE, construct a new integer
% expression NewIE which replaces all occurences of trace 
% OldTrace with NewTrace.
 predicate scalar_trace_rename(in OrigIE:scalar, in OldTrace: t_trace, in
NewTrace:t_trace, out NewIE:scalar).

% Given a vr, repplace all traces OldTrace with NewTrace.
predicate vr_trace_rename(in OrigVR:g_vecrep, in OldTrace:t_trace, in NewTrace:t_trace, out NewVR:g_vecrep).

predicate strename_list(in GS: list[guarded_scalar], in OldTrace:t_trace, in NewTrace: t_trace, NewGS:  list[guarded_scalar]).


% -------------------------------------------------------------------------------
%  vr_trace_rename rules

% Note: X's are intentially used to represent anything we don't care
% about...  
vr_trace_rename(vr_t{OldTrace},
	OldTrace, NewTrace, vr_t{NewTrace}) :- .

% Keep original trace if specified trace does not occur.
vr_trace_rename(vr_t{Trace},
	OldTrace, NewTrace, vr_t{Trace}) :- 
	~equal(Trace, OldTrace).

vr_trace_rename(vr_st{OldTrace, P, E}, OldTrace, 
	NewTrace, vr_st{NewTrace, P, E}) :- .

vr_trace_rename(vr_st{Trace, P, E}, OldTrace, 
	NewTrace, vr_st{Trace, P, E}) :- 
	~equal(Trace, OldTrace).



vr_trace_rename(X, Old, New,X):- equal(X, vr_i{_,_}).      


vr_trace_rename(vr_c{X,P,Old}, Old, New, vr_c{X, P, New}) :-.
vr_trace_rename(vr_c{X,P,T},
	OldTrace, NewTrace, vr_c{X, P, T}) :- 
	~equal(T, OldTrace). 

vr_trace_rename(E, Old, New, vr_instscalar{S, P, NewGV}) :-
	equal(E, vr_instscalar{S, P, GV}),
	vr_trace_rename(GV, Old, New, NewGV).
	


vr_trace_rename(vr_buffersize{Old}, Old, New, vr_buffersize{New}) :-.   
vr_trace_rename(vr_buffersize{T}, Old, New, vr_buffersize{T}) :- ~equal(T, Old).   


vr_trace_rename(vr_u{IE}, Old, New, vr_u{NewIE}) :-
	scalar_trace_rename(IE, Old, New, NewIE).

% ---------------------------------------------------------------------------

% scalar_trace_rename rules

scalar_trace_rename(IE, OldTrace, NewTrace, IE) :- equal(IE,  s_int_const{_, _,_}).   
scalar_trace_rename(IE, OldTrace,NewTrace, IE):-  equal(IE, s_float_const{_, _}).

scalar_trace_rename(s_unconstrained{VR, Y}, Old, New, s_unconstrained{NewVR, Y}) :-
	vr_trace_rename(VR, Old, New, NewVR).


scalar_trace_rename(s_cast{IE, X}, OldTrace, NewTrace, 
	s_cast{NewIE,X}) :-
    	scalar_trace_rename(IE, OldTrace, NewTrace, NewIE).

?scalar_trace_rename(s_guard{GS}, OldTrace, NewTrace, _),
	map_to_sorted_list(GS, GSList),
	strename_list(GSList, OldTrace, NewTrace, NewGSList),
	map_of_list(NewGSList, NewGS),
	+scalar_trace_rename(s_guard{GS}, OldTrace, NewTrace, s_guard{NewGS}).

?strename_list(List1, OldTrace, NewTrace, _),
	List1 = [pair{GS1, Guard}|Rest1],
	scalar_trace_rename(GS1, OldTrace, NewTrace, GS2),
	strename_list(Rest1, OldTrace, NewTrace, Rest2),
	List2 = [pair{GS2, Guard}|Rest2],
	+strename_list(List1, OldTrace, NewTrace, List2).

+strename_list([], _, _, []).

scalar_trace_rename(s_unop{X, IE}, OldTrace, NewTrace, s_unop{X, NewIE}) :-
    scalar_trace_rename(IE, OldTrace, NewTrace, NewIE).

scalar_trace_rename(s_binop{X, LE, RE}, OldTrace, NewTrace, s_binop{X, NewLE, NewRE}) :-
    scalar_trace_rename(LE, OldTrace, NewTrace, NewLE), 
    scalar_trace_rename(RE, OldTrace, NewTrace, NewRE).

% ---------------------------------------------------------------------------

predicate scalar_trace_to_scalar(in OrigIE:scalar, in OldTrace: t_trace, in
Scalar:scalar, out NewIE:scalar).

predicate tr2sc_list(in GS: list[guarded_scalar], in OldTrace:t_trace, in Scalar: scalar, NewGS:  list[guarded_scalar]).

% scalar_trace_to_scalar rules rules

scalar_trace_to_scalar(IE, OldTrace,Scalar, IE) :- equal(IE,  s_int_const{_, _,_}).   
scalar_trace_to_scalar(IE, OldTrace,Scalar, IE):-  equal(IE, s_float_const{_, _}).

scalar_trace_to_scalar(s_unconstrained{vr_t{T}, _}, T, Scalar, Scalar) :- .


scalar_trace_to_scalar(s_unconstrained{vr_st{T, _, _}, _}, T, Scalar, Scalar) :- .

scalar_trace_to_scalar(s_unconstrained{X, Y}, T, Scalar, s_unconstrained{X, Y}) :- 
	~(equal(X, vr_t{T}); equal(X, vr_st{T,_,_})).


scalar_trace_to_scalar(s_cast{IE, X}, OldTrace, Scalar, 
	s_cast{NewIE,X}) :-
    	scalar_trace_to_scalar(IE, OldTrace, Scalar, NewIE).

?scalar_trace_to_scalar(s_guard{GS}, OldTrace, Sc, _),
	map_to_sorted_list(GS, GSList),
	tr2sc_list(GSList, OldTrace, Sc, NewGSList),
	map_of_list(NewGSList, NewGS),
	+scalar_trace_to_scalar(s_guard{GS}, OldTrace,Sc, s_guard{NewGS}).

?tr2sc_list(List1, OldTrace, Sc, _),
	List1 = [pair{GS1, Guard}|Rest1],
	scalar_trace_to_scalar(GS1, OldTrace, Sc, GS2),
	tr2sc_list(Rest1, OldTrace,Sc, Rest2),
	List2 = [pair{GS2, Guard}|Rest2],
	+tr2sc_list(List1, OldTrace,Sc, List2).

+tr2sc_list([], _, _, []).

scalar_trace_to_scalar(s_unop{X, IE}, OldTrace,Sc, s_unop{X, NewIE}) :-
    scalar_trace_to_scalar(IE, OldTrace, Sc, NewIE).

scalar_trace_to_scalar(s_binop{X, LE, RE}, OldTrace,Sc, s_binop{X, NewLE, NewRE}) :-
    scalar_trace_to_scalar(LE, OldTrace,Sc, NewLE), 
    scalar_trace_to_scalar(RE, OldTrace, Sc, NewRE).





% --------------------------------------------------------------------------------------------
% Replacing expressions within guards
% -----------------------------------------------------------------------------------------
% Given a guard which involves a comparison between a return value and an integer,
% replace the return value expression with a specific numeric value.
predicate replace_exp_guard(in OrigG: g_guard, in NewExp:scalar, out ReplacedG: g_guard).

% T is a trace that originates from the return value of this function.
% Note that the representation of the return trace is determined by the 
% top-down correlation analysis.
predicate return_trace(in T:t_trace).



return_trace(T) :- 
	cil_curfn(F), trace_root(T, temp{"", F}).



% Case 1. Scalar comparison
% first integer expression is the return value expression.
replace_exp_guard(OrigG, NewExp, ReplacedG) :- 
	 #id_g(SG, OrigG), equal(SG, br_cmp{sc{OP, IE1, IE2}}),
	 scalar_trace_used(IE1, T), return_trace(T),
	 #id_g(br_cmp{sc{OP, NewExp, IE2}}, ReplacedG).



% Case 2.: Scalar comparison, not negated and
% second integer expression is the return value expr.
replace_exp_guard(OrigG, NewExp, ReplacedG) :- 
	 #id_g(SG, OrigG), equal(SG, br_cmp{sc{OP, IE1, IE2}}),
	 scalar_trace_used(IE2, T), return_trace(T),
	 #id_g(br_cmp{sc{OP, IE1, NewExp}}, ReplacedG).


% Case 3.a: sc_eqz, not negated
replace_exp_guard(OrigG,  NewExpr, ReplacedG) :- 
	 #id_g(SG, OrigG), equal(SG, br_cmp{sc_eqz{IE}}),
	 scalar_trace_used(IE, T), return_trace(T),
	 #id_g(br_cmp{sc_eqz{NewExpr}}, ReplacedG).

% Case 3: #not
replace_exp_guard(OrigG, NewExpr, ReplacedG) :-
	#not(NotOrigG, OrigG), 
	replace_exp_guard(NotOrigG, NewExpr, NotReplacedG),
	#not(NotReplacedG, ReplacedG).

% Case 4: #and 
replace_exp_guard(OrigG, NewExpr, ReplacedG) :-
	#and(G1, G2, OrigG),
	replace_exp_guard(G1, NewExpr, ReplacedG1),
	replace_exp_guard(G2, NewExpr, ReplacedG2),
	#and(ReplacedG1, ReplacedG2, ReplacedG).

% Case 5: #or 
replace_exp_guard(OrigG, NewExpr,  ReplacedG) :-
	#or(G1, G2, OrigG), 
	replace_exp_guard(G1, NewExpr, ReplacedG1), 
	replace_exp_guard(G2, NewExpr, ReplacedG2), 
	#or(ReplacedG1, ReplacedG2, ReplacedG).



% -----------------------------------------------------------------------------------
% Mapping guards that involve return values of functions
% ------------------------------------------------------------------------------------
% Why do we need this?
% If there is a correlation between the return value of a function and a property 
% of interest, the guard that needs to be translated involves the return value.
% However, the existing inst_guard cannot translate things that involve return values,
% so we must do this seperately.
% -------------------------------------------------------------------------------------

predicate replace_guard_trace(in OrigG: g_guard, in OldTrace: t_trace, in NewTrace: t_trace, out NewG:g_guard).



% Case 1. id_g: Scalar comparison
replace_guard_trace(OrigG, OldTrace, NewTrace, ReplacedG) :- 
	 #id_g(SG, OrigG), equal(SG, br_cmp{sc{OP, IE1, IE2}}),	
	 scalar_trace_rename(IE1, OldTrace, NewTrace, NewExp1), 
	 scalar_trace_rename(IE2, OldTrace, NewTrace, NewExp2),
	#id_g(br_cmp{sc{OP, NewExp1, NewExp2}}, ReplacedG). 

 

% Case 2: id_g: sc_eqz
replace_guard_trace(OrigG, OldTrace, NewTrace, ReplacedG) :- 
	 #id_g(SG, OrigG), equal(SG, br_cmp{sc_eqz{IE}}),
	scalar_trace_rename(IE, OldTrace, NewTrace, NewExpr),
	 #id_g(br_cmp{sc_eqz{NewExpr}}, ReplacedG).

% Case 3: #not
replace_guard_trace(OrigG, OldTrace, NewTrace, ReplacedG) :-
	#not(NotOrigG, OrigG), 
	replace_guard_trace(NotOrigG, OldTrace, NewTrace, NotReplacedG),
	#not(NotReplacedG, ReplacedG).

% Case 4: #and 
replace_guard_trace(OrigG, OldTrace, NewTrace, ReplacedG) :-
	#and(G1, G2, OrigG),
	replace_guard_trace(G1, OldTrace, NewTrace, ReplacedG1),
	replace_guard_trace(G2, OldTrace, NewTrace, ReplacedG2),
	#and(ReplacedG1, ReplacedG2, ReplacedG).

% Case 5: #or 
replace_guard_trace(OrigG, OldTrace, NewTrace, ReplacedG) :-
	#or(G1, G2, OrigG), 
	replace_guard_trace(G1, OldTrace, NewTrace, ReplacedG1), 
	replace_guard_trace(G2, OldTrace, NewTrace, ReplacedG2), 
	#or(ReplacedG1, ReplacedG2, ReplacedG).

% ------------------------------------------------------------------------

% Replace the trace in the given gurad with the provided scalar.
% This predicate is mostly used for replacing guards used in loop
% termination conditions. For instance, it is used to convert i<size to
% 0<size if 0 is the initial value of i.This allows us to determine
% conditions under which the loop never executes.
predicate guard_trace_to_scalar(in OrigG: g_guard, in OldTrace: t_trace, in Scalar: scalar, out NewG:g_guard).


% Case 1. id_g: Scalar comparison
guard_trace_to_scalar(OrigG, OldTrace, Scalar, ReplacedG) :- 
	 #id_g(SG, OrigG), equal(SG, br_cmp{sc{OP, IE1, IE2}}),	
	 scalar_trace_to_scalar(IE1, OldTrace,Scalar, NewExp1), 
	 scalar_trace_to_scalar(IE2, OldTrace,Scalar, NewExp2),
	 #id_g(br_cmp{sc{OP, NewExp1, NewExp2}}, ReplacedG). 

 

% Case 2: id_g: sc_eqz
guard_trace_to_scalar(OrigG, OldTrace, Scalar, ReplacedG) :- 
	 #id_g(SG, OrigG), equal(SG, br_cmp{sc_eqz{IE}}),
	scalar_trace_to_scalar(IE, OldTrace, Scalar, NewExpr),
	 #id_g(br_cmp{sc_eqz{NewExpr}}, ReplacedG).

% Case 3: #not
guard_trace_to_scalar(OrigG, OldTrace, Scalar, ReplacedG) :-
	#not(NotOrigG, OrigG), 
	guard_trace_to_scalar(NotOrigG, OldTrace,Scalar, NotReplacedG),
	#not(NotReplacedG, ReplacedG).

% Case 4: #and 
guard_trace_to_scalar(OrigG, OldTrace,Scalar, ReplacedG) :-
	#and(G1, G2, OrigG),
	guard_trace_to_scalar(G1, OldTrace,Scalar, ReplacedG1),
	guard_trace_to_scalar(G2, OldTrace,Scalar, ReplacedG2),
	#and(ReplacedG1, ReplacedG2, ReplacedG).

% Case 5: #or 
guard_trace_to_scalar(OrigG, OldTrace, Scalar, ReplacedG) :-
	#or(G1, G2, OrigG), 
	guard_trace_to_scalar(G1, OldTrace, Scalar, ReplacedG1), 
	guard_trace_to_scalar(G2, OldTrace, Scalar, ReplacedG2), 
	#or(ReplacedG1, ReplacedG2, ReplacedG).


% -------------------------------------------------------------------------

predicate replace_subguard(in OrigG: g_guard, in OldSubguard:g_guard,
	 in NewSubguard:g_guard, out ReplacedG:g_guard).

% Base case: found
replace_subguard(OrigG, OldSG, NewSG, ReplacedG) :-
	#not(OrigG, NotOrigG),
	equal(OldSG, NotOrigG), #not(NewSG, ReplacedG).

replace_subguard(OrigG, OldSG, NewSG, ReplacedG) :-
	equal(OldSG, OrigG), ReplacedG=NewSG.


% Base case: Not found
replace_subguard(OrigG, OldSG, NewSG, OrigG) :- 
	 #id_g(_, OrigG), ~equal(OrigG, OldSG),
	#not(OrigG, NotOrigG), ~equal(NotOrigG, OldSG).


% Case 2: #not
replace_subguard(OrigG, OldSG, NewSG, ReplacedG) :-
	#not(NotOrigG, OrigG), 
	replace_subguard(NotOrigG, OldSG, NewSG, NotReplacedG),
	#not(NotReplacedG, ReplacedG).

% Case 4: #and 
replace_subguard(OrigG, OldSG, NewSG, ReplacedG) :-
	#and(G1, G2, OrigG),
	replace_subguard(G1, OldSG, NewSG, ReplacedG1),
	replace_subguard(G2, OldSG, NewSG, ReplacedG2),
	#and(ReplacedG1, ReplacedG2, ReplacedG).

% Case 5: #or 
replace_subguard(OrigG, OldSG, NewSG, ReplacedG) :-
	#or(G1, G2, OrigG), 
	replace_subguard(G1, OldSG, NewSG, ReplacedG1), 
	replace_subguard(G2, OldSG, NewSG, ReplacedG2), 
	#or(ReplacedG1, ReplacedG2, ReplacedG).


% ---------------------------------------------------------------------------
% Analyses using the correlation analysis must use corr_inst_guard instead of 
% inst_guard.

predicate corr_inst_guard(in SUM: sum, in P:pp, in G:g_guard, out TG: g_guard).

% Don't do anything special for loops.
corr_inst_guard(s_loop{I}, P, G, TG) :-
	inst_guard(s_loop{I}, P, G, TG).

corr_inst_guard(s_call{I}, P, G, TG) :-
	~callret(I, _),
	inst_guard(s_call{I}, P, G, TG).


corr_inst_guard(s_call{I}, P, G, TG) :-
	callret(I, R),
	dircall(I, F), 
	guard_trace_used(G, drf{root{temp{"", F}}}),
	replace_guard_trace(G, drf{root{temp{"", F}}}, drf{root{R}}, TG).

 
	
corr_inst_guard(s_call{I}, P, G, TG) :-
	callret(I, R),
	dircall(I, F), 
	(~guard_trace_used(G, drf{root{temp{"", F}}});
	~replace_guard_trace(G, drf{root{temp{"", F}}}, drf{root{R}}, _)),
	inst_guard(s_call{I}, P, G, TG).	

corr_inst_guard(s_call{I}, P, G, TG) :-
	~dircall(I, _), 
	inst_guard(s_call{I}, P, G, TG).
