%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%   Peter Hawkins  <hawkinsp@cs.stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% Modelling integer operations using bit vectors

import "memory.clp".

% ANALYSIS REPRESENTATIONS

using solve_bitvector.

% representation for bit vectors
type g_vecrep ::=
    vr_t{t_trace}    % initial value of some trace
  | vr_i{pp,c_exp}.  % imprecise intermediate result

% representation for regular unconstrained bits, which can be the bits used
% by upstream analyses, or the extra type g_axrep for extension by downstream
% analyses
type g_xrep.
type g_arep ::=
    ar_extra{g_xrep}
  | ar_path{pp,pp}
  | ar_entry{pp,t_trace,t_trace}
  | ar_exit{sum,pp,t_trace,t_trace}.

type g_vec = bvec[g_vecrep,g_arep].
type g_bit = bvecbit[g_vecrep,g_arep].

% intermediate integer/floating-point values are just bvecs
type t_nrep = g_vec.

% PREDICATES

% fix the length of all bit vectors at N. if not supplied, the default lengths
% given by the type will be used
predicate fixed_vectors(N:int).

% at program point P, expression E evaluates to bit vector BV
predicate evalbv(in P:pp,in E:c_exp,BV:g_vec).

% as interpreted by the call/loop I at point P, guard RG maps to NG
predicate inst_guard(in I:sum,in P:pp,in RG:g_guard,NG:g_guard).

% As interpreted by the call/loop I at point P, bitvector RV maps to NV
predicate inst_bvec(in I:sum,in P:pp,in RV:g_vec,NV:g_vec).

% convert a trace_val to a bitvector, using the trace's type to determine
% sign and length for unconstrained initial trace values if necessary
predicate trace_vec(in t_trace_val,g_vec).

% convert a specific bit of a trace_val to a guard
predicate trace_bit(in t_trace_val,in int,g_guard).

% CUSTOMIZATION

% add omit_evalbv(P,E) to prevent evalbv(P,E,_) from being computed in the
% standard way, allowing for customization of evalbv(P,E,_)
predicate omit_evalbv(in P:pp,in E:c_exp).

% RULES

% implement satisfiability predicates for upstream analyses
type g_guard_asn = bval_asn[g_bit].
?guard_sat(G), #sat(G), +guard_sat(G).
?guard_sat_asn(G,ASN), #satasn(G,ASN), +guard_sat_asn(G,ASN).
?asn_guard(ASN,G), #asng(ASN,G), +asn_guard(ASN,G).
?asn_value(ASN,V,_), trace_vec(V,BV), #asnbv(ASN,BV,N), +asn_value(ASN,V,N).

% implement unconstrained bits for upstream analyses
?path_bit(P0,P1,_), #id_g(b_abit{ar_path{P0,P1}},G), +path_bit(P0,P1,G).
?entry_bit(P,S,T,_), #id_g(b_abit{ar_entry{P,S,T}},G), +entry_bit(P,S,T,G).
?exit_bit(SUM,P,S,T,_),
    #id_g(b_abit{ar_exit{SUM,P,S,T}},G), +exit_bit(SUM,P,S,T,G).

% integer kind K is modelled with sign S and bit length LEN. this is equal to
% ikind_bytes*8, unless fixed_vectors is supplied for some width
predicate ikind_bvec(K:ikind,S:bool,LEN:int).
fixed_vectors(LEN), ikind_bytes(K,S,_), +ikind_bvec(K,S,LEN).
~fixed_vectors(_), ikind_bits(K,S,Len),
    +ikind_bvec(K,S,Len).

% model pointers as bitvectors with sign S and length LEN. 
predicate ptr_bvec(S:bool,LEN:int).
fixed_vectors(LEN), +ptr_bvec(false,LEN).
~fixed_vectors(_), pointer_bits(Len), +ptr_bvec(false, Len).

% float kind K is modelled with sign S and bit length LEN
predicate fkind_bvec(K:fkind,S:bool,LEN:int).
fixed_vectors(LEN), fkind_bytes(K,_), +fkind_bvec(K,true,LEN).
~fixed_vectors(_), fkind_bits(K,LEN),
    +fkind_bvec(K,true,LEN).

% generate an unconstrained vector for E evaluated at P. The OF bit is true
% if the vector may be the result of an overflowing operation, or false
% otherwise
predicate unconstrained(in P:pp,in E:c_exp,in OF:bool,V:g_vec).
?unconstrained(P,E,OF,_), exp_tint(E,K), ikind_bvec(K,S,LEN),
    #id_bv(vr_i{P,E},S,OF,LEN,V), +unconstrained(P,E,OF,V).

% convert trace values to vecs/bits

% get trace sign/length for int/float traces. assume pointer for other traces
predicate ifkind_bvec(in T:t_trace,S:bool,LEN:int).
?ifkind_bvec(drf{T},_,_), trace_type(s_func,T,y_int{K}),
    ikind_bvec(K,S,LEN), +ifkind_bvec(drf{T},S,LEN).
?ifkind_bvec(drf{T},_,_), trace_type(s_func,T,y_flt{K}),
    fkind_bvec(K,S,LEN), +ifkind_bvec(drf{T},S,LEN).

?trace_vec(trace{T},_), ifkind_bvec(T,S,LEN),
    #id_bv(vr_t{T},S,false,LEN,V), +trace_vec(trace{T},V).
?trace_vec(trace{T},_), ~ifkind_bvec(T,_,_), ptr_bvec(S,LEN),
    #id_bv(vr_t{T},S,false,LEN,V), +trace_vec(trace{T},V).
+trace_vec(nrep{V},V).

?trace_bit(trace{T},N,_), #id_g(b_vbit{vr_t{T},N},G), +trace_bit(trace{T},N,G).
?trace_bit(nrep{V},N,_), #bv_bit(V,N,G), +trace_bit(nrep{V},N,G).

% floating point expressions

% float expressions are "don't care", get an unconstrained value for the result
exp_tflt(E,K), ?eval(P,E,_,_), ~omit_eval(P,E),
    #bool_g(true,G), fkind_bytes(K,BYTES), int_mul(BYTES,8,LEN),
    #id_bv(vr_i{P,E},true,true,LEN,V), +eval(P,E,nrep{V},G).

% convert bitvectors to locations/guards

% for non-lvalue integer expressions (memory.clp handles all lvalue exps),
% evaluate the expression as a bitvector and use that representation
exp_tint(E,_), ~cil_exp_lval(E,_), ?eval(P,E,_,_), ~omit_eval(P,E),
    evalbv(P,E,BV), #bool_g(true,G), +eval(P,E,nrep{BV},G).

% handle scalar to pointer casts
cil_exp_cast(E,CE,_), exp_tptr(E), ~exp_tptr(CE), ?eval(P,E,_,_),
    ~omit_eval(P,E), eval(P,CE,V,G), +eval(P,E,V,G).

% for integer expressions evaluated for a boolean condition, evaluate as
% a bitvector and compute the NEZ condition
exp_tint(E,_), ~exp_boolean(E), ?beval(P,E,_,_), ~omit_beval(P,E),
    evalbv(P,E,BV), #bv_nez(BV,BG), #bool_g(true,G), +beval(P,E,BG,G).

% bitvector expression evaluation

% integer constant expressions, just get the appropriate bitvector
cil_exp_const(E,_), cil_exp_x_intval(E,N),
    ?evalbv(P,E,_), ~omit_evalbv(P,E),
    exp_tint(E,K), ikind_bvec(K,S,LEN), #int_bv(N,S,LEN,BV), +evalbv(P,E,BV).
cil_exp_x_sizeof_alignof(E), cil_exp_x_intval(E,N),
    ?evalbv(P,E,_), ~omit_evalbv(P,E),
    exp_tint(E,K), ikind_bvec(K,S,LEN), #int_bv(N,S,LEN,BV), +evalbv(P,E,BV).

% CIL should be constant folding these &(0->field) but it's not.
% just get an unconstrained value
cil_exp_x_intval(MCE,0), cil_exp_cast(ME,MCE,_), cil_lval_mem(LV,ME,_),
    cil_exp_addr(AE,LV), cil_exp_cast(E,AE,_),
    ?evalbv(P,E,_), ~omit_evalbv(P,E),
    unconstrained(P,E,true,V), +evalbv(P,E,V).

% adding will merge G=>BV into the evaluation of E at P
predicate bvmerge(in P:pp,in E:c_exp,BV:g_vec,G:g_guard).
?evalbv(P,E,_), ~omit_evalbv(P,E), bvmerge(P,E,_,_),
    pp_order(P,DEP,DIST), exp_order(E,EDIST), $order(DEP,DIST,2,EDIST),
    \/bvmerge(P,E,BV,G):#bv_all(BV,G,MBV), +evalbv(P,E,MBV).

% evaluate integer lvals using #id_bv by casting all possible targets of the
% lvalue to vectors and merging them together according to the guards
cil_exp_lval(E,_), exp_tint(E,_), ?bvmerge(P,E,_,_),
    eval(P,E,T,G), trace_vec(T,BV), +bvmerge(P,E,BV,G).

% cast expressions
cil_exp_cast(E,CE,_), exp_tint(E,K), ikind_bvec(K,S,LEN),
    ?evalbv(P,E,_), ~omit_evalbv(P,E),
    evalbv(P,CE,CBV), #bv_cast(CBV,S,LEN,BV), +evalbv(P,E,BV).

% pointer/float conversion to bitvectors

% we need a bit vector for pointer targets for some situations such as
% a ptr->uint cast
exp_tptr(E), ?bvmerge(P,E,_,_),
    eval(P,E,T,G), trace_vec(T,BV), +bvmerge(P,E,BV,G).

% cast floats to integers (yurgh)
exp_tflt(E,_), ?bvmerge(P,E,_,_),
    eval(P,E,T,G), trace_vec(T,BV), +bvmerge(P,E,BV,G).

% convert guards to bit vectors

% only needs to be done for exp_boolean expressions, for which evalbv is not
% defined. do the cast by evalbv'ing to a bit vector with the lower bit set
exp_boolean(E), exp_tint(E,K), ikind_bvec(K,S,LEN), ?bvmerge(P,E,_,_),
    beval(P,E,BG,G), #g_bv(BG,S,LEN,BV), +bvmerge(P,E,BV,G).

% unop expressions

% the result of applying unop OP to RBV is BV
predicate bvunop(in OP:unop,in RBV:g_vec,BV:g_vec).

% model the CIL arithmetic unops
?bvunop(u_neg,RBV,_), #bv_neg(RBV,BV), +bvunop(u_neg,RBV,BV).
?bvunop(u_bnot,RBV,_), #bv_bnot(RBV,BV), +bvunop(u_bnot,RBV,BV).

cil_exp_unop(E,OP,RE,_), ?evalbv(P,E,_), ~omit_evalbv(P,E),
    evalbv(P,RE,RBV), bvunop(OP,RBV,BV), +evalbv(P,E,BV).

% arithmetic binop expressions

% get an unconstrained value for arithmetic operations on pointers (minuspp)
% XXX: Actually it's possible for overflow to occur on pointer-pointer
% subtraction. We'll ignore it for now.
pointer_int_op(OP), cil_exp_binop(E,OP,_,_,_),
    ?evalbv(P,E,_), ~omit_evalbv(P,E),
    unconstrained(P,E,false,V), +evalbv(P,E,V).

% the precise result of applying binop OP to vectors LV and RV is V
predicate bvbinop(in OP:binop,in LV:g_vec,in RV:g_vec,V:g_vec).

% model the CIL arithmetic binops precisely
?bvbinop(b_plusa,LV,RV,_),  #bv_add(LV,RV,V),  +bvbinop(b_plusa,LV,RV,V).
?bvbinop(b_minusa,LV,RV,_), #bv_sub(LV,RV,V),  +bvbinop(b_minusa,LV,RV,V).
?bvbinop(b_band,LV,RV,_),   #bv_band(LV,RV,V), +bvbinop(b_band,LV,RV,V).
?bvbinop(b_bxor,LV,RV,_),   #bv_bxor(LV,RV,V), +bvbinop(b_bxor,LV,RV,V).
?bvbinop(b_bor,LV,RV,_),    #bv_bor(LV,RV,V),  +bvbinop(b_bor,LV,RV,V).

% model Mul/Shiftlt/Shiftrt precisely when the LHS/RHS is constant
?bvbinop(b_mult,LV,RV,_), #bv_int(LV,N),
    #bv_mul(RV,N,V), +bvbinop(b_mult,LV,RV,V).
?bvbinop(b_mult,LV,RV,_), #bv_int(RV,N), ~#bv_int(LV,_),
    #bv_mul(LV,N,V), +bvbinop(b_mult,LV,RV,V).
?bvbinop(b_shiftlt,LV,RV,_), #bv_int(RV,N),
    #bv_shl(LV,N,V), +bvbinop(b_shiftlt,LV,RV,V).
?bvbinop(b_shiftrt,LV,RV,_), #bv_int(RV,N),
    #bv_shr(LV,N,V), +bvbinop(b_shiftrt,LV,RV,V).

% get precise results for bvbinop results, otherwise get an unconstrained value
arith_op(OP), cil_exp_binop(E,OP,LE,RE,_), ?evalbv(P,E,_), ~omit_evalbv(P,E),
    evalbv(P,LE,LV), evalbv(P,RE,RV), bvbinop(OP,LV,RV,V), +evalbv(P,E,V).
arith_op(OP), cil_exp_binop(E,OP,LE,RE,_), ?evalbv(P,E,_), ~omit_evalbv(P,E),
    evalbv(P,LE,LV), evalbv(P,RE,RV), ~bvbinop(OP,LV,RV,_), 
    unconstrained(P,E,true,V), +evalbv(P,E,V).

% comparison binop expressions

% the result of applying binop OP to LBV and RBV is guard G
predicate bvbinop_g(in P:binop,in LBV:g_vec,in RBV:g_vec,G:g_guard).

% model the CIL comparison binops precisely
?bvbinop_g(b_eq,LV,RV,_), #bv_eq(LV,RV,G), +bvbinop_g(b_eq,LV,RV,G).
?bvbinop_g(b_ne,LV,RV,_), #bv_ne(LV,RV,G), +bvbinop_g(b_ne,LV,RV,G).
?bvbinop_g(b_lt,LV,RV,_), #bv_lt(LV,RV,G), +bvbinop_g(b_lt,LV,RV,G).
?bvbinop_g(b_gt,LV,RV,_), #bv_gt(LV,RV,G), +bvbinop_g(b_gt,LV,RV,G).
?bvbinop_g(b_le,LV,RV,_), #bv_le(LV,RV,G), +bvbinop_g(b_le,LV,RV,G).
?bvbinop_g(b_ge,LV,RV,_), #bv_ge(LV,RV,G), +bvbinop_g(b_ge,LV,RV,G).

% compare_op binops should be beval'ed, according to paths.clp
compare_op(OP), cil_exp_binop(E,OP,LE,RE,_), ?beval(P,E,_,_), ~omit_beval(P,E),
    evalbv(P,LE,LBV), evalbv(P,RE,RBV), bvbinop_g(OP,LBV,RBV,BG),
    #bool_g(true,G), +beval(P,E,BG,G).

% condition re-evaluation

?inst_guard(SUM,P,G,_), #bool_g(true, G), +inst_guard(SUM,P,G,G).
?inst_guard(SUM,P,G,_), #bool_g(false,G), +inst_guard(SUM,P,G,G).

% as interpreted by the call/loop SUM at point P, bit N of callee vector CV
% holds if caller guard G holds
predicate inst_gmerge(in SUM:sum,in P:pp,in CV:g_vecrep,in N:int,G:g_guard).

% revaluate bits for the initial value of callee traces. use inst_trace to get
% the corresponding value in the caller, then extract the Nth bit
?inst_gmerge(SUM,P,vr_t{CT},N,_), inst_trace(SUM,P,CT,T,WG), trace_bit(T,N,NG),
    #and(NG,WG,G), +inst_gmerge(SUM,P,vr_t{CT},N,G).

% just give a warning for re-evaluating the results of imprecise operations
?inst_gmerge(SUM,P,vr_i{PP,EXP},_,_),
    +warning("Can't get inst_guard for imprecise bit",SUM,P,PP,EXP).

% reval complex conditions by switching on what they could be: #and, #or, #not,
% or a #id_g {T,N}
?inst_guard(SUM,P,RG,_), #and_lv(RG0,RG1,LEVEL,RG),
    inst_guard(SUM,P,RG0,NG0), inst_guard(SUM,P,RG1,NG1),
    #and_lv(NG0,NG1,LEVEL,NG), +inst_guard(SUM,P,RG,NG).
?inst_guard(SUM,P,RG,_), #or_lv(RG0,RG1,LEVEL,RG),
    inst_guard(SUM,P,RG0,NG0), inst_guard(SUM,P,RG1,NG1),
    #or_lv(NG0,NG1,LEVEL,NG), +inst_guard(SUM,P,RG,NG).
?inst_guard(SUM,P,RG,_), #not_lv(RG0,LEVEL,RG),
    inst_guard(SUM,P,RG0,NG0),
    #not_lv(NG0,LEVEL,NG), +inst_guard(SUM,P,RG,NG).
?inst_guard(SUM,P,RG,_), #id_g(b_vbit{T,N},RG),
    pp_order(P,DEP,DIST), $order(DEP,DIST,3,0),
    \/inst_gmerge(SUM,P,T,N,G):#or_all(G,MG), +inst_guard(SUM,P,RG,MG).

?inst_guard(SUM,P,RG,_), #id_g(b_abit{BIT},RG),
    +warning("Can't get inst_guard for unknown extra bit",SUM,P,BIT).

% -----------------------------------------------------------------------------

% map a list of guards using inst_guard
predicate inst_bits(in sum,in pp,in list[g_guard],list[g_guard]).
+inst_bits(_,_,[],[]).
?inst_bits(SUM,P,[B0|BS0],_),
    inst_guard(SUM,P,B0,B), inst_bits(SUM,P,BS0,BS),
    +inst_bits(SUM,P,[B0|BS0],[B|BS]).

% map a bit vector out of a loop/call by splitting it up into its component
% bits which are mapped using inst_guard and then recombining it.
?inst_bvec(SUM,P,RV,_), #bv_split(RV,SIGN,OFLOW0,BITS0),
    inst_guard(SUM,P,OFLOW0,OFLOW), inst_bits(SUM,P,BITS0,BITS),
    #bv_split(NV,SIGN,OFLOW,BITS), +inst_bvec(SUM,P,RV,NV).
