%%
% Authors:
%   Brian Hackett  <bhackett@stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

% Path-sensitive intraprocedural points-to and value-aliasing information.
% If the alias analysis has not already been run, assumes non-aliasing at
% entry and side-effect free callees

% this file leaves the type t_nrep of intermediate int/float values abstract.
% files importing this which need to manipulate such values must define t_nrep
type t_nrep.

import "traces.clp".
import "paths.clp".

% PREDICATES

% pointers are normally assumed to be non-aliased at entry, that the targets of
% two different traces are themselves two different traces. any extra points-to
% information in the sum_entry, sum_exit, and other summaries from traces.clp
% are used to relax this assumption

% possible values a trace can hold
type t_trace_val ::=
    trace{t_trace}  % address of another trace (points-to relationship)
  | nrep{t_nrep}.   % intermediate integer/float result

% integer values are ignored here. to model integers, an additional file such
% as intignore.clp or intbvec.clp must be imported, which will fabricate
% additional abstract traces for integer values

% note that all per-pp guards, i.e. val, eval, lval, inst_trace, etc.
% are dependent on P actually being reachable, guard(P,G). to get the conds
% under which both P is reachable *AND* a property val(P,S,T,_) holds, conjoin
% guard(P,GG) and val(P,S,T,G).

% at program point P, the memory location with canonical name S contains value 
% T when G holds (if S is a pointer, then S points to T). Note that S must be 
% the canonical name for the location -- this predicate does NOT work for
% arbitrary access paths.
predicate val(in P:pp,in S:t_trace,T:t_trace_val,G:g_guard).

% at program point P, expression E points to T if G holds.
% only defined for memory operations in this file
predicate eval(in P:pp,in E:c_exp,T:t_trace_val,G:g_guard).

% at program point P, lval LV refers to trace T if G holds
predicate lval(in P:pp,in LV:c_lval,T:t_trace,G:g_guard).

% as interpreted by the call/loop SUM at point P, trace CT maps to T if G holds
predicate inst_trace(in SUM:sum,in P:pp,in CT:t_trace,T:t_trace_val,G:g_guard).

% top-down inst_trace; caller trace T is mapped to by trace CT if G holds
predicate td_inst_trace(in SUM:sum,in P:pp,in T:t_trace,CT:t_trace,G:g_guard).

% val() summary transfer function. S has value T if G holds after transfer
% over the call/loop at P
predicate inst_transfer(in P:pp,in S:t_trace,T:t_trace_val,G:g_guard).

% T is directly accessed at program point P as a read or write (not deepwrite).
% Accesses through callees are not included here.
predicate access(P:pp,T:t_trace,access_type).

% get AT and AI such that index{AT,EY,AI} refers to the same location as T
predicate trace_coerce_index(in SUM:sum,in T:t_trace,in EY:t_type,
                             AT:t_trace,AI:t_trace_val).

% INTEGER EXTENSIONS

% these must be supplied by any integer or other analysis incorporating
% this file

% get an unconstrained bit to distinguish the condition under which
% S points to T at entry point P or as side effect to SUM.
% must be provided once the type of g_bit is actually known
predicate entry_bit(in P:pp,in S:t_trace,in T:t_trace,BG:g_guard).
predicate exit_bit(in SUM:sum,in P:pp,in CS:t_trace,in CT:t_trace,BG:g_guard).

% get the particular scalar value of a trace_val under an assignment
% generated by guard_sat_asn
predicate asn_value(in ASN:g_guard_asn,in NR:t_trace_val,N:int).

% operators implemented downstream needed to compute trace indexes
predicate trace_val_int(in K:ikind,in N:int,NV:t_trace_val).
predicate trace_val_add(in LV:t_trace_val,in RV:t_trace_val,V:t_trace_val).
predicate trace_val_mul(in LV:t_trace_val,in N:int,V:t_trace_val).
predicate trace_val_div(in LV:t_trace_val,in N:int,V:t_trace_val).

% CS is updated to point into array CT as a side effect to SUM at P.
% V is the (unconstrained) index to use in updating the call site state
predicate exit_index(in SUM:sum,in P:pp,in CS:t_trace,in CT:t_trace,
                     V:t_trace_val).

% as interpreted by call/loop SUM at point P, CV maps to V if G holds
predicate inst_trace_val(in SUM:sum,in P:pp,in CV:t_trace_val,
                         V:t_trace_val,G:g_guard).

% SUMMARY TRANSFER EXTENSIONS

% G is a necessary condition for the target of S to be preserved across the
% call/loop SUM at P
predicate inst_transfer_necessary(in SUM:sum,in P:pp,in S:t_trace,G:g_guard).

% the call/loop SUM at P may write S with unconstrained index I into trace T.
% if not specified then trace_val_int(0) will be used
predicate inst_write_array(in P:pp,in S:t_trace,in T:t_trace,I:t_trace_val).

% the call/loop SUM at P may write S with unconstrained integer V.
% if not specified then the value of S before the call will be used
predicate inst_write_int(in P:pp,in S:t_trace,V:t_trace_val).

% CUSTOMIZATION

% val() will not be computed anywhere - disables memory analysis
predicate disable_val().

% entry aliases are not used to relax the entry points-to graph
predicate no_entry_aliasing().

% ignore side effects on SUM at P
predicate no_exit_aliasing(SUM:sum,P:pp).

% each guard G in val(P,S,T,G) is 'true', i.e. no path sensitivity
predicate no_val_guard().

% do not attempt to apply integer or index side effects at summaries
predicate no_inst_may_write().

% prevent eval(P,E,_,_) from being computed in the standard way,
% allowing for customization of eval(P,E,_,_)
predicate omit_eval(in P:pp,in E:c_exp).

% S may be written 
predicate call_effect(in SUM:sum,in P:pp,in S:t_trace).

% only generate inst_trace for arguments included in the function prototype
predicate matching_arg_inst_trace().

% do not generate rel_inst_trace
predicate no_rel_inst_trace().

% RULES

% CIL location offsets

% CIL offset OFF of root trace RT as calculated at P is T if G holds
predicate trace_offset(in P:pp,in RT:t_trace,in OFF:c_offset,
                       T:t_trace,G:g_guard).

% empty offsets preserve the original location
#bool_g(true,G), cil_off_none(OFF), +trace_offset(_,RT,OFF,RT,G).

% field offsets take the specified field of the original location
cil_off_field(OFF,FX,NOFF), cil_field_name(FX,F,_,_,_),
    base_off_type(OFF,TYP), cil_type_comp(TYP,C),
    ?trace_offset(P,RT,OFF,_,_),
    sum_body(SUM,P), trace_simplify(SUM,fld{RT,F,C},FT),
    trace_offset(P,FT,NOFF,T,G), +trace_offset(P,RT,OFF,T,G).

% index offsets are handled only if generate_index_traces() is being used,
% otherwise ignored and preserving the original location
~generate_index_traces(), cil_off_index(OFF,_,NOFF),
    ?trace_offset(P,RT,OFF,_,_), trace_offset(P,RT,NOFF,T,G),
    +trace_offset(P,RT,OFF,T,G).
generate_index_traces(), cil_off_index(OFF,E,NOFF),
    ?trace_offset(P,RT,OFF,_,_), sum_body(SUM,P), eval(P,E,I,IG),
    base_off_type(NOFF,ETYP), convert_type(ETYP,EY),
    trace_simplify(SUM,index{RT,EY,I},IT),
    trace_offset(P,IT,NOFF,T,TG), #and(IG,TG,G), +trace_offset(P,RT,OFF,T,G).

% reverse RT to T such that trace_offset(T,OFF,RT).
predicate trace_rev_offset(in RT:t_trace,in OFF:c_offset,T:t_trace).
cil_off_none(OFF), +trace_rev_offset(RT,OFF,RT).
cil_off_field(OFF,FX,NOFF), cil_field_name(FX,F,_,_,_),
    base_off_type(OFF,TYP), cil_type_comp(TYP,C),
    ?trace_rev_offset(RT,OFF,_), trace_rev_offset(RT,NOFF,MT),
    trace_simplify(s_func,rfld{MT,F,C},T), +trace_rev_offset(RT,OFF,T).
~generate_index_traces(), cil_off_index(OFF,_,NOFF),
    ?trace_rev_offset(RT,OFF,_),
    trace_rev_offset(RT,NOFF,T), +trace_rev_offset(RT,OFF,T).
generate_index_traces(), cil_off_index(OFF,_,_),
    ?trace_rev_offset(_,OFF,_),
    +warning("Can't compute trace_rev_offset for index trace").

% get a fixed number of bytes represented by an offset
predicate offset_bytes(in OFF:c_offset,N:int).
cil_off_none(OFF), +offset_bytes(OFF,0).
cil_off_field(OFF,FX,NOFF), cil_field_name(FX,_,_,BBS,_), int_div(BBS,8,BS),
    ?offset_bytes(OFF,_), offset_bytes(NOFF,NN), int_add(BS,NN,N),
    +offset_bytes(OFF,N).
cil_off_index(OFF,_,_), ?offset_bytes(OFF,_),
    +warning("Can't compute offset_bytes for index trace").

% pointer casts

% casting T to type Y yields CT. not be defined for all T
predicate trace_val_cast(in SUM:sum,in T:t_trace_val,in Y:t_type,
                         CT:t_trace_val).

predicate fld_chain(in SUM:sum,in T:t_trace,in FLIST:list[t_pair[string,string]],NT:t_trace).
+fld_chain(_,T,[],T).
fld_chain(SUM,T,[pair{F,C}|FLIST],NT) :-
    trace_simplify(SUM,fld{T,F,C},MT), fld_chain(SUM,MT,FLIST,NT).

predicate rfld_chain(in SUM:sum,in T:t_trace,in FLIST:list[t_pair[string,string]],NT:t_trace).
+rfld_chain(_,T,[],T).
rfld_chain(SUM,T,[pair{F,C}|FLIST],NT) :-
    rfld_chain(SUM,T,FLIST,MT), trace_simplify(SUM,rfld{MT,F,C},NT).

% look for field upcasts, where a pointer to some comp C is cast to the type of
% one of its first fields, or downcasts, where a pointer to some type Y is
% cast to a type which contains Y as its first field
predicate field_cast(in SUM:sum,in T:t_trace,in Y:t_type,NT:t_trace).
field_cast(SUM,T,Y,NT) :- trace_type(SUM,T,y_comp{C}),
    first_field(C,Y,FLIST), fld_chain(SUM,T,FLIST,NT).
field_cast(SUM,T,y_comp{C},NT) :- trace_type(SUM,T,Y),
    first_field(C,Y,FLIST), rfld_chain(SUM,T,FLIST,NT).

% apply field_cast where possible, otherwise preserve the trace
?trace_val_cast(SUM,trace{T},Y,_), field_cast(SUM,T,Y,NT),
    +trace_val_cast(SUM,trace{T},Y,trace{NT}).
?trace_val_cast(SUM,trace{T},Y,_), ~field_cast(SUM,T,Y,_),
    +trace_val_cast(SUM,trace{T},Y,trace{T}).
+trace_val_cast(_,nrep{V},_,nrep{V}).

% lvalue evaluation

% compute the trace for variable lvals. this is an offset from the var_trace
% for the variable
cil_lval_var(LV,X,OFF), var_trace(X,RT),
    ?lval(P,LV,_,_), trace_offset(P,root{RT},OFF,T,G), +lval(P,LV,T,G).

% compute the trace for memory lvals. evaluate the pointer for its targets,
% then take the offset of each target
cil_lval_mem(LV,ME,OFF),
    ?lval(P,LV,_,_), eval(P,ME,trace{RT},MG), trace_offset(P,RT,OFF,T,OG),
    #and(MG,OG,G), +lval(P,LV,T,G).

% constructing indexes

% cast an index BI of elem width BBS to index NI of elem width NBS,
% scaling scaling BI as appropriate. we currently only support clean
% multiplication/division of sizes
predicate index_cast(in BI:t_trace_val,in BBS:int,in NBS:int,NI:t_trace_val).
+index_cast(I,BS,BS,I).
?index_cast(BI,BBS,NBS,_), int_gt(BBS,NBS),
    int_div(BBS,NBS,DN), int_mul(DN,NBS,BBS),
    trace_val_mul(BI,DN,NI), +index_cast(BI,BBS,NBS,NI).
?index_cast(BI,BBS,NBS,_), int_lt(BBS,NBS),
    int_div(NBS,BBS,DN), int_mul(DN,BBS,NBS),
    trace_val_div(BI,DN,NI), +index_cast(BI,BBS,NBS,NI).

% if we are already at an offset into an array, scale the offset according
% to the old vs. new type. otherwise just get index{T,EY,0}
?trace_coerce_index(SUM,index{AT,XY,XI},EY,_,_),
    type_width(XY,XBS), type_width(EY,EBS), index_cast(XI,XBS,EBS,AI),
    +trace_coerce_index(SUM,index{AT,XY,XI},EY,AT,AI).
?trace_coerce_index(SUM,T,EY,_,_), T\=index{_,_,_},
    trace_val_int(iint,0,ZI), +trace_coerce_index(SUM,T,EY,T,ZI).

% expression evaluation

% constant expressions (just strings, these have pointer type)
cil_const_str(C,STR), cil_exp_const(E,C), ?eval(P,E,_,_), ~omit_eval(P,E),
    #bool_g(true,G), +eval(P,E,trace{root{cstr{STR}}},G).

% evaluate lval expressions by getting val() for the referenced trace S.
% note that all variables are evaluated here, not just pointers.
% an 'abstract target' will then be generated for integer variables which
% are evaluated, i.e. int trace T will initially point to drf{T}
cil_exp_lval(E,LV), ?eval(P,E,_,_), ~omit_eval(P,E),
    lval(P,LV,S,SG), +access(P,S,read),
    val(P,S,T,TG), #and(SG,TG,G), +eval(P,E,T,G).

% address-of lvalue expressions
cil_exp_addr(E,LV), ?eval(P,E,_,_), ~omit_eval(P,E),
    lval(P,LV,T,G), +eval(P,E,trace{T},G).

% expression E is the sum of pointer LE and integer RE (scaled by sizeof(*LE)).
% fabricate a negation for b_minuspi
predicate exp_psum(E:c_exp,LE:c_exp,RE:c_exp).
cil_exp_binop(E,b_pluspi,LE,RE,_), +exp_psum(E,LE,RE).
cil_exp_binop(E,b_minuspi,LE,RE,_), cil_make_exp(RE,"neg",NRE),
    exp_type(RE,RTYP), +cil_exp_unop(NRE,u_neg,RE,RTYP), +exp_psum(E,LE,NRE).

% pointer arithmetic is handled only if generate_index_traces() is being used,
% otherwise ignored and preserving the targets of the LHS
~generate_index_traces(), exp_psum(E,LE,_), ?eval(P,E,_,_), ~omit_eval(P,E),
    eval(P,LE,T,G), +eval(P,E,T,G).
generate_index_traces(), exp_psum(E,LE,RE), ?eval(P,E,_,_), ~omit_eval(P,E),
    sum_body(SUM,P), exp_type(E,TYP), convert_type(TYP,y_ptr{DY}),
    eval(P,LE,trace{T},LG), trace_coerce_index(SUM,T,DY,BT,LI),
    eval(P,RE,RI,RG), trace_val_add(LI,RI,I),
    #and(LG,RG,G), +eval(P,E,trace{index{BT,DY,I}},G).

% add the cast for pointer to pointer cast expressions except when
% the target is not a trace. scalar to pointer casts will be handled by
% the scalar representation
cil_exp_cast(E,CE,TYP), exp_tptr(E), exp_tptr(CE),
    convert_type(TYP,y_ptr{DY}), ?eval(P,E,_,_), ~omit_eval(P,E),
    eval(P,CE,T,G), sum_body(SUM,P), trace_val_cast(SUM,T,DY,CT),
    +eval(P,E,CT,G).

% lift pointer expressions to boolean guards by fabricating an integer cast
% and evaluating that (passing the buck to any integer extensions).

% E is a pointer expression used as a boolean, with CE the cast result
predicate exp_tptr_as_boolean(E:c_exp).
cil_exp_unop(_,u_lnot,E,_), exp_tptr(E), +exp_tptr_as_boolean(E).
logic_op(OP), cil_exp_binop(_,OP,E,_,_), exp_tptr(E), +exp_tptr_as_boolean(E).
logic_op(OP), cil_exp_binop(_,OP,_,E,_), exp_tptr(E), +exp_tptr_as_boolean(E).
cil_stmt_if(_,E,_,_), exp_tptr(E), +exp_tptr_as_boolean(E).

exp_tptr_as_boolean(E),
    cil_make_exp(E,"intcast",CE), cil_make_type(E,"inttype",TYP),
    +cil_exp_cast(CE,E,TYP), +cil_type_int(TYP,iuint),
    ?beval(P,E,_,_), ~omit_beval(P,E), beval(P,CE,BG,G), +beval(P,E,BG,G).

% function/loop entry aliasing

% at entry to SUM, S may additionally point to T
predicate entry_points(in SUM:sum,in S:t_trace,T:t_trace).

% summary specific entry aliasing information
predicate inst_entry_points(SUM:sum,S:t_trace,T:t_trace).
cil_curfn(FN), sum_bound(SUM,_,_), ~omit_entry_alias_summary(SUM),
    sum_entry(FN,SUM)->spoints(S,T), +inst_entry_points(SUM,S,T).

% add entry_points for each inst_entry_points. watch for additional S->drf{S}
% edges which might be generated by the alias analysis
inst_entry_points(SUM,S,T), ~trace_simplify(SUM,drf{S},T),
    ~omit_entry_points(SUM,S,T), +entry_points(SUM,S,T).

% get entry points-to information from global and composite type summaries
?entry_points(SUM,S,_), ~omit_entry_alias_summary(SUM),
    trace_root(S,glob{G}), sum_glob(G)->spoints(S,T),
    ~trace_simplify(s_func,drf{S},T),
    ~omit_entry_points(SUM,S,T), +entry_points(SUM,S,T).
?entry_points(SUM,S,_), ~omit_entry_alias_summary(SUM),
    trace_sub(S,fld{STR,_,C},_), comp_to_relative(STR,S,RS),
    sum_comp(C)->spoints(RS,RT), comp_from_relative(STR,RT,T),
    ~omit_entry_points(SUM,S,T), +entry_points(SUM,S,T).

% instruction trace evaluation

% get call arguments which match the actual function type, if desired
predicate use_call_arg(in I:c_instr,in A:int).
matching_arg_inst_trace(), cil_instr_call(I,FNE),
    exp_type(FNE,FNTYP), cil_type_func_arg(FNTYP,A,_,_), +use_call_arg(I,A).
~matching_arg_inst_trace(), +use_call_arg(_,_).

% map arguments out of a call or asm
?inst_trace(s_call{I},P,root{arg{A}},_,_), use_call_arg(I,A),
    cil_instr_call_arg(I,A,E), cil_exp_lval(E,LV), lval(P,LV,T,G),
    +inst_trace(s_call{I},P,root{arg{A}},trace{T},G).
?inst_trace(s_asm{I},P,root{asm_in{A}},_,_),
    cil_instr_asm_in(I,A,_,_,E), cil_exp_lval(E,LV), lval(P,LV,T,G),
    +inst_trace(s_asm{I},P,root{asm_in{A}},trace{T},G).
?inst_trace(s_asm{I},P,root{asm_out{A}},_,_),
    cil_instr_asm_out(I,A,_,_,LV), lval(P,LV,T,G),
    +inst_trace(s_asm{I},P,root{asm_out{A}},trace{T},G).

% for arguments that are NOT l-values, we can only map out drf{arg{_}}
% through use of eval
?inst_trace(s_call{I},P,drf{root{arg{A}}},_,_), use_call_arg(I,A),
    cil_instr_call_arg(I,A,E), ~cil_exp_lval(E,_), eval(P,E,T,G),
    +inst_trace(s_call{I},P,drf{root{arg{A}}},T,G).
?inst_trace(s_asm{I},P,drf{root{asm_in{A}}},_,_),
    cil_instr_asm_in(I,A,_,_,E), ~cil_exp_lval(E,_), eval(P,E,T,G),
    +inst_trace(s_asm{I},P,drf{root{asm_in{A}}},T,G).

% map the return value out of a call
?inst_trace(s_call{I},P,root{return},_,_), callret(I,R), #bool_g(true,G),
    +inst_trace(s_call{I},P,root{return},trace{root{R}},G).

% map arguments and locals out of a loop. just preserve the trace
?inst_trace(s_loop{PH},P,root{arg{A}},_,_), #bool_g(true,G),
    +inst_trace(s_loop{PH},P,root{arg{A}},trace{root{arg{A}}},G).
?inst_trace(s_loop{PH},P,root{local{L}},_,_), #bool_g(true,G),
    +inst_trace(s_loop{PH},P,root{local{L}},trace{root{local{L}}},G).

% map globals out of a call/loop. just preserve the trace
?inst_trace(I,P,root{glob{G}},_,_), #bool_g(true,GG),
    +inst_trace(I,P,root{glob{G}},trace{root{glob{G}}},GG).

% map constant strings out of a call/loop by preserving the trace.
?inst_trace(I,P,root{cstr{S}},_,_), #bool_g(true,GG),
    +inst_trace(I,P,root{cstr{S}},trace{root{cstr{S}}},GG).

% map drf{T} out of a call/loop. since drf{T} means the target at entry to the
% call/loop, evaluate T for its location and take its target at P.
% special case drf{root{return}} due to the implicit CIL call casting
?inst_trace(I,P,drf{CPT},_,_), CPT\=root{return},
    inst_trace(I,P,CPT,trace{PT},PG),
    val(P,PT,T,TG), #and(PG,TG,G), +inst_trace(I,P,drf{CPT},T,G).
?inst_trace(I,P,drf{root{return}},_,_), sum_body(SUM,P),
    inst_trace(I,P,root{return},trace{PT},G),
    trace_type(I,drf{root{return}},Y),
    trace_simplify(SUM,drf{PT},DPT), trace_val_cast(SUM,trace{DPT},Y,CDPT),
    +inst_trace(I,P,drf{root{return}},CDPT,G).

% map fields out of a call/loop, just map out the parent trace
?inst_trace(I,P,fld{CT,F,C},_,_),
    inst_trace(I,P,CT,trace{FT},G), sum_body(SUM,P),
    trace_simplify(SUM,fld{FT,F,C},T), +inst_trace(I,P,fld{CT,F,C},trace{T},G).
?inst_trace(I,P,rfld{CT,F,C},_,_),
    inst_trace(I,P,CT,trace{FT},G), sum_body(SUM,P),
    trace_simplify(SUM,rfld{FT,F,C},T), +inst_trace(I,P,rfld{CT,F,C},trace{T},G).

% map out an index using inst_nrep
?inst_trace(I,P,index{CT,DY,CVI},_,_), sum_body(SUM,P),
    inst_trace(I,P,CT,trace{T},TG), inst_trace_val(I,P,CVI,VI,VG),
    trace_coerce_index(SUM,T,DY,AT,AI), trace_val_add(VI,AI,NI),
    #and(TG,VG,G), +inst_trace(I,P,index{CT,DY,CVI},trace{index{AT,DY,NI}},G).

% top-down instruction trace propagation

% to get the set of callee traces which *could* map to a given caller trace,
% we have to know which traces might not be in their usual invariant state.
% this includes all traces that have been previously written within the summary
% as well as any with extra entry aliasing. generate all inst_trace relevant
% to these accessed traces and store via rel_inst_trace

% T is a relevant trace accessed either at P or at an earlier point
% in the containing SUM
predicate rel_trace(P:pp,T:t_trace).

% rel_trace for traces written directly or through an inner sum
access(AP,T,write), trace_sub(drf{T},DT,_),
    pp_reach(AP,P), AP\=P, +rel_trace(P,DT).

% some interesting other operations
cil_exp_addr(E,LV),
    ~(cil_lval_mem(LV,ME,_), cil_exp_cast(ME,MCE,_), cil_exp_x_intval(MCE,_)),
    lval_type(LV,TYP), ~cil_type_func(TYP,_,_),
    exp_point(E,AP), lval(AP,LV,T,_), trace_sub(drf{T},DT,_),
    pp_reach(AP,P), +rel_trace(P,DT).

% rel_trace for entry points-to relations that apply over the entire SUM
inst_entry_points(SUM,S,T), ~trace_simplify(SUM,drf{S},T),
    sum_body(SUM,P), trace_sub(drf{S},DS,_), +rel_trace(P,DS).

% now, for each call/loop site, crawl down through any relevant roots, staying
% within the rel_trace for points previous to that call/loop
predicate rel_get_trace(I:sum,P:pp,CT:t_trace).
predicate rel_inst_trace(I:sum,P:pp,CT:t_trace,T:t_trace,G:g_guard).
~no_rel_inst_trace(), rel_get_trace(I,P,CT), inst_trace(I,P,CT,trace{T},G),
    +rel_inst_trace(I,P,CT,T,G).

% get the traces for any local/global/args accessed by loops/calls, and for
% all arguments on calls. for calls we need both __arg and *__arg to account
% for both non-lvalue arguments and for structs passed on the stack
~no_rel_inst_trace(), isum(P,_,I),
    rel_trace(P,root{R}), (I\=s_call{_}; R\=arg{_}),
    inst_may_access(I,root{R},read), +rel_get_trace(I,P,root{R}).
~no_rel_inst_trace(), icall(P,_,I), cil_instr_call_arg(I,AN,_), 
    +rel_get_trace(s_call{I},P,root{arg{AN}}),
    +rel_get_trace(s_call{I},P,drf{root{arg{AN}}}).

% arguments to assembly statements. input arguments may be passed either by
% value or by address, so just get the trace regardless of type.
% output arguments are always by address
~no_rel_inst_trace(), iasm(P,_,I), cil_instr_asm_in(I,AN,_,_,_),
    +rel_get_trace(s_asm{I},P,root{asm_in{AN}}),
    +rel_get_trace(s_asm{I},P,drf{root{asm_in{AN}}}).
~no_rel_inst_trace(), iasm(P,_,I), cil_instr_asm_out(I,AN,_,_,_),
    +rel_get_trace(s_asm{I},P,root{asm_out{AN}}).

% walk down dereferences and fields according to rel_trace
rel_inst_trace(I,P,CT,T,_), rel_trace(P,drf{T}),
    trace_simplify(I,drf{CT},DCT), +rel_get_trace(I,P,DCT).
rel_inst_trace(I,P,CT,T,_), rel_trace(P,fld{T,F,C}),
    trace_simplify(I,fld{CT,F,C},FCT), +rel_get_trace(I,P,FCT).
rel_inst_trace(I,P,CT,T,_), rel_trace(P,rfld{T,F,C}),
    trace_simplify(I,rfld{CT,F,C},FCT), +rel_get_trace(I,P,FCT).
rel_inst_trace(I,P,CT,T,_), rel_trace(P,index{T,DY,TI}),
    trace_simplify(I,index{CT,DY,TI},ICT), +rel_get_trace(I,P,ICT).

% finally, we can compute td_inst_trace by looking for any subtraces in
% rel_inst_trace, which will reflect all trace modifications performed
% by the caller. look only for the largest subtrace in rel_inst_trace
% to avoid overapproximating the guard
rel_inst_trace(I,P,CT,T,G), +td_inst_trace(I,P,T,CT,G).
?td_inst_trace(I,P,T,_,_), trace_sub(T,ST,RT), rel_inst_trace(I,P,CST,ST,G),
    (trace_sub(T,drf{ST},_), NT=drf{ST};
     trace_sub(T,fld{ST,F,C},_), NT=fld{ST,F,C};
     trace_sub(T,rfld{ST,F,C},_), NT=rfld{ST,F,C};
     trace_sub(T,index{ST,DY,TI},_), NT=index{ST,DY,TI}),
    pp_order(P,DEP,DIST), $order(DEP,DIST,0,0), ~rel_trace(P,NT),
    trace_compose(I,CST,RT,CT), +td_inst_trace(I,P,T,CT,G).

% also look to see if the trace maps back to itself via inst_trace.
% this covers cases where the trace was left in its invariant state
% by the caller
?td_inst_trace(I,P,T,_,_), (~trace_root(T,arg{_}); I=s_loop{_}),
    inst_trace(I,P,T,trace{T},G), +td_inst_trace(I,P,T,T,G).

% val() merging

% adding will merge G into the condition under which S points to T at P
predicate vmerge(in P:pp,in S:t_trace,T:t_trace_val,G:g_guard).
~disable_val(), ~no_val_guard(), ?val(P,S,_,_), vmerge(P,S,T,_),
    pp_order(P,DEP,DIST), $order(DEP,DIST,0,0),
    \/vmerge(P,S,T,G):#or_all(G,MG), ~#bool_g(false,MG), +val(P,S,T,MG).
~disable_val(), no_val_guard(), #bool_g(true,TG),
    ?val(P,S,_,_), vmerge(P,S,T,G), ~#bool_g(false,G), +val(P,S,T,TG).

% initial value information

% absent any entry aliasing info, trace S has value T at entry
% for most values S this is just a (simplified) *S. if we are generating
% indexes however, then if drf{S} is an extra array (S may be used in pointer
% arithmetic later), then S initially points to S[0]
predicate base_entry_target(in SUM:sum,in S:t_trace,T:t_trace).
~generate_index_traces(), ?base_entry_target(SUM,S,_),
    trace_simplify(SUM,drf{S},T), +base_entry_target(SUM,S,T).
generate_index_traces(), ?base_entry_target(SUM,S,_),
    trace_simplify(SUM,drf{S},BT), trace_val_int(iint,0,ZV),
    (trace_soft(SUM,BT,ptarray), trace_type(SUM,BT,EY), T=index{BT,EY,ZV};
     ~trace_soft(SUM,BT,ptarray), T=BT),
    +base_entry_target(SUM,S,T).

% use entry_bit() to compute the initial targets and (disjoint) guards for S.
% TGTS includes all traces besides drf{S} that S might point to, and DTGTS
% includes drf{S} as well (with all leftover conditions as the guard)
predicate split_entry_guard(in P:pp,in S:t_trace,
                            in TGTS:list[t_trace],in PG:g_guard,
                            DTGTS:list[t_pair[t_trace,g_guard]]).
?split_entry_guard(P,S,[],PG,_), sum_body(SUM,P), base_entry_target(SUM,S,T),
    +split_entry_guard(P,S,[],PG,[pair{T,PG}]).
?split_entry_guard(P,S,[T|TAIL],PG,_), entry_bit(P,S,T,BG),
    #and(BG,PG,CPG), #not(BG,NBG), #and(NBG,PG,NPG),
    split_entry_guard(P,S,TAIL,NPG,DTAIL),
    +split_entry_guard(P,S,[T|TAIL],PG,[pair{T,CPG}|DTAIL]).

% resolve val() at summary entry points by using entry_points, getting disjoint
% guards unless no_val_guard() is set
~no_entry_aliasing(), ~no_val_guard(), sum_bound(SUM,P,_), ?vmerge(P,S,_,_),
    \/entry_points(SUM,S,XT):list_all(XT,TGTS),
    #bool_g(true,TG), split_entry_guard(P,S,TGTS,TG,DTGTS),
    list_mem(DTGTS,pair{T,G}), +vmerge(P,S,trace{T},G).
~no_entry_aliasing(), no_val_guard(), sum_bound(SUM,P,_), ?vmerge(P,S,_,_),
    #bool_g(true,G), base_entry_target(SUM,S,T), +vmerge(P,S,trace{T},G),
    entry_points(SUM,S,XT), +vmerge(P,S,trace{XT},G).
no_entry_aliasing(), sum_bound(SUM,P,_), ?vmerge(P,S,_,_),
    #bool_g(true,G), base_entry_target(SUM,S,T), +vmerge(P,S,trace{T},G).

% transfer functions

% assigning to S implicitly casts T to CT. CIL introduces explicit casts for
% everything but function returns *grumble*
predicate sassign_cast(in SUM:sum,in S:t_trace,in T:t_trace_val,
                       CT:t_trace_val).
?sassign_cast(SUM,S,T,_), trace_type(SUM,S,y_ptr{Y}),
    trace_val_cast(SUM,T,Y,CT), +sassign_cast(SUM,S,T,CT).
?sassign_cast(SUM,S,T,_), ~trace_type(SUM,S,y_ptr{_}),
    +sassign_cast(SUM,S,T,T).

% the assignment at P of T to S when BG holds also assigns TF to SF of S
% when G holds. this covers both the direct assignment T -> S as well as any
% indirect structure field assignments resulting from T -> S
predicate sassign(in P:pp,in S:t_trace,in T:t_trace_val,in BG:g_guard,
                  in SF:t_trace,TF:t_trace_val,G:g_guard).
?sassign(P,S,T,G,S,_,_), sum_body(SUM,P), sassign_cast(SUM,S,T,CT),
    +sassign(P,S,T,G,S,CT,G).
?sassign(P,S,trace{drf{T}},BG,SF,_,_), trace_fld_sub(SF,S,SR), SF\=S,
    sum_body(SUM,P), trace_compose(SUM,T,SR,PTF), val(P,PTF,TF,TG),
    #and(BG,TG,G), +sassign(P,S,trace{drf{T}},BG,SF,TF,G).

% S is written with T at point P. mark S as written, as well as any field F
% of S that is written to T
predicate sassign_write(P:pp,S:t_trace,T:t_trace_val,G:g_guard).
sassign_write(P,S,_,_), +access(P,S,write).
sassign_write(P,S,trace{drf{T}},_), access(PP,TF,write),
    trace_fld_sub(TF,T,TR), pp_order_prec(PP,P),
    sum_body(SUM,P), trace_compose(SUM,S,TR,SF), +access(P,SF,write).

% if S is assigned to by a set, merge the targets of the RHS with its targets
% after the set
iset(P0,P1,I), cil_instr_set(I,LV,E), lval(P0,LV,S,LG), eval(P0,E,T,RG),
    #and(LG,RG,BG), +sassign_write(P0,S,T,BG),
    ?vmerge(P1,SF,_,_), sassign(P0,S,T,BG,SF,TF,G),
    eguard(P0,P1,G,EG), +vmerge(P1,SF,TF,EG).

% if X is NOT assigned to by a set, nor is it a subfield of the structure
% assigned to by the set, merge its old targets with its targets after the set
iset(P0,P1,I), cil_instr_set(I,LV,_),
    lval(P0,LV,OS,LG), ?vmerge(P1,S,_,_), ~trace_fld_sub(S,OS,_),
    val(P0,S,T,RG), #and(LG,RG,G),
    eguard(P0,P1,G,EG), +vmerge(P1,S,T,EG).

% if X is soft, merge its old targets with its targets after the set
iset(P0,P1,_), sum_body(SUM,P0), ?vmerge(P1,S,_,_), trace_soft(SUM,S,_),
    val(P0,S,T,G), eguard(P0,P1,G,EG), +vmerge(P1,S,T,EG).

% summary transfer function, use inst_transfer
isum(P0,P1,_), ?vmerge(P1,S,_,_), inst_transfer(P0,S,T,G),
    eguard(P0,P1,G,EG), +vmerge(P1,S,T,EG).

% branches leave all values alone
branch(P,P0,_,_), ?vmerge(P0,S,_,_), val(P,S,T,G),
    eguard(P,P0,G,EG), +vmerge(P0,S,T,EG).
branch(P,_,P1,_), ?vmerge(P1,S,_,_), val(P,S,T,G),
    eguard(P,P1,G,EG), +vmerge(P1,S,T,EG).

% summary transfer function

% introduce new fresh traces based on inst_unified if the caller does not
% have an existing representation for the trace, i.e. the dereference of an
% integer value:
% foo(int **x) { *x = malloc(); }
% bar() { int *x = 0; foo(&x); }
%isum(P0,P1,SUM), ~no_exit_aliasing(SUM,P0), inst_unified(SUM,CT,FT),
%    rel_get_trace(SUM,P0,drf{CT}), inst_trace(SUM,P0,CT,trace{T},TG),
%    val(P0,T,nrep{_},NG), #and(TG,NG,G),
%    +inst_trace(SUM,P0,drf{CT},trace{FT},G), ?vmerge(P1,T,_,_),
%    eguard(P0,P1,G,EG), +vmerge(P1,T,trace{FT},EG).

% all-in-one fuzzing. S is written with T by the isum at P. T is either the
% target given by a pointer side effect on the summary, or is the previous
% value of S. NT is the final result of the write
predicate inst_write_val(in P:pp,in CT:t_trace,in S:t_trace,in T:t_trace_val,
                         NT:t_trace_val).

% for pointers to array indexes, add an unconstrained index
generate_index_traces(), ?inst_write_val(P,CT,S,trace{T},_),
    isum(P,_,CSUM), sum_body(SUM,P),
    trace_strip_index(CT,NCT), trace_soft(CSUM,NCT,SFT),
    (SFT=array; SFT=ptarray),
    trace_type(SUM,S,y_ptr{DY}), trace_coerce_index(SUM,T,DY,AT,I),
    inst_write_array(P,S,AT,AI), trace_val_add(I,AI,NI),
    +inst_write_val(P,CT,S,trace{T},trace{index{AT,DY,NI}}).

% for integers, get an unconstrained value
?inst_write_val(P,CT,S,T,_), sum_body(SUM,P), trace_type(SUM,S,Y),
    (Y=y_int{_}; Y=y_enum{_}; Y=y_flt{_}),
    inst_write_int(P,S,NT), +inst_write_val(P,CT,S,T,NT).

% for other traces, just preserve the original value
predicate inst_write_any(in P:pp,in CT:t_trace,in S:t_trace,in T:t_trace_val,
                         NT:t_trace_val).
?inst_write_any(P,CT,S,T,_), inst_write_val(P,CT,S,T,NT),
    +inst_write_any(P,CT,S,T,NT).
?inst_write_any(P,CT,S,T,_), pp_order(P,DEP,DIST), $order(DEP,DIST,4,0),
    ~inst_write_val(P,CT,S,T,_), +inst_write_any(P,CT,S,T,T).

% SUM at P may write a new value to S if G holds. use per-trace usemod
% information if available
predicate inst_may_write(in SUM:sum,in P:pp,in S:t_trace,CS:t_trace,G:g_guard).
~no_inst_may_write(), ?inst_may_write(SUM,P,S,_,_), trace_strip_index(S,NS),
    td_inst_trace(SUM,P,NS,CS,G), inst_may_access(SUM,CS,write),
    +inst_may_write(SUM,P,S,CS,G).

% isum transfer preserves all targets when the inst_transfer_necessary conds
% are met. get the conditions under which a write occurs and inst_write_any
isum(P,_,SUM), ?inst_transfer(P,S,_,_), val(P,S,T,LG),
    pp_order(P,DEP,DIST), $order(DEP,DIST,5,0), 
    \/inst_transfer_necessary(SUM,P,S,TG):#and_all(TG,MTG), #and(LG,MTG,LMTG),
    \/inst_may_write(SUM,P,S,_,WG):#or_all(WG,MWG), #not(MWG,NMWG),
    #and(LMTG,NMWG,NVWG), +inst_transfer(P,S,T,NVWG),
    inst_may_write(SUM,P,S,CS,CWG), trace_simplify(SUM,drf{CS},DCS),
    inst_write_any(P,DCS,S,T,NT),
    #and(LMTG,CWG,VWG), +inst_transfer(P,S,NT,VWG).

% compute side effects for each icall/iloop, mapping back sum_exit to the
% caller and updating the value afterwards. make fresh bits to distinguish
% between different side effects and the default targets-preserved behavior
predicate split_exit(SUM:sum,P:pp,CS:t_trace,
                     TGTS:list[t_trace],PG:g_guard).
split_exit(SUM,P,CS,[],PG),
    inst_trace(SUM,P,CS,trace{S},SG),
    #not(SG,NSG), #or(NSG,PG,G), +inst_transfer_necessary(SUM,P,S,G).
split_exit(SUM,P,CS,[CT|TAIL],PG), exit_bit(SUM,P,CS,CT,BG),
    #and(BG,PG,CPG), #not(BG,NBG), #and(NBG,PG,NPG),
    +split_exit(SUM,P,CS,TAIL,NPG),
    inst_trace(SUM,P,CS,trace{S},SG), inst_trace(SUM,P,CT,T,TG),
    #and(SG,TG,MG), #and(MG,CPG,XG), +sassign_write(P,S,T,XG),
    ?inst_transfer(P,SF,_,_), sassign(P,S,T,XG,SF,TF,G),
    inst_write_any(P,CT,SF,TF,NTF), +inst_transfer(P,SF,NTF,G).

~no_val_guard(), isum(P,_,SUM), ~no_exit_aliasing(SUM,P),
    pp_order(P,DEP,DIST), $order(DEP,DIST,0,0),
    inst_exit_points(SUM,CS,_), \/inst_exit_points(SUM,CS,CT):list_all(CT,TGT),
    #bool_g(true,G), +split_exit(SUM,P,CS,TGT,G).
no_val_guard(), isum(P,_,SUM), ~no_exit_aliasing(SUM,P),
    inst_exit_points(SUM,CS,CT),
    inst_trace(SUM,P,CS,trace{S},SG), inst_trace(SUM,P,CT,T,TG),
    #and(SG,TG,XG), +sassign_write(P,S,T,XG),
    ?inst_transfer(P,SF,_,_), sassign(P,S,T,XG,SF,TF,G),
    inst_write_any(P,CT,SF,TF,NTF), +inst_transfer(P,SF,NTF,G).

% if a function returns an alias for earlier data, kill the default
% return value by setting 'false' as a necessary condition
icall(P,_,I), ~no_exit_aliasing(s_call{I},P),
    inst_exit_points(s_call{I},root{return},_), callret(I,S),
    #bool_g(false,FG), +inst_transfer_necessary(s_call{I},P,root{S},FG).

% check that downstream analyses are doing their job

?entry_bit(P,S,T,_), ~entry_bit(P,S,T,_),
    +warning("Integer analysis must provide entry_bit:",P,S,T).
?exit_bit(SUM,P,S,T,_), ~exit_bit(SUM,P,S,T,_),
    +warning("Integer analysis must provide exit_bit:",SUM,P,S,T).

?asn_value(ASN,NR,_), ~asn_value(ASN,NR,_),
    +warning("Integer analysis must provide asn_value:",ASN,NR).

?trace_val_int(K,N,_), ~trace_val_int(K,N,_),
    +warning("Integer analysis must provide trace_val_int:",K,N).
?trace_val_add(LV,RV,_), ~trace_val_add(LV,RV,_),
    +warning("Integer analysis must provide trace_val_add:",LV,RV).
?trace_val_mul(LV,N,_), ~trace_val_mul(LV,N,_),
    +warning("Integer analysis must provide trace_val_mul:",LV,N).
?trace_val_div(LV,N,_), ~trace_val_div(LV,N,_),
    +warning("Integer analysis must provide trace_val_div:",LV,N).

?exit_index(SUM,P,S,T,_), ~exit_index(SUM,P,S,T,_),
    +warning("Integer analysis must provide exit_index:",SUM,P,S,T).

?inst_trace_val(I,P,CV,_,_), ~inst_trace_val(I,P,CV,_,_),
    +warning("Integer analysis must provide inst_trace_val:",I,P,CV).

% baked in

import "memorybaked.clp".
