%%
% Authors:
%   Peter Hawkins  <hawkinsp@cs.stanford.edu>
%
% Copyright (c) 2006,
%   The Board of Trustees of The Leland Stanford Junior University
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
% 
% 1. Redistributions of source code must retain the above copyright notice,
% this list of conditions and the following disclaimer.
% 
% 2. Redistributions in binary form must reproduce the above copyright notice,
% this list of conditions and the following disclaimer in the documentation
% and/or other materials provided with the distribution.
% 
% 3. The names of the contributors may not be used to endorse or promote
% products derived from this software without specific prior written
% permission.
% 
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

%-----------------------------------------------------------------------------%
% scalar.clp
%
% Integer/float operations with a high level symbolic representation.
%
% This is useful in combination with scalar_sat which handles SAT conversion
% and constraint solving.
%
%-----------------------------------------------------------------------------%

%-----------------------------------------------------------------------------%
% Interface
%-----------------------------------------------------------------------------%

import "memory.clp".
using solve_sat.
using map.

% ----------------------------------------------------------------------------
% Scalar expressions and guard representation

type scalar.
type scalar_cmp.

    % Names for unconstrained trace values.
type g_vecrep ::=
        % Initial value of some trace at entry point to a summary.
      vr_t{t_trace}             

        % Initial values of a soft trace. Since in general a soft trace
        % represents a set of values, we need to allow for the possibility for
        % different references to the trace's initial value referring to 
        % distinct memory locations. We use the program point and expression 
        % (if any) of the access site to obtain a unique name for each access, 
        % thereby simulating the fact that each access may return a different 
        % element of the set.
    | vr_st{t_trace, pp, maybe[c_exp]}        
    
        % Imprecise intermediate result. These are generated by operations that
        % we don't model and instead treat as uninterpreted functions.
    | vr_i{pp, c_exp}           

        % Unconstrained value of a trace after being clobbered by a call or 
        % loop side-effect.
    | vr_c{sum, pp, t_trace}    

        % Result of mapping an imprecise or clobbered value using inst_scalar.
        % Some unconstrained values can't be sensibly renamed when mapping
        % a scalar out of a callsite. To give a name to such values we
        % use the following recursive constructor to name the value by the 
        % callsite.
    | vr_instscalar{sum, pp, g_vecrep}

    % Values not generated by scalar.clp
    | vr_buffersize{t_trace}    % Size of a buffer in bytes

    % Values generated by scalar_sat.clp
    | vr_u{scalar}              % Unmodelled expression treated as an 
                                % uninterpreted function. 
    .

    % Representation for regular unconstrained bits, which can be the bits used
    % by upstream analyses, or the extra type g_axrep for extension by
    % downstream analyses
type g_xrep.
type g_arep ::=
      ar_extra{g_xrep}
    | ar_path{pp, pp}
    | ar_entry{pp, t_trace, t_trace}
    | ar_exit{sum, pp, t_trace, t_trace}
    | ar_clobber{sum, pp, t_trace}
    | ar_instguard{sum, pp, g_arep}
    .

    % Guard leaf formulae
type g_bit ::= 
      br_cmp{scalar_cmp}            % Result of a scalar comparison operation
    | br_ioverflow{scalar}          % Overflow bit of an integer expression
    | br_abit{g_arep}               % Additional bit created by the memory model
    .

% ----------------------------------------------------------------------------

    % A scalar with a guard attached.
type guarded_scalar = t_pair[scalar, g_guard].

    % A set of guarded scalars. We use a map here so we can group similar 
    % scalars with different guards.
type guarded_scalar_set = map[scalar, g_guard].

    % The possible kinds of scalar expressions.
type t_scalar_kind ::= 
      sk_int{Signed:bool, Bits:int} 
    | sk_float{Bits:int}.

type scalar_unop.
type scalar_binop.

    % An scalar expression
    % Width fields are given as a number of bits
type scalar ::=
        % Integer and floating point constants
      s_int_const{int, Signed:bool, Bits:int}   
    | s_float_const{float, Bits:int}

        % Uninterpreted int/float expressions which represent something we 
        % can't model directly, such as the initial value of a variable or 
        % the value of an inexact operation. 
    | s_unconstrained{g_vecrep, t_scalar_kind} 

    | s_cast{scalar, t_scalar_kind}             % Signedness and/or width cast
    | s_guard{guarded_scalar_set}               % A list of guarded integer 
                                                % expressions

    | s_unop{scalar_unop, scalar}               % Unary operations
    | s_binop{scalar_binop, scalar, scalar}     % Binary operations

    .

    % Unary integer operations
type scalar_unop ::=
      su_neg                                % Unary negation 
    | su_not                                % Binary negation
    .

    % Binary integer operations
type scalar_binop ::=
      sb_plus                               % Addition
    | sb_minus                              % Subtraction
    | sb_mul                                % Multiplication
    | sb_div                                % Division
    | sb_mod                                % Modulus 
    | sb_shiftlt                            % Left shift
    | sb_shiftrt                            % Right shift 
    | sb_or                                 % Disjunction
    | sb_and                                % Conjunction
    | sb_xor                                % eXclusive OR
    .

    % Integer comparisons
type t_scalar_cmp ::=
      sc_eq     % ==
    | sc_ne     % !=
    | sc_lt     % <
    | sc_le     % <=
    | sc_gt     % >
    | sc_ge     % >=
    .

    % scalar_cmp_op(SC, Str, Prec) holds for each possible sc_... constant.
    % Str is a string representation of the comparison operator, and Prec
    % is a relative operator precedence.
predicate scalar_cmp_op(t_scalar_cmp, string, int).

    % cil_cmp_binop_to_scalar_cmp(Op, SC)
    % Map CIL comparison ops to our types
predicate cil_cmp_binop_to_scalar_cmp(out binop, out t_scalar_cmp).

type scalar_cmp ::=
      sc{t_scalar_cmp, scalar, scalar}
    | sc_eqz{scalar}        % Shorthand for 0 == ...
    .
    
    % Intermediate integer/floating-point values are scalars
type t_nrep = scalar.

% ----------------------------------------------------------------------------
% Publically exported predicates 

    % evals(P, E, I)
    % At program point P, expression E evaluates to scalar I
predicate evals(in pp, in c_exp, out scalar).

    % vals(P, T, IE)
    % At program point P, trace T has scalar value IE
predicate vals(in pp, in t_trace, out scalar).


    % inst_guard(I, P, RG, NG)
    % As interpreted by the call/loop I at point P, callee guard RG maps to 
    % caller guard NG
predicate inst_guard(in sum, in pp, in g_guard, out g_guard).

    % inst_scalar(I, P, RI, NI)
    % As interpreted by the call/loop I at point P, scalar expression RI maps 
    % to NI
predicate inst_scalar(in sum, in pp, in scalar, out scalar).

    % inst_scalar_cmp(I, P, RI, G)
    % As interpreted by the call/loop I at point P, scalar comparision RI maps 
    % to G
predicate inst_scalar_cmp(in sum, in pp, in scalar_cmp, out g_guard).

% applies simple constant folding etc. to a scalar, preserving its input
% if no simplifications could be applied
predicate simplify_scalar(in scalar, out scalar).
predicate simplify_scalar_cmp(in scalar_cmp, out g_guard).

    % xyz_scl_used(I, S)
    % Scalar S is referenced by expression/guard/comparison I
predicate scalar_scl_used(in scalar, out scalar).
predicate scalar_cmp_scl_used(in scalar_cmp, out scalar).
predicate guard_scl_used(in g_guard, out scalar).

    % xyz_trace_used(I, T)
    % Trace T is referenced by integer expression/guard/comparison I
predicate scalar_trace_used(in scalar, out t_trace).
predicate scalar_cmp_trace_used(in scalar_cmp, out t_trace).
predicate guard_trace_used(in g_guard, out t_trace).

% ----------------------------------------------------------------------------

    % Given a scalar, determine its kind.
predicate scalar_kind(in scalar, out t_scalar_kind).

    % Given a trace, determine its kind.
predicate trace_scalar_kind(in sum, in t_trace, out t_scalar_kind).

    % Given a type, determine whether it is a scalar type (ie. an int, float
    % or enum).
predicate type_scalar(in t_type).

    % Given a type, determine the corresponding scalar_kind that should be
    % used to represent it.
    % NB. Some non-scalar types (e.g. pointers) will be assigned kinds
    % by this predicate.
predicate type_scalar_kind(in t_type, out t_scalar_kind).

    % traceval_scalar(P, V, Kind, IE)
    % Convert a trace_val V at program point P to an scalar expression IE.
    % The initial value of traces at function or loop entry is an 
    % unconstrained vector, or the integer expression itself in
    % other cases. Takes a program point and optional expression in order
    % to uniquely name different occurrences of soft traces.
predicate traceval_scalar(in pp, in maybe[c_exp], in t_trace_val, 
    in t_scalar_kind, out scalar).

% ----------------------------------------------------------------------------
% Guard construction predicates

% Equivalents of bv_eqz, bv_nez, bv_lt, bv_gt etc 
predicate g_eqz(in IE: scalar,                out G:g_guard).
predicate g_nez(in IE: scalar,                out G:g_guard).
predicate g_eq( in IE1:scalar, in IE2:scalar, out G:g_guard).
predicate g_ne( in IE1:scalar, in IE2:scalar, out G:g_guard).
predicate g_gt( in IE1:scalar, in IE2:scalar, out G:g_guard).
predicate g_lt( in IE1:scalar, in IE2:scalar, out G:g_guard).
predicate g_le( in IE1:scalar, in IE2:scalar, out G:g_guard).
predicate g_ge( in IE1:scalar, in IE2:scalar, out G:g_guard).

    % g_cmp(Op, S1, S2, G)
    % Produce a comparison guard between two scalars using the specified
    % operator.
predicate g_cmp(in t_scalar_cmp, in IE1:scalar, in IE2:scalar, 
    out G:g_guard).

    % g_range(S, Min, Max, G)
    % Produce a guard G that holds if S is between Min and Max inclusive.
    % Only defined for integer-valued scalars.
predicate g_range(in scalar, in int, in int, out g_guard).

    % Extract the overflow bit of an integer expression
predicate g_intoverflow(in scalar, out g_guard).

% ----------------------------------------------------------------------------
% Scalar construction predicates

    % Create a scalar value representing an integer constant.
    % The scalar size will be chosen to be large enough to represent
    % the integer value without overflow.
predicate sc_int(in int, out scalar).

    % Cast a scalar to a different kind
predicate sc_cast(in scalar, in t_scalar_kind, out scalar).

    % Given a list of guarded scalars, simplify the list
    % if possible and return a single scalar representing the set.
predicate sc_guarded(in list[guarded_scalar], out scalar).

% ----------------------------------------------------------------------------
% Stringification

predicate guard_string(in g_guard, out string).
predicate scalar_string(in scalar, out string).
predicate scalar_cmp_string(in scalar_cmp, out string).

% ----------------------------------------------------------------------------

    % If debug_scalar is defined, warn on failure of predicates such as
    % eval and evals.
predicate debug_scalar().

    % Don't clobber scalars as a result of call/loop side effects.
predicate no_clobber_scalars().

    % sum_write(Sum, T)
    % Summary Sum or its callees may write to trace T.
predicate sum_write(sum, t_trace).

% ----------------------------------------------------------------------------
% Implementation.
% ----------------------------------------------------------------------------

scalar_cmp_op(sc_eq, "==", 5) :- .
scalar_cmp_op(sc_ne, "!=", 5) :- .
scalar_cmp_op(sc_lt, "<", 6) :- .
scalar_cmp_op(sc_le, "<=", 6) :- .
scalar_cmp_op(sc_ge, ">=", 6) :- .
scalar_cmp_op(sc_gt, ">", 6) :- .

cil_cmp_binop_to_scalar_cmp(b_eq, sc_eq) :- .
cil_cmp_binop_to_scalar_cmp(b_ne, sc_ne) :- .
cil_cmp_binop_to_scalar_cmp(b_lt, sc_lt) :- .
cil_cmp_binop_to_scalar_cmp(b_gt, sc_gt) :- .
cil_cmp_binop_to_scalar_cmp(b_le, sc_le) :- .
cil_cmp_binop_to_scalar_cmp(b_ge, sc_ge) :- .

% ----------------------------------------------------------------------------

g_eqz(RES, G)     :- simplify_scalar_cmp(sc_eqz{RES}, G).
g_nez(RES, NG)    :- simplify_scalar_cmp(sc_eqz{RES}, G), #not(G, NG).
g_eq(IE1, IE2, G) :- simplify_scalar_cmp(sc{sc_eq, IE1, IE2}, G).
g_ne(IE1, IE2, G) :- simplify_scalar_cmp(sc{sc_ne, IE1, IE2}, G).
g_lt(IE1, IE2, G) :- simplify_scalar_cmp(sc{sc_lt, IE1, IE2}, G).
g_gt(IE1, IE2, G) :- simplify_scalar_cmp(sc{sc_gt, IE1, IE2}, G).
g_le(IE1, IE2, G) :- simplify_scalar_cmp(sc{sc_le, IE1, IE2}, G).
g_ge(IE1, IE2, G) :- simplify_scalar_cmp(sc{sc_ge, IE1, IE2}, G).

g_cmp(Cmp, IE1, IE2, G) :- #id_g(br_cmp{sc{Cmp, IE1, IE2}}, G).

g_range(S, Min, Max, G) :-
    scalar_kind(S, sk_int{Signed, Bits}),
    g_ge(S, s_int_const{Min, Signed, Bits}, G1),
    g_le(S, s_int_const{Max, Signed, Bits}, G2),
    #and(G1, G2, G).

g_intoverflow(IE, G) :- #id_g(br_ioverflow{IE}, G).

% ----------------------------------------------------------------------------

    % nbits(N, Bits)
    % Non-negative integer N can be represented exactly using Bits bits.
predicate nbits(in int, out int).

nbits(0, 0) :- .
nbits(N, B) :-
    int_gt(N, 0),
    int_div(N, 2, N2),
    nbits(N2, B2),
    int_add(B2, 1, B).

    % Negative integer constant to scalar
sc_int(N, s_int_const{N, true, NBits}) :-
    int_lt(N, 0),
    int_neg(N, M),
    nbits(M, MBits),
    int_add(MBits, 1, NBits).

    % Positive integer constant to scalar
sc_int(N, s_int_const{N, false, NBits}) :-
    int_ge(N, 0),
    nbits(N, NBits).

% ----------------------------------------------------------------------------

sc_cast(S0, Kind, S0) :-
    scalar_kind(S0, OldKind),
    OldKind = Kind.

sc_cast(S0, Kind, S) :-
    scalar_kind(S0, OldKind),
    OldKind \= Kind,
    simplify_scalar(s_cast{S0, Kind}, S).

% ----------------------------------------------------------------------------

predicate guarded_scalar_list_to_map(in list[guarded_scalar],
    in guarded_scalar_set, out guarded_scalar_set).

predicate trivial_guarded_expr(in guarded_scalar_set, out scalar).

sc_guarded(GS, T) :-
    map_empty(M0),
    guarded_scalar_list_to_map(GS, M0, M),
    trivial_guarded_expr(M, T).

sc_guarded(GS, s_guard{M}) :-
    map_empty(M0),
    guarded_scalar_list_to_map(GS, M0, M),
    ~trivial_guarded_expr(M, _).

guarded_scalar_list_to_map([], M, M) :- .
guarded_scalar_list_to_map([pair{S, G0}|GSs], M0, M) :-
    map_search(M0, S, G1),
    #or(G0, G1, G),
    map_insert(S, G, M0, M1),
    guarded_scalar_list_to_map(GSs, M1, M).

guarded_scalar_list_to_map([pair{S, G}|GSs], M0, M) :-
    ~map_search(M0, S, _),
    map_insert(S, G, M0, M1),
    guarded_scalar_list_to_map(GSs, M1, M).

trivial_guarded_expr(M, I) :-
    map_to_sorted_list(M, [pair{I, G}]),
    #bool_g(true, G).

% ----------------------------------------------------------------------------
% Unconstrained bits for upstream analyses 


path_bit(P0, P1, G) :- #id_g(br_abit{ar_path{P0, P1}}, G).
entry_bit(P, S, T, G) :- #id_g(br_abit{ar_entry{P, S, T}}, G).
exit_bit(SUM, P, S, T, G) :- #id_g(br_abit{ar_exit{SUM, P, S, T}}, G).

% ----------------------------------------------------------------------------
% Value clobbering through call/loop side effects.

import "mustmod.clp".

predicate clobber_bit(in sum, in pp, in t_trace, out g_guard).

    % Must-mod traces are always clobbered.
clobber_bit(SUM, P, T, G) :- 
    ~inst_mustmod(SUM, T),
    #id_g(br_abit{ar_clobber{SUM, P, T}}, G).
clobber_bit(SUM, P, T, G) :- 
    inst_mustmod(SUM, T),
    #bool_g(true, G).

predicate clobber(in sum, in pp, t_trace, t_trace, t_scalar_kind, g_guard).

    % Scalar call side-effects. Requires precise usemod information.
    % For every scalar valued trace possibly modified by a call, introduce a new
    % unconstrained value for the value of the trace after the call.
    % We don't handle return values here since they already have temporary
    % values created for them by the CFG construction code.
clobber(ISum, P0, ST, T, Kind, FG) :-
    ~no_clobber_scalars(),
    sum_body(Sum, P0),
    inst_usemod(ISum, ST, write),
    ST \= root{return},
    inst_trace(ISum, P0, ST, trace{T}, G),
    trace_type(Sum, T, Type),
    type_scalar(Type),
    type_scalar_kind(Type, Kind),
    clobber_bit(ISum, P0, ST, CG),
    #and(CG, G, FG). 
    

inst_transfer_necessary(ISum, P0, T, PG) :-
    isum(P0, _, ISum),
    pp_order(P0, Dep, Dist),
    $order(Dep, Dist, 100, 0),
    clobber(ISum, P0, _, T, _, FG),
    #not(FG, PG). % By Demorgan's law PG = (not CG) or (not G)

vmerge(P1, T, nrep{s_unconstrained{vr_c{ISum, P0, ST}, Kind}}, FG) :-
    isum(P0, P1, ISum),
    clobber(ISum, P0, ST, T, Kind, G),
    eguard(P0, P1, G, FG).

% ----------------------------------------------------------------------------

    % Values clobbered by callers are written in the summary
sum_write(Sum, T) :-
    isum(P0, _, ISum),
    clobber(ISum, P0, _, T, _, _),
    sum_body(Sum, P0).

    % Values directly modified by the summary are modified in the summary
sum_write(Sum, T) :-
    access(P, T, write),
    sum_body(Sum, P).

% ----------------------------------------------------------------------------

type_scalar(y_int{_}) :- .
type_scalar(y_flt{_}) :- .
type_scalar(y_enum{_}) :- .

% ----------------------------------------------------------------------------

type_scalar_kind(y_int{IKind}, sk_int{Signed, Bits}) :-
    ikind_bits(IKind, Signed, Bits).

type_scalar_kind(y_enum{_}, sk_int{Signed, Bits}) :-
    ikind_bits(iint, Signed, Bits).

type_scalar_kind(y_flt{FKind}, sk_float{Bits}) :-
    fkind_bits(FKind, Bits).

type_scalar_kind(y_ptr{_}, sk_int{false, Bits}) :-
    pointer_bits(Bits).


% Kinded version
traceval_scalar(P, E, trace{T}, Kind, s_unconstrained{vr_t{T}, Kind}) :- 
    sum_body(Sum, P), ~trace_soft(Sum, T, _).
traceval_scalar(P, E, trace{T}, Kind, s_unconstrained{vr_st{T, P, E}, Kind}) :- 
    sum_body(Sum, P), trace_soft(Sum, T, _).

traceval_scalar(_, _, nrep{V}, _Kind, V) :- .

% Kindless version
predicate traceval_scalar_nokind(in t_trace_val, out scalar).
traceval_scalar_nokind(trace{drf{T}}, s_unconstrained{vr_t{drf{T}}, Kind}) :- 
    trace_type(s_func, T, Type),
    type_scalar_kind(Type, Kind).

traceval_scalar_nokind(trace{T}, s_unconstrained{vr_t{T}, sk_int{false, Bits}})
    :- 
    T \= drf{_},
    pointer_bits(Bits).

traceval_scalar_nokind(nrep{V}, V) :- .

?traceval_scalar_nokind(T, _), ~traceval_scalar_nokind(T, _),     
    +warning("traceval_scalar failed", T).

%?- trace_type(A,B,C).

trace_scalar_kind(Sum, T, K) :-
    trace_type(Sum, T, Type),
    type_scalar_kind(Type, K).

% ----------------------------------------------------------------------------
% Convert bitvectors to locations/guards

    % For non-lvalue scalar expressions (memory.clp handles all lvalue exps),
    % evaluate the expression as an scalar and use that representation
eval(P, E, nrep{Scalar}, G) :-
    ~omit_eval(P, E),
    exp_type(E, EType), 
    convert_type(EType, Type),
    type_scalar(Type),
    ~cil_exp_lval(E,_), 
    evals(P, E, Scalar), 
    #bool_g(true, G).

eval(P, E, nrep{Scalar}, G) :-
    ~omit_eval(P, E),
    cil_exp_cast(E, CE, _),
    exp_tptr(E),
    ~exp_tptr(CE),
    evals(P, E, Scalar),
    #bool_g(true, G).

    % For scalar expressions evaluated as a boolean condition, evaluate as
    % a bitvector and compute the NEZ condition
beval(P, E, BG, G) :-
    ~omit_beval(P,E),
    ~exp_boolean(E), 
    ~exp_tptr_as_boolean(E),
    evals(P, E, IExpr), 
    g_nez(IExpr, BG),
    #bool_g(true, G). 

% ----------------------------------------------------------------------------
% evals - Integer expression evaluation

    % Integer constant expressions
evals(P, E, s_int_const{N, Signed, Bits}) :-
    cil_exp_const(E, _), 
    cil_exp_x_intval(E, N), 
    exp_tint(E, K), 
    ikind_bits(K, Signed, Bits).

    % Floating-point constant expressions
evals(P, E, s_float_const{N, Bits}) :-
    cil_exp_const(E, C), 
    cil_const_real(C, _, N), 
    exp_tflt(E, K), 
    fkind_bits(K, Bits).

    % Sizeof/alignof expressions
evals(P, E, s_int_const{N, S, Length}) :-
    cil_exp_x_sizeof_alignof(E), 
    cil_exp_x_intval(E, N), 
    exp_tint(E, K), 
    ikind_bits(K, S, Length).

    % CIL should be constant folding these &(0->field) but it's not.
    % just get an unconstrained value
evals(P, E, s_unconstrained{vr_i{P, E}, sk_int{Signed, Length}}) :-
    cil_exp_x_intval(MCE, 0), 
    cil_exp_cast(ME, MCE, _), 
    cil_lval_mem(LV, ME, _),
    cil_exp_addr(AE, LV),
    cil_exp_cast(E, AE, _),
    exp_tint(E, K), 
    ikind_bits(K, Signed, Length).

    % Cast expressions.
evals(P, E, S) :-
    cil_exp_cast(E, CE, _),
    exp_type(E, NewType),
    convert_type(NewType, NewY),
    type_scalar_kind(NewY, Kind),
    evals(P, CE, CS),
    sc_cast(CS, Kind, S).

% ----------------------------------------------------------------------------
% Evaluation of guarded scalar expressions

    % imerge(P, E, I)
    % Adding will merge I into the evaluation of E at P
predicate imerge(in pp, in c_exp, out guarded_scalar).

evals(P, E, I) :-
    imerge(P, E, _), 
    pp_order(P, Dep, Dist), exp_order(E, EDist), $order(Dep, Dist, 2, EDist), 
    \/imerge(P, E, IM):list_all(IM, Is), 
    sc_guarded(Is, I).

% Evaluate integer lvals by casting all possible targets of the
% lvalue to vectors and merging them together according to the guards
imerge(P, E, pair{IE, G}) :-
    cil_exp_lval(E, _), 
    exp_type(E, EType),
    convert_type(EType, Type),
    type_scalar(Type),
    eval(P, E, T, G), 
    ~#bool_g(false, G),             % Omit trivially false cases
    type_scalar_kind(Type, Kind),
    traceval_scalar(P, yes{E}, T, Kind, IE).

% we need a bit vector for pointer targets for some situations such as, e.g.
% a ptr->uint cast.
imerge(P, E, pair{IE, G}) :-
    exp_tptr(E),
    eval(P, E, T, G),
    ~#bool_g(false, G),             % Omit trivially false cases
    exp_type(E, Type),
    convert_type(Type, Y),
    type_scalar_kind(Y, Kind),
    traceval_scalar(P, yes{E}, T, Kind, IE).

% Convert guards to bit vectors

% Only needs to be done for exp_boolean expressions, for which evals is not
% defined. Do the cast by returning a guarded scalar '1'.
imerge(P, E, pair{s_int_const{1, Signed, Length}, TG}) :-
    exp_boolean(E), 
    exp_tint(E, K), 
    ikind_bits(K, Signed, Length),
    beval(P, E, BG, G),
    #and(BG, G, TG).

% ----------------------------------------------------------------------------
% Arithmetic unary expressions

    % sunop(Op, RI, I)
    % The result of applying arithmetic unary operation Op to RI is I
predicate sunop(out unop, out scalar_unop).
sunop(u_neg,  su_neg) :- .
sunop(u_bnot, su_not) :- .

evals(P, E, S) :-
    cil_exp_unop(E,OP,RE,_),
    evals(P, RE, RI),
    sunop(OP, IOp),
    simplify_scalar(s_unop{IOp, RI}, S).

% ----------------------------------------------------------------------------
% Arithmetic binary operation expressions

    % Get an unconstrained value for arithmetic operations on pointers
    % (minuspp) 
    % XXX: actually subtraction on pointers can overflow. We'll deal
    % with this later.
evals(P, E, s_unconstrained{vr_i{P, E}, sk_int{Signed, Length}}) :- 
    pointer_int_op(OP), 
    cil_exp_binop(E, OP, _, _, _),
    exp_tint(E, K),
    ikind_bits(K, Signed, Length).


predicate commutative_binop(out binop, out scalar_binop).
commutative_binop(b_plusa,   sb_plus) :- .
commutative_binop(b_band,    sb_and) :- .
commutative_binop(b_bxor,    sb_xor) :- .
commutative_binop(b_bor,     sb_or) :- .
commutative_binop(b_mult,    sb_mul) :- .

predicate noncommutative_binop(out binop, out scalar_binop).
noncommutative_binop(b_minusa,  sb_minus) :- .
noncommutative_binop(b_div,     sb_div) :- .
noncommutative_binop(b_mod,     sb_mod) :- .
noncommutative_binop(b_shiftlt, sb_shiftlt) :- .
noncommutative_binop(b_shiftrt, sb_shiftrt) :- .

evals(P, E, S) :-
    cil_exp_binop(E, OP, LE, RE, _), 
    commutative_binop(OP, IOp),
    evals(P, LE, LS),
    evals(P, RE, RS),
    % Lexicographic comparison for canonicalization
    (
        lessthan(LS, RS),
        S1 = LS, S2 = RS
    ;   
        LS = RS,
        S1 = LS, S2 = RS
    ;   
        lessthan(RS, LS),
        S1 = RS, S2 = LS
    ),
    simplify_scalar(s_binop{IOp, S1, S2}, S).

evals(P, E, S) :- 
    cil_exp_binop(E, OP, LE, RE, _), 
    noncommutative_binop(OP, IOp),
    evals(P, LE, LS),
    evals(P, RE, RS),
    simplify_scalar(s_binop{IOp, LS, RS}, S).

% ----------------------------------------------------------------------------
% vals - trace evaluation

    % vimerge(P, T, IE)
    % IE is a possible value for trace T at program point P
predicate vimerge(in pp, in t_trace, out guarded_scalar).

vals(P, Trace, I) :-
    vimerge(P, Trace, _),
    pp_order(P, Dep, Dist), $order(Dep, Dist, 1, 0),
    \/vimerge(P, Trace, IM):list_all(IM, Is),
    sc_guarded(Is, I).

vimerge(P, T, pair{IE, G}) :-
    val(P, T, V, G),
    trace_type(s_func, T, Type),
    type_scalar_kind(Type, Kind),
    traceval_scalar(P, no, V, Kind, IE).

% ----------------------------------------------------------------------------

predicate inst_trace_scalar_kind(in sum, in pp, in t_scalar_kind, in t_trace, out scalar).

% Kinded version
predicate itmerge_kind(in sum, in pp, in t_scalar_kind, in t_trace, 
    out guarded_scalar).
inst_trace_scalar_kind(Sum, P, Kind, Trace, I) :-
    itmerge_kind(Sum, P, Kind, Trace, _),
    pp_order(P, Dep, Dist), $order(Dep, Dist, 3, 0),
    \/itmerge_kind(Sum, P, Kind, Trace, IM):list_all(IM, Is),
    sc_guarded(Is, I).

itmerge_kind(Sum, P, Kind, CalleeTrace, pair{E, G}) :-
    inst_trace(Sum, P, CalleeTrace, TraceVal, G),
    ~#bool_g(false, G),             % Omit trivially false cases
    traceval_scalar(P, no, TraceVal, Kind, E).

itmerge_kind(Sum, P, Kind, CalleeTrace, pair{TS, True}) :-
    trace_root(CalleeTrace, temp{_, _}), 
    #bool_g(true, True),
    TS = s_unconstrained{
        vr_instscalar{Sum, P, vr_t{CalleeTrace}},
        Kind
    }.

itmerge_kind(Sum, P, Kind, CalleeTrace, pair{TS, True}) :-
    Sum = s_call{_},
    trace_root(CalleeTrace, local{_}), 
    #bool_g(true, True),
    TS = s_unconstrained{
        vr_instscalar{Sum, P, vr_t{CalleeTrace}},
        Kind
    }.
    

% -----------------------------------------------------------------------------
% Binary comparison expressions


% compare_op binops should be beval'ed, according to paths.clp
beval(P, E, BG, G) :-
    ~omit_beval(P,E),
    cil_exp_binop(E, OP, LE, RE, _), 
    compare_op(OP), 
    evals(P, LE, LBI), 
    evals(P, RE, RBI), 
    cil_cmp_binop_to_scalar_cmp(OP, Cmp),
    simplify_scalar_cmp(sc{Cmp, LBI, RBI}, BG),
    #bool_g(true, G).

% -----------------------------------------------------------------------------
% Guard reevaluation

    % Base cases -- trivially true or false
inst_guard(SUM, P, G, G) :- #bool_g(true, G).
inst_guard(SUM, P, G, G) :- #bool_g(false, G).

    % Conjunction
inst_guard(SUM, P, RG, NG) :- 
    #and(RG0, RG1, RG),
    inst_guard(SUM, P, RG0, NG0), 
    inst_guard(SUM, P, RG1, NG1),
    #and(NG0, NG1, NG).

    % Disjunction
inst_guard(SUM, P, RG, NG) :-
    #or(RG0, RG1, RG),
    inst_guard(SUM, P, RG0, NG0), 
    inst_guard(SUM, P, RG1, NG1),
    #or(NG0, NG1, NG).

    % Negation
inst_guard(SUM, P, RG, NG) :-
    #not(RG0, RG),
    inst_guard(SUM, P, RG0, NG0), 
    #not(NG0, NG).

    % Scalar comparisons
inst_guard(Sum, P, RG, NG) :-
    #id_g(br_cmp{RC}, RG),
    inst_scalar_cmp(Sum, P, RC, NG).

    % Integer overflow bits
inst_guard(Sum, P, RG, NG) :-
    #id_g(br_ioverflow{RE}, RG),
    inst_scalar(Sum, P, RE, NE),
    #id_g(br_ioverflow{NE}, NG).

    % Additional bits
inst_guard(Sum, P, RG, NG) :-
    #id_g(br_abit{RE}, RG),
    #id_g(br_abit{ar_instguard{Sum, P, RE}}, NG).

?inst_guard(Sum, P, RG, _), ~inst_guard(Sum, P, RG, _),
    +warning("inst_guard failed on", Sum, P, RG).

% -----------------------------------------------------------------------------
% Scalar expression reevaluation

inst_scalar(Sum, P, s_unconstrained{vr_t{Trace}, Kind}, IE) :-
    inst_trace_scalar_kind(Sum, P, Kind, Trace, IE).

inst_scalar(Sum, P, s_unconstrained{VR, Kind}, s_unconstrained{NVR, Kind}) :-
    VR \= vr_t{_},
    NVR = vr_instscalar{Sum, P, VR}.

predicate inst_sgexprs(in sum, in pp, in list[guarded_scalar], 
    out list[guarded_scalar]).

inst_scalar(Sum, P, s_guard{GIS0}, s_guard{GIS}) :- 
    map_to_sorted_list(GIS0, GL0),
    inst_sgexprs(Sum, P, GL0, GL),
    map_of_list(GL, GIS).

inst_sgexprs(_Sum, _P, [], []) :- .
inst_sgexprs(Sum, P, [pair{I0, G0}|Is0], [pair{I, G}|Is]) :- 
    inst_guard(Sum, P, G0, G),
    inst_scalar(Sum, P, I0, I),
    inst_sgexprs(Sum, P, Is0, Is).

inst_scalar(Sum, P, s_int_const{V, S, W}, s_int_const{V, S, W}) :- .
inst_scalar(Sum, P, s_float_const{V, W}, s_float_const{V, W}) :- .
inst_scalar(Sum, P, s_cast{IE0, Kind}, S) :-
    inst_scalar(Sum, P, IE0, IE),
    simplify_scalar(s_cast{IE, Kind}, S).
inst_scalar(Sum, P, s_unop{Op, I0}, S) :- 
    inst_scalar(Sum, P, I0, I),
    simplify_scalar(s_unop{Op, I}, S).
inst_scalar(Sum, P, s_binop{Op, A0, B0}, S) :-
    inst_scalar(Sum, P, A0, A),
    inst_scalar(Sum, P, B0, B),
    simplify_scalar(s_binop{Op, A, B}, S).

?inst_scalar(Sum, P, S, _), ~inst_scalar(Sum, P, S, _),
    +warning("inst_scalar failed on", Sum, P, S).

% -----------------------------------------------------------------------------
% Comparison reevaluation

inst_scalar_cmp(Sum, P, sc{Cmp, A0, B0}, G) :-
    inst_scalar(Sum, P, A0, A),
    inst_scalar(Sum, P, B0, B),
    simplify_scalar_cmp(sc{Cmp, A, B}, G).
inst_scalar_cmp(Sum, P, sc_eqz{A0}, G) :-
    inst_scalar(Sum, P, A0, A),
    simplify_scalar_cmp(sc_eqz{A}, G).

?inst_scalar_cmp(Sum, P, SC, _), ~inst_scalar_cmp(Sum, P, SC, _),
    +warning("inst_scalar_cmp failed on", Sum, P, SC).

% ----------------------------------------------------------------------------
% Scalar simplification

% try to simplify a scalar
predicate try_simplify_scalar(in scalar, out scalar).
simplify_scalar(S,NS) :- try_simplify_scalar(S,NS).
simplify_scalar(S,S) :- ~try_simplify_scalar(S,_).

% try to simplify a scalar comparison
predicate try_simplify_scalar_cmp(in scalar_cmp, out g_guard).
simplify_scalar_cmp(S,G) :- try_simplify_scalar_cmp(S,G).
simplify_scalar_cmp(S,G) :- ~try_simplify_scalar_cmp(S,_), #id_g(br_cmp{S},G).

% compute the signed modulus of BN according to SIGN/LEN
predicate signed_modulus(in BN:int, in SIGN:bool, in LEN:int, out N:int).
signed_modulus(BN,false,LEN,N) :-
    int_exp(2,LEN,EXP),
    int_mod(BN,EXP,N).
signed_modulus(BN,true,LEN,N) :-
    int_exp(2,LEN,EXP),
    int_sub(LEN,1,MLEN), int_exp(2,MLEN,MEXP),
    int_mod(BN,EXP,BMN),
    (int_lt(BMN,MEXP), N=BMN;
     int_ge(BMN,MEXP), int_sub(BMN,EXP,N)).

try_simplify_scalar(s_cast{CS, sk_int{SIGN,LEN}}, S) :-
    CS = s_int_const{CN,_,_},
    signed_modulus(CN,SIGN,LEN,NN),
    S = s_int_const{NN,SIGN,LEN}.

predicate const_fold_binop(in scalar_binop, in LN:int, in RN:int, out N:int).

const_fold_binop(sb_plus,    LN,RN,N) :- int_add(LN,RN,N).
const_fold_binop(sb_and,     LN,RN,N) :- int_band(LN,RN,N).
const_fold_binop(sb_xor,     LN,RN,N) :- int_bxor(LN,RN,N).
const_fold_binop(sb_or,      LN,RN,N) :- int_bor(LN,RN,N).
const_fold_binop(sb_mul,     LN,RN,N) :- int_mul(LN,RN,N).
const_fold_binop(sb_minus,   LN,RN,N) :- int_sub(LN,RN,N).
const_fold_binop(sb_div,     LN,RN,N) :- int_div(LN,RN,N).
const_fold_binop(sb_mod,     LN,RN,N) :- int_mod(LN,RN,N).
const_fold_binop(sb_shiftlt, LN,RN,N) :- int_exp(2,RN,EXP), int_mul(LN,EXP,N).
const_fold_binop(sb_shiftrt, LN,RN,N) :- int_exp(2,RN,EXP), int_div(LN,EXP,N).

try_simplify_scalar(s_binop{OP, LS, RS}, S) :-
    LS = s_int_const{LN,SIGN,LEN},
    RS = s_int_const{RN,SIGN,LEN},
    const_fold_binop(OP,LN,RN,BN),
    signed_modulus(BN,SIGN,LEN,NN),
    S = s_int_const{NN,SIGN,LEN}.

predicate const_fold_cmp(in t_scalar_cmp, in LN:int, in RN:int).

const_fold_cmp(sc_eq, LN,RN) :- int_eq(LN,RN).
const_fold_cmp(sc_ne, LN,RN) :- int_ne(LN,RN).
const_fold_cmp(sc_lt, LN,RN) :- int_lt(LN,RN).
const_fold_cmp(sc_le, LN,RN) :- int_le(LN,RN).
const_fold_cmp(sc_gt, LN,RN) :- int_gt(LN,RN).
const_fold_cmp(sc_ge, LN,RN) :- int_ge(LN,RN).

try_simplify_scalar_cmp(sc{OP, LS, RS}, G) :-
    LS = s_int_const{LN,SIGN,LEN},
    RS = s_int_const{RN,SIGN,LEN},
    (const_fold_cmp(OP,LN,RN), #bool_g(true,G);
     ~const_fold_cmp(OP,LN,RN), #bool_g(false,G)).

try_simplify_scalar_cmp(sc_eqz{S}, G) :-
    S = s_int_const{N,_,_},
    (int_eq(N,0), #bool_g(true,G);
     int_ne(N,0), #bool_g(false,G)).

% -----------------------------------------------------------------------------
% Set of scalars/vectors/traces used by an integer expression

    % Expressions
scalar_scl_used(S, S) :- .
scalar_scl_used(s_cast{IE, _}, S) :- 
    scalar_scl_used(IE, S).
scalar_scl_used(s_guard{GS}, S) :- 
    map_to_sorted_list(GS, GEL),
    list_mem(GEL, pair{Expr, _Guard}),
    scalar_scl_used(Expr, S).
scalar_scl_used(s_guard{GS}, S) :- 
    map_to_sorted_list(GS, GEL),
    list_mem(GEL, pair{_Expr, Guard}),
    guard_scl_used(Guard, S).
scalar_scl_used(s_unop{_, IE}, S) :-
    scalar_scl_used(IE, S).
scalar_scl_used(s_binop{_, LE, _}, S) :-
    scalar_scl_used(LE, S).
scalar_scl_used(s_binop{_, _, RE}, S) :-
    scalar_scl_used(RE, S).

    % Comparisons 
scalar_cmp_scl_used(sc{_, LE, _}, S) :-
    scalar_scl_used(LE, S).
scalar_cmp_scl_used(sc{_, _, RE}, S) :-
    scalar_scl_used(RE, S).
scalar_cmp_scl_used(sc_eqz{IE}, S) :-
    scalar_scl_used(IE, S).

    % Guards 
guard_scl_used(RG, S) :- 
    #and(RG0, _RG1, RG),
    guard_scl_used(RG0, S). 
guard_scl_used(RG, S) :- 
    #and(_RG0, RG1, RG),
    guard_scl_used(RG1, S). 
guard_scl_used(RG, S) :- 
    #or(RG0, _RG1, RG),
    guard_scl_used(RG0, S). 
guard_scl_used(RG, S) :- 
    #or(_RG0, RG1, RG),
    guard_scl_used(RG1, S). 
guard_scl_used(RG, S) :- 
    #not(RG0, RG),
    guard_scl_used(RG0, S). 
guard_scl_used(RG, S) :- 
    #id_g(br_cmp{RC}, RG),
    scalar_cmp_scl_used(RC, S). 
guard_scl_used(RG, S) :- 
    #id_g(br_ioverflow{RE}, RG),
    scalar_scl_used(RE, S).

scalar_trace_used(S,T) :- scalar_scl_used(S,s_unconstrained{vr_t{T},_}).
scalar_trace_used(S,T) :- scalar_scl_used(S,s_unconstrained{vr_st{T,_,_},_}).
scalar_cmp_trace_used(SC,T) :- scalar_cmp_scl_used(SC,s_unconstrained{vr_t{T},_}).
scalar_cmp_trace_used(SC,T) :- scalar_cmp_scl_used(SC,s_unconstrained{vr_st{T,_,_},_}).
guard_trace_used(G,T) :- guard_scl_used(G,s_unconstrained{vr_t{T},_}).
guard_trace_used(G,T) :- guard_scl_used(G,s_unconstrained{vr_st{T,_,_},_}).

% -----------------------------------------------------------------------------

predicate guarded_scalar_kind(in list[guarded_scalar], in t_scalar_kind).

scalar_kind(s_int_const{_, Signed, Bits}, sk_int{Signed, Bits}) :- .
scalar_kind(s_float_const{_, Bits},       sk_float{Bits}) :- .
scalar_kind(s_unconstrained{_, Kind},     Kind) :- .
scalar_kind(s_cast{_, Kind},              Kind) :- .

?scalar_kind(s_guard{M}, _),
    map_empty(M),
    +warning("Empty guarded expression passed to scalar_kind").

scalar_kind(s_guard{M}, Kind) :-
    map_to_sorted_list(M, [pair{S, _G}|Gs]),
    scalar_kind(S, Kind),
    guarded_scalar_kind(Gs, Kind).

scalar_kind(s_unop{_, S}, Kind) :-
    scalar_kind(S, Kind).

scalar_kind(s_binop{_, S1, _S2}, K1) :-
    scalar_kind(S1, K1).

    % Check for a type mismatch in the arguments to a binary operator.
    % Left and right shifts can have different operand and operator types.
?scalar_kind(s_binop{Op, S1, S2}, _),
    scalar_kind(S1, K1),
    scalar_kind(S2, K2),
    K1 \= K2,
    Op \= sb_shiftlt,
    Op \= sb_shiftrt,
    +warning("Mismatched kinds of scalar in arguments to binary operation",
        K1, K2).

guarded_scalar_kind([], K) :- .
guarded_scalar_kind([pair{S, G}|GSs], K) :- 
    guarded_scalar_kind(GSs, K).
?guarded_scalar_kind([pair{S, _G}|_], K),
    scalar_kind(S, SK),
    K \= SK,
    +warning("Mismatched kinds of scalar in guarded scalar expression",
        K, S, SK).
    
% -----------------------------------------------------------------------------
% Stringification

    % maybe_bracket(PrecContext, PrecTopOp, Str, MaybeBracketedStr)
    % Add brackets to a string if the top-level operator's precedence is lower
    % than a given context precedence.
predicate maybe_bracket(in int, in int, in string, out string).
maybe_bracket(PC, P, S0, S) :-
    int_lt(P, PC),
    str_cat3("(", S0, ")", S).
maybe_bracket(PC, P, S0, S) :-
    int_ge(P, PC),
    S = S0.

    % Convert a guard/scalar/scalar comparison to a string, given the 
    % precendence context in which the string is to appear.
predicate guard_string_x(in g_guard, in int, out string).
predicate scalar_cmp_string_x(in scalar_cmp, in int, out string).
predicate scalar_string_x(in scalar, in int, out string).
predicate arep_string_x(in g_arep, in int, out string).

guard_string_x(G, _, BS) :- #bool_g(B, G), tostring(B, BS).
guard_string_x(G, PC, BS) :- #id_g(B, G), 
    (
        B = br_cmp{SC},
        scalar_cmp_string_x(SC, PC, BS)
    ;
        B = br_ioverflow{S},
        scalar_string_x(S, PC, SS),
        str_cat_list(["overflow(",SS,")"], BS)
    ;
        B = br_abit{AR},
        arep_string_x(AR, PC, BS)
    ).
guard_string_x(G, PC, S) :- 
    (
        #or(G1, G2, G), P = 0, Sep = "||"
    ;
        #and(G1, G2, G), P = 1, Sep = "&&"
    ),
    guard_string_x(G1, P, S1), guard_string_x(G2, P, S2),
    str_cat5(S1, " ", Sep, " ", S2, SB), maybe_bracket(PC, P, SB, S).

guard_string_x(G, PC, S) :- 
    P = 10,
    #not(G1, G),
    (
        #id_g(br_cmp{sc_eqz{LS}}, G1), scalar_string_x(LS, P, SB)
    ;
        ~#id_g(br_cmp{sc_eqz{_}}, G1), guard_string_x(G1, P, S1),
        str_cat("!", S1, SB)
    ),
    maybe_bracket(PC, P, SB, S).

scalar_cmp_string_x(sc_eqz{LS}, PC, S) :-
    P = 10, scalar_string_x(LS, P, LSS), 
    str_cat("!", LSS, SB), maybe_bracket(PC, P, SB, S).

scalar_cmp_string_x(sc{Op, LS, RS}, PC, S) :-
    scalar_cmp_op(Op, OpS, P),
    scalar_string_x(LS, P, LSS), scalar_string_x(RS, P, RSS),
    str_cat5(LSS, " ", OpS, " ", RSS, SB), maybe_bracket(PC, P, SB, S).

scalar_string_x(s_int_const{N,_,_},_,SS) :- tostring(N,SS).
scalar_string_x(s_float_const{N,_},_,SS) :- tostring(N,SS).

scalar_string_x(s_unconstrained{vr_t{T},_},_,SS) :-
    trace_string(T,SS).

scalar_string_x(s_unconstrained{vr_st{T,P,_},_},_,SS) :-
    trace_string(T,TS), tostring(P,PS),
    str_cat_list(["soft(",TS,",",PS,")"],SS).

scalar_string_x(s_unconstrained{vr_i{P,E},_},_,SS) :-
    exp_string(E,ES), tostring(P,PS),
    str_cat_list(["uninterpreted(",ES,",",PS,")"],SS).

scalar_string_x(s_unconstrained{vr_c{SUM,_,T},_},_,SS) :-
    trace_string(T,TS), tostring(SUM,SUMS),
    str_cat_list(["clobber(",TS,",",SUMS,")"],SS).

scalar_string_x(s_unconstrained{vr_instscalar{SUM,_,V},K},P,SS) :-
    scalar_string_x(s_unconstrained{V,K},P,VS), tostring(SUM,SUMS),
    str_cat_list(["instscalar(",VS,",",SUMS,")"],SS).

scalar_string_x(s_cast{RS,sk_int{SIGN,LEN}},P,SS) :-
    scalar_string_x(RS,P,RSS),
    (SIGN=true, SIGNS="int"; SIGN=false, SIGNS="uint"), tostring(LEN,LENS),
    str_cat_list(["(",SIGNS,LENS,")(",RSS,")"],SS).

scalar_string_x(s_guard{GSET},P,SS) :-
    $order(0), % TODO: probably broken for nested s_guard
    \/(map_search(GSET,V,G),
       scalar_string_x(V,P,VS), guard_string_x(G,P,GS),
       str_cat4(GS," -> ",VS,", ",VGS)):list_all(VGS,SLIST),
    str_cat_list(SLIST,SS0), str_cat(SS1,", ",SS0),
    str_cat3("[",SS1,"]",SS).

scalar_string_x(s_unop{OP,RS},P,SS) :-
    scalar_string_x(RS,P,RSS), sunop(UOP,OP), unop_str(UOP,OPS),
    str_cat_list([OPS,"(",RSS,")"],SS).

scalar_string_x(s_binop{OP,LS,RS},P,SS) :-
    scalar_string_x(LS,P,LSS), scalar_string_x(RS,P,RSS),
    (commutative_binop(BOP,OP);
     noncommutative_binop(BOP,OP)),
    binop_str(BOP,OPS),
    str_cat_list(["(",LSS,OPS,RSS,")"],SS).

arep_string_x(ar_extra{XR},_,SS) :-
    tostring(XR,XRS),
    str_cat_list(["extra_bit(",XRS,")"],SS).

arep_string_x(ar_path{P0,P1},_,SS) :-
    tostring(P0,P0S), tostring(P1,P1S),
    str_cat_list(["path_bit(",P0S,",",P1S,")"],SS).

arep_string_x(ar_entry{P,S,T},_,SS) :-
    tostring(P,PS), trace_string(S,S0), trace_string(T,S1),
    str_cat_list(["entry_bit(",PS,",",S0,",",S1,")"],SS).

arep_string_x(ar_exit{SUM,_,S,T},_,SS) :-
    tostring(SUM,SUMS), trace_string(S,S0), trace_string(T,T1),
    str_cat_list(["exit_bit(",SUMS,",",S0,",",T1,")"],SS).

arep_string_x(ar_clobber{SUM,_,T},_,SS) :-
    tostring(SUM,SUMS), trace_string(T,TS),
    str_cat_list(["clobber_bit(",SUMS,",",TS,")"],SS).

arep_string_x(ar_instguard{SUM,_,AR},PC,SS) :-
    tostring(SUM,SUMS), arep_string_x(AR,PC,PS),
    str_cat_list(["instguard_bit(",SUMS,",",PS,")"],SS).

    % If no precedence is provided, assume an arbitrarily low precedence 
    % context.
guard_string(G, S) :-
    guard_string_x(G, -1, S).
scalar_string(G, S) :-
    scalar_string_x(G, -1, S).
scalar_cmp_string(G, S) :-
    scalar_cmp_string_x(G, -1, S).

% -----------------------------------------------------------------------------

debug_scalar(), ?eval(P, E, _, _), ~eval(P, E, _, _), 
    point_location(P, File, Line),
    exp_string(E, ES),
    +warning("eval failed", P, E, File, Line, ES).

debug_scalar(), ?evals(P, E, _), ~evals(P, E, _), 
    point_location(P, File, Line),
    exp_string(E, ES),
    exp_type(E, Type),
    convert_type(Type, Y),
    Y \= y_func,
    Y \= y_valist,
    Y \= y_comp{_},
    type_string(Type, TS),
    +warning("evals failed", P, E, File, Line, ES, TS).

debug_scalar(), ?scalar_string(S,_), ~scalar_string(S,_),
    +warning("scalar_string failed",S).
debug_scalar(), ?guard_string(G,_), ~guard_string(G,_),
    +warning("guard_string failed",G).

% -----------------------------------------------------------------------------
