(*
 * Authors:
 *   Peter Hawkins  <hawkinsp@cs.stanford.edu>
 *   Brian Hackett  <bhackett@stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(* AST to spec rule conversion and stratification *)

open Spec
open Clpautil
open Flags
open Spectypes

module IO = Streamio
module A = Array 
module H = Hashtbl 

(* Rule Translation *)

(* get an rpred with no priority *)
let no_priority_rpred kind (spo, pred) = {
  r_priority = (0,rep2predo None);
  r_kind = kind;
  r_session = spo;
  r_pred = pred;
}

let debug_lambda_sfas = register_debug "lambda_sfas"
  "Print original SFA for each generated $lambda predicate"

let make_spec_rules (rules:ast_rule list) : t_rule list =
  let xlate_find sp kind =
    let mgc =
      if get_magic sp.s_pred then
        [no_priority_rpred Tp_add (sp.s_session, magic_pred sp.s_pred)]
      else [] in
    ([], mgc @ [no_priority_rpred kind (sp.s_session, sp.s_pred)]) in
  (* Take a goal and return a set of rules and a rule fragment such that 
   * evaluating the rule fragment together with the rules
   * is equivalent to evaluating the goal *)
  let rec lambda_rpred goal menv (gs : ast_goal list) (kind:t_rpred_kind) =
    match gs with [{body = Ag_find sp}] -> xlate_find sp kind | _ ->
    let name = "$lambda" ^ (string_of_int goal.id) in
      if debug_lambda_sfas() then
        IO.printf "Generated predicate %s: %s:%d\n"
          name (fst goal.sfa) (snd goal.sfa);
    let used = VS.diff menv.used menv.bound in
    let inargs = List.map var2val (VS.elements used) in
    let outargs = List.map var2val (VS.elements menv.bound) in
    let mgcpred = rep2pred (magic_name name,
        Array.of_list (inargs @ (List.map (fun _ -> any_val) outargs))) in
    let headpred = rep2pred (name, Array.of_list (inargs @ outargs)) in
    let (xrs, drfs) = List.split (List.map xlate_goal gs) in
    let drs = List.map (fun rf ->
      (if inargs = [] then rf else
          (no_priority_rpred Tp_find (rep2predo None, mgcpred))::rf) @
      [no_priority_rpred Tp_add (rep2predo None, headpred)]) drfs in
    let orf = (if inargs = [] then [] else
        [no_priority_rpred Tp_add (rep2predo None, mgcpred)]) @
      [no_priority_rpred kind (rep2predo None, headpred)] in
    (List.flatten (drs::xrs), orf)

  (* Goal to interpreter rule translation. Translating a goal produces a
   * (rules, rule fragment) tuple. *)
  and xlate_goal goal = match goal.body with
      Ag_conjunction cs -> List.fold_left (fun (rs,r) c ->
        let (crs, cr) = xlate_goal c in (rs @ crs, r @ cr)) ([],[]) cs
    | Ag_disjunction ds -> lambda_rpred goal goal.menv ds Tp_find 
    | Ag_find sp -> xlate_find sp Tp_find
    | Ag_wait pred -> 
        ([], [no_priority_rpred Tp_find (rep2predo None, magic_pred pred)])
    | Ag_add sp ->
        ([], [no_priority_rpred Tp_add (sp.s_session, sp.s_pred)])
    | Ag_negation g -> lambda_rpred goal goal.menv [g] Tp_negated 
    | Ag_collection (g,c) -> lambda_rpred goal g.menv [g] (Tp_collect c) in
    
  let xlate_rule rule = match rule with 
      Ar_flat goal -> let (rs, rf) = xlate_goal goal in (goal.sfa, rf::rs)
    | Ar_horn_clause (_, head, body) -> 
       let (rs, rf) = xlate_goal body in 
       let (name,args) = pred2rep head.s_pred in
       let allout = Clpautil.forall_array ((=) false) (get_mode (name,args)) in
       let r = (if allout then rf else (no_priority_rpred Tp_find 
                (head.s_session, magic_pred head.s_pred))::rf) 
         @ [no_priority_rpred Tp_add (head.s_session, head.s_pred)] in
       (body.sfa, r::rs)
    | Ar_query (pred,sfa) -> 
       let (name,args) = pred2rep pred in
       let qpred = rep2pred (query_pred_name,A.append [|str2val name|] args) in
       (sfa, [[no_priority_rpred Tp_find (rep2predo None, pred);
               no_priority_rpred Tp_add  (rep2predo None, qpred)]])
    in
  List.fold_left (fun rs ar -> (snd (xlate_rule ar)) @ rs) [] rules

(* Rule Ordering and Stratification *)

(* (fname,fargs) is an underconstrained magic version of (aname,args):
 * mgc$pred(A,B,_,_), +pred(A,B,C,D). *)
let magic_match (fname,fargs) (aname,aargs) =
  if fname <> (magic_name aname) then false
  else match make_vmap aargs fargs with
      Some vm -> (A.map (vmap_val vm) aargs) = aargs
    | None -> false

module OrderedRPred = struct type t = t_rpred let compare = compare end
module RPredG = Graph.Make (OrderedRPred)

let debug_strat_sccs = register_debug "strat_sccs"
  "Print each stratification SCC"

(* set priorities for all negation and collection operations, generating
 * an interference graph and stratifying the negation/collections based on
 * a topsort of the result, using $order() to resolve any cycles *)
let set_negation_priorities nextpri rules =
  let has_vmap args0 args1 =
    match args0 with None -> true | Some args0 ->
      make_vmap args0 (A.map (rename_val "%s") args1) <> None in
  let get_session_preds name =
    try H.find sessions name with Not_found -> H.create 10 in
  (* (n,args*must): (n,args) will always be added if must is true, or may be
   * added if must is false *)
  let r_adds : (string,t_val array option * bool) H.t = H.create 10 in
  (* (n,args*tail) is a find operation performed, with remaining rule tail *)
  let r_finds : (string,t_val array * t_rule) H.t = H.create 10 in
  (* (n,pred) is a negation/collection operation performed *)
  let r_negs : (string,t_rpred) H.t = H.create 10 in
  (* prepass to get add/find/negates in rules. mustx indicates this rule
   * will *always* be executed *)
  let rec get_rule_kinds mustx = function
      [] -> ()
    | rp::rulexs ->
        let (name,args) = pred2rep rp.r_pred in
          match predo2rep rp.r_session with
            Some sess -> 
              if name = merge_pred_name then 
                (H.iter (fun p () -> H.add r_adds p (None,false) ) 
                  (get_session_preds (fst (pred2rep sess)));
                get_rule_kinds mustx rulexs)
              else
                get_rule_kinds false rulexs
          | None -> (
            let mustx = match rp.r_kind with
                Tp_add -> H.add r_adds name (Some args,mustx); mustx
              | Tp_find -> H.add r_finds name (args,rulexs); false
              | Tp_negated
              | Tp_collect _ -> H.add r_negs name rp; false in
            get_rule_kinds mustx rulexs 
          ) in
    List.iter (get_rule_kinds true) rules;
  (* rule portions which are statically dead and can be ignored *)
  let deadrps : (t_rpred,unit) H.t = H.create 10 in
  (* preds for which a find/negate can succeed WITHOUT any explicit adds
   * add operations occurring in the rules. *)
  let dead_candidate rp =
    let initial_preds = get_session_preds !analysis_session in
    let name = fst (pred2rep rp.r_pred) in
      if predo2rep rp.r_session <> None then false
      else if H.mem find_handlers name then false
      else not (H.mem initial_preds name) in
    List.iter (fun rule ->
      ignore (List.fold_left (fun dead rp ->
        let (rname,rargs) = pred2rep rp.r_pred in
        if dead then (H.add deadrps rp (); true)
        else if not (dead_candidate rp) then false
        else if rp.r_kind = Tp_find then
          List.for_all (fun (args,_) ->
              not (has_vmap args rargs))
            (H.find_all r_adds rname)
        else if rp.r_kind = Tp_negated then
          List.exists (fun (argso,mst) ->
              match argso with None -> false | Some args ->
                mst && args = rargs && (forall_array is_val_const args))
            (H.find_all r_adds rname)
        else false) false rule)) rules;
  (* the interference graph. each Tp_add and Tp_negate/Tp_collect
   * in the program is a node in this graph, with the following edges:
   * add r0 -> add r1 :
   *   adding r0 can cause r1 to be added, through a Tp_find preceding r1.
   *   doesn't follow matching magic edges mgc$pred(ARGS) => pred(ARGS)
   * negate r0 -> add r1 :
   *   resolving r0 can cause r1 to be added later in the same rule.
   * add r0 -> negate r1 :
   *   adding r0 can change the result of r1
   * cycles in this graph containing negations are bad, and must be resolved
   * through the use of $order preceding all such negations *)
  let interferes : RPredG.t = RPredG.create() in
  let inter_edge rp0 rp1 =
    let srcn = RPredG.mk_node interferes rp0 in
    let tgtn = RPredG.mk_node interferes rp1 in
      ignore (RPredG.mk_edge interferes srcn tgtn) in
  (* $order() preds for each negation/collection which has them *)
  let neg_orders : (t_rpred,t_pred) H.t = H.create 10 in
  let rec get_interferes pnegs order = function
      [] -> ()
    | rp::rulexs ->
        let (name,args) = pred2rep rp.r_pred in
        if (H.mem deadrps rp) || (predo2rep rp.r_session <> None &&
            name <> merge_pred_name) then
          get_interferes pnegs order rulexs
        else begin
          let (pnegs,order) = match rp.r_kind with
            Tp_add ->
              let preds = match predo2rep rp.r_session with
                  None -> [(name, Some args)]
                | Some s -> H.fold (fun p () l -> (p, None)::l) 
                    (get_session_preds (fst (pred2rep s))) [] in
              List.iter (fun (name,args) ->
                (* add -> add edges *)
                List.iter (fun (fargs,ftail) ->
                  if has_vmap args fargs then
                    List.iter (fun frp ->
                      let frp_rep = pred2rep frp.r_pred in
                        if frp.r_kind = Tp_add then
                          if not (magic_match (name,fargs) frp_rep) then
                           inter_edge rp frp) ftail) (H.find_all r_finds name);
                (* negate -> add edges *)
                List.iter (fun neg -> inter_edge neg rp) pnegs;
                (* add -> negate edges *)
                List.iter (fun neg ->
                  if has_vmap args (snd (pred2rep neg.r_pred)) then
                    inter_edge rp neg) (H.find_all r_negs name)
              ) preds;
              (pnegs,order)
          | Tp_find ->
              if name = order_pred_name then (pnegs,Some rp.r_pred)
              else (pnegs,order)
          | Tp_negated
          | Tp_collect _ ->
              (match order with
                   Some order -> H.add neg_orders rp order
                 | None -> ());
              (rp::pnegs,order) in
          get_interferes pnegs order rulexs
        end in
  List.iter (get_interferes [] None) rules;
  (* get sccs and topsort *)
  let interscc = RPredG.compute_scc interferes in
  let sccmap : (t_rpred,t_rpred list) H.t = H.create 10 in
    RPredG.iter_scc interferes interscc (H.add sccmap);
  let intertpo = RPredG.topsort interferes interscc in
  (* is there a cycle for the given rpreds in an SCC? *)
  let has_cycle scclist =
    if List.length scclist > 1 then true
    else (* self cycle? *)
      let hd = RPredG.get_node interferes (List.hd scclist) in
        RPredG.query_edge interferes hd hd in
  (* compute new priorities for each negate/collect rpred *)
  let neg_pris : (t_rpred,int) H.t = H.create 10 in
  let _ = List.fold_left (fun nextpri rp ->
    if H.mem sccmap rp then begin
      let scclist = H.find sccmap rp in
      let neglist = List.filter (fun rp -> rp.r_kind <> Tp_add) scclist in
        if neglist <> [] && has_cycle scclist then begin
          let unordered = List.filter (o not (H.mem neg_orders)) neglist in
          let unordered_string neg =
            (if List.mem neg unordered then "(*)" else "")
            ^ string_of_rpred neg in
          if unordered <> [] then IO.printf
            "WARNING: Missing $order in negation/collection cycle: %s\n"
              (sep_string_of_list " " unordered_string neglist)
          else
            let order_count neg =
              let neg_order = H.find neg_orders neg in
                A.length (snd (pred2rep neg_order)) in
            let orders = List.map order_count neglist in
            let order_count_string neg =
              "(" ^ (string_of_int (order_count neg)) ^ ")"
              ^ (string_of_rpred neg) in
            if List.exists ((<>)(List.hd orders)) orders then IO.printf
              "WARNING: Inconsistent $order in negation/collection cycle: %s\n"
                (sep_string_of_list " " order_count_string neglist)
        end;
        if debug_strat_sccs() then
          print_endline (sep_string_of_list " " string_of_rpred scclist);
        List.iter (fun rp -> H.add neg_pris rp nextpri) neglist;
        nextpri + 1
    end else nextpri) nextpri intertpo in
  List.map (List.map (fun rp ->
    try let pri = H.find neg_pris rp in
      try let order = H.find neg_orders rp in
        { rp with r_priority = (pri,rep2predo (Some order)) }
      with Not_found ->
        { rp with r_priority = (pri,rep2predo None) }
    with Not_found -> rp)) rules

(* set rule priorities *)
let set_priorities rules =
  let prinone = rep2predo None in
  let cvt_rule nextpri rule =
    let (nextpri,revrule) = List.fold_left (fun (nextpri,revxs) rp ->
        let rp = { rp with r_priority = (nextpri,prinone) } in
          (nextpri+1,rp::revxs)) (nextpri,[]) rule in
    let rule = List.rev revrule in 
    (nextpri,rule) in
  let (nextpri,revrules) = List.fold_left (fun (nextpri,revxs) rule ->
    let (nextpri,crule) = cvt_rule nextpri rule in
      (nextpri,crule::revxs)) (0,[]) rules in
  set_negation_priorities nextpri (List.rev revrules)

let print_rules rules = 
  List.iter (fun r -> IO.printf "%s\n" (string_of_rule r)) rules
let debug_dump_rules = register_debug "dump_rules"
  "Dump interpreter rules after the compound goal removal pass"

(* Convert a list of rules to an analysis to be executed by the interpreter *)
let make_analysis rules = 
  check_rules rules;
  let spec_rules = make_spec_rules rules in
  if debug_dump_rules () then print_rules spec_rules;
  {
  Specio.a_name = !analysis_name;
  Specio.a_session = !analysis_session;
  Specio.a_topdown =
    (match !analysis_topdown with
         Some topdown -> topdown
       | None -> false);
  Specio.a_eager = !analysis_eager;
  Specio.a_priority =
    (match !analysis_priority with
         Some prio -> prio
       | None -> 0);
  Specio.a_packages =
    H.fold (fun name () packs -> name::packs) packages [];
  Specio.a_rules = Specio.dump_rules_str (set_priorities spec_rules);
  }
