(*
 * Authors:
 *   Peter Hawkins  <hawkinsp@cs.stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(* Translation of CLPA programs to Coq source code. *)

open Clpautil
open Flags
open Spec
open Specio

module A = Array
module H = Hashtbl
module PQ = Prioqueue
module S = String
module IO = Streamio
module V = Vector
module U = Unix
module E = Libevent

open Spectypes

(* CLP parsing *)

let stats_clp_sizes = register_stats "clp_sizes"
  "Print number of rules in each .clp file"
let debug_clp_rules = register_debug "clp_rules"
  "Print the rules in each .clp file"

(* List of directories that should be searched for source code *)
let source_dirs : string list ref = ref [Filename.current_dir_name]

let find_source_file filename searchdirs =
  let rec try_source_dirs file dirs =
    match dirs with 
        []    -> raise (Spectypes.File_not_found file)
      | d::ds -> let f = Filename.concat d file in
                 if Sys.file_exists f then f else try_source_dirs file ds in
  let len = String.length filename in
  let filename = if filename.[0] = '~' then 
    (Sys.getenv "HOME") ^ (String.sub filename 1 (len - 1)) else filename in
  if Filename.is_relative filename then
    try_source_dirs filename searchdirs 
  else if Sys.file_exists filename then filename
  else raise (Spectypes.File_not_found filename)

let parse_clp file = try
  let parsed : (string,unit) H.t = H.create 10 in
  let counts = Stack.create() in
  Stack.push 0 counts;
  let rec inner_parse (file : string) (extrasearchdirs : string list) =
    let base = Filename.basename file in
    if not (H.mem parsed base) then begin
      H.add parsed base ();
      if stats_clp_sizes() then Stack.push 0 counts;
      let searchdirs = extrasearchdirs @ !source_dirs in
      let file = find_source_file file searchdirs in
      let inch = open_in file in
      let inbuf = Lexing.from_channel inch in
      let oldsfa = changesfa (file,1) in
      let rules = Specparser.sentences Speclexer.token inbuf in
        ignore (changesfa oldsfa);
        if stats_clp_sizes() then begin
          let count = Stack.pop counts in
          let tot = List.length rules in
          let sub = tot - count in
            IO.printf "%s: %d (%d) rules\n" file sub tot;
            Stack.push (Stack.pop counts + List.length rules) counts
        end; rules
    end else [] in
    Spectypes.parse_file := inner_parse;
    List.iter Spectypes.add_package (get_builtin_packages());
    let rules = inner_parse file [] in
    Spectypes.check_rules rules;
      rules
  with 
      Spectypes.File_not_found file ->
        Printf.eprintf "File not found: %s\n" file ; exit 1 
    | Parse_error (file,line,message) ->
        Printf.eprintf "%s:%d: parse error: %s\n" file line message; exit 1
let parse_clp =
  make_track_1 "parse" parse_clp

(* Grouping of rules into SCCs *)

(* For each predicate name, the list of rules that can be used to produce it *)
let constructors : (string, ast_rule list) H.t = H.create 100

(* For each predicate P, on which other predicates does the definition of P
 * depend *)
module Pg = Graph.Make(OrderedString)
let depgraph : Pg.t = Pg.create ()

(* Add a rule to the constructor list, and compute the dependencies *)
let add_rule_deps (r:ast_rule) : unit = 
  let rec add_goal_deps (hp:Pg.n) (g:ast_goal) : unit =
    match g.body with
    | Ag_conjunction gs -> List.iter (add_goal_deps hp) gs
    | Ag_disjunction gs -> List.iter (add_goal_deps hp) gs
    | Ag_find sp -> 
        let predname = fst (pred2rep sp.s_pred) in
        if (predo2rep sp.s_session) = None && 
            not (H.mem find_handlers predname) then
        ignore (Pg.mk_edge depgraph hp (Pg.mk_node depgraph predname))
    | Ag_add _ ->
        rule_error g.sfa "Rules with add goals cannot be converted to Coq"
    | Ag_wait _ ->
        rule_error g.sfa "Rules with wait goals cannot be converted to Coq"
    | Ag_negation sg -> add_goal_deps hp sg
    | Ag_collection (_,_) -> failwith "Collection predicates not yet supported"
    in
  match r with 
  | Ar_flat g -> 
      rule_error g.sfa "Non-horn clause rules cannot be converted to Coq"
  | Ar_query (_,_) -> ()
  | Ar_horn_clause (nameo, head, body) ->
      if (predo2rep head.s_session) <> None then 
        failwith "Predicates added to sessions not yet supported";
      let (hname,_) = pred2rep head.s_pred in
      (* Add rule to the constructor list *)
      let cs = try H.find constructors hname with Not_found -> [] in
      H.replace constructors hname (r::cs);
      add_goal_deps (Pg.mk_node depgraph hname) body

let xlate_type_constructor tc = match tc with 
  | "maybe" -> "option"
  | _ -> tc

let xlate_fsym fs = match fs with
  | "yes" -> "Some"
  | "no" -> "None"
  | _ -> fs

let rec xlate_type needbrackets typ = 
  try
    let (tc,targs) = break_type typ in
    let tcstr = (xlate_type_constructor tc) in
      if A.length targs = 0 then tcstr
      else
        let tstr = tcstr ^ " " ^ 
          (sep_string_of_array " " (xlate_type true) targs) in
        if needbrackets then "(" ^ tstr ^ ")" else tstr
  with Unexpected _ ->
    let var = rev_vt_var typ in
    var

let xlate_predicate_type pred =
  let psym = H.find symbols pred in
  if A.length psym.s_args = 0 then
    "Prop"
  else
    (sep_string_of_array " -> " (xlate_type false) psym.s_args) ^ " -> Prop"

let dvid = ref 0

let rec xlate_term t =
  if t = any_val then 
    let dv = "dv" ^ (string_of_int !dvid) in
    incr dvid;
    (VS.singleton dv, dv)
  else try let name = val2var t in
    (VS.singleton name, name)
  with Unexpected _ -> try 
    (VS.empty, val2str t)
  with Unexpected _ -> try 
    (VS.empty, Big_int.string_of_big_int (val2int t))
  with Unexpected _ -> try 
    (VS.empty, string_of_float (val2flt t))
  with Unexpected _ -> try 
    let (name, args) = val2sum t in
    let (vars, astrs) = List.fold_right (fun a (vars, astrs) ->
        let (avars, astr) = xlate_term a in
        (VS.union vars avars, astr::astrs)
    ) (A.to_list args) (VS.empty, []) in
    if astrs = [] then (vars, xlate_fsym name)
    else 
      (vars, "(" ^ (xlate_fsym name) ^ " " ^ (sep_string_of_list " " id astrs) 
             ^ ")")
  with Unexpected _ -> failwith "Unexpected value type in xlate_term"

let xlate_predicate p =
  let (pred,args) = pred2rep p in
  if A.length args = 0 then (VS.empty, pred) else
    let (vars,cargs) = List.fold_left (fun (vars,cargs) t ->
      let (tvars, tstr) = xlate_term t in
      (VS.union vars tvars, tstr::cargs)) (VS.empty,[]) (A.to_list args) in
    (vars, pred ^ " " ^ (sep_string_of_list " " id (List.rev cargs)))

let rec xlate_goal toplevel g = 
  match g.body with 
  | Ag_conjunction gs -> 
      let (sgvars, sgs) = List.fold_right (fun sg (vars, cs) ->
      let (sgvars, sgs) = xlate_goal false sg in
      (VS.union vars sgvars, sgs::cs)) gs (VS.empty, []) in
      if toplevel then
        (sgvars, sep_string_of_list "\n      -> " id sgs)
      else
      (sgvars, sep_string_of_list " /\\ " id sgs)
  | Ag_find sp ->
      if (predo2rep sp.s_session) <> None then failwith "Unimplemented" else
      xlate_predicate sp.s_pred
  | _ -> (VS.empty, "unimpl")

    

let xlate_rule pred n r = match r with 
  | Ar_horn_clause (nameo, head, body) ->
    let name = match nameo with 
        | Some name -> name 
        | None -> pred ^ (string_of_int n) in
    let (bvars, bodys) = xlate_goal true body in
    let (hvars, heads) = (xlate_predicate head.s_pred) in
    let vars = VS.union bvars hvars in
      "| " ^ name ^ ":" ^
      (if not (VS.is_empty vars) then 
          " forall " ^ (sep_string_of_list " " id (VS.elements vars)) ^ "," 
       else "") ^ 
       (if bodys = "" then " " else "\n      " ^ bodys ^ "\n      -> ") ^
       heads
  | _ -> unreachable ()

let xlate_predicate_def pred =
  let cs = try H.find constructors pred with Not_found -> [] in
  pred ^ ": " ^ (xlate_predicate_type pred) ^ " :=\n  " ^
  let ss = List.rev (snd (List.fold_left (fun (n,rs) c ->
    (n+1, (xlate_rule pred n c)::rs)) (1,[]) cs)) in
  (sep_string_of_list "\n  " id ss)
      
(* Translate a CLP file to Coq source *)
let xlate_program program =
  let rules = parse_clp program in
  List.iter add_rule_deps rules;
  
  (* Compute a topological sort of the dependency graph *)
  let depscc = Pg.compute_scc depgraph in
  let depsccmems = H.create 10 in
  Pg.iter_scc depgraph depscc (H.add depsccmems);
  let deptpo = Pg.topsort depgraph depscc in 
  IO.printf "(* Coq translation of %s *)\n" program;
  List.iter (fun sccid ->
    let mems = H.find depsccmems sccid in
    IO.printf "Inductive %s.\n\n" 
        (sep_string_of_list "\nwith " xlate_predicate_def mems)
  ) (List.rev deptpo)

  

(* Main *)

let usage = "Usage: clpa [options] program.clp*"

let main () = 
  let programs : string list ref = ref [] in
  let add_program = prepend programs in
  let options = Arg.align [
    ("-I", Arg.String (prepend source_dirs),
     "dir  Extra directory to be searched for source code");
    ("--debug", Arg.String set_debug,
     "flag  Set debug flag");
    ("--list-debug", Arg.Unit list_all_debug,
     "  List all possible debug flags");
  ] in
  Arg.parse options (add_program) usage;
  if !programs = [] then (Arg.usage options usage; exit 1);
  List.iter xlate_program !programs
let main = make_track_1 "main" main

let _ = main()
