(*
 * Authors:
 *   Brian Hackett  <bhackett@stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(* Predicates for generating Dot graph files *)

open Spec
open Clpautil
open Flags

module H = Hashtbl
module V = Vector
module IO = Streamio

(* Dot graphs *)

let vt_dotgraph = make_type "dotgraph" [||]

type dgval = int
let vk_dgval = register_val_kind {
  vki_tostring = (fun _ v -> "DOTG#" ^ (string_of_int v));
  vki_compare = compare;
  vki_dump = (fun _ _ -> failwith "Cannot dump dot graph");
  vki_load = (fun _ -> failwith "Cannot load dot graph");
}

let sval2dgval = val2cst vk_dgval
let dgval2sval = cst2val vk_dgval

type dgattrs = (string * string) list ref

type dotgraph = {
  dg_name : string;
  dg_directed : bool;

  dg_graphattrs : dgattrs;
  dg_nodeattrs : dgattrs;
  dg_edgeattrs : dgattrs;

  dg_nodes : (t_val,dgattrs) H.t;
  dg_edges : (t_val,(t_val*t_val*dgattrs)) H.t;
}

let new_dotgraph name directed = {
  dg_name = name;
  dg_directed = directed;

  dg_graphattrs = ref [];
  dg_nodeattrs = ref [];
  dg_edgeattrs = ref [];

  dg_nodes = H.create 10;
  dg_edges = H.create 10;
}

let output_dotgraph fout dg =
  let out = IO.fprintf fout "%s" in
  let attr2str (attr,vs) = attr ^ "=\"" ^ (String.escaped vs) ^ "\"" in
  let attrs2str btxt atxt attrs =
    if !attrs = [] then ""
    else btxt ^ " [" ^ (sep_string_of_list "," attr2str !attrs) ^ "]" ^ atxt in
  let escp = o String.escaped (string_of_val !print_level) in
  let (gs,es) = if dg.dg_directed then ("digraph","->") else ("graph","--") in
    out (gs ^ " \"" ^ (String.escaped dg.dg_name) ^ "\" {\n");
    out (attrs2str "  graph" ";\n" dg.dg_graphattrs);
    out (attrs2str "  node"  ";\n" dg.dg_nodeattrs);
    out (attrs2str "  edge"  ";\n" dg.dg_edgeattrs);
    H.iter (fun vname attrs ->
      out ("  \"" ^ (escp vname) ^ "\"");
      out ((attrs2str "" "" attrs) ^ ";\n")) dg.dg_nodes;
    H.iter (fun _ (sname,tname,attrs) ->
      out ("  \"" ^ (escp sname) ^ "\" " ^ es ^ " \"" ^ (escp tname) ^ "\"");
      out ((attrs2str "" "" attrs) ^ ";\n")) dg.dg_edges;
    out "}\n"

let dotgraphs : dotgraph V.t = V.create 10
let dotgraph_hash : (string,dgval) H.t = H.create 10

let debug_dotty_database = register_debug "dotty_database"
  "Write dotty graphs into dotty.db"

let write_dotgraphs () =
  V.iter (fun graph ->
    if H.length graph.dg_nodes > 0 then
      let fout =
        if debug_dotty_database() then
          IO.open_dbentry "dotty" graph.dg_name IO.Pt_excl
        else IO.open_file (graph.dg_name ^ ".dot") IO.Pt_excl in
      output_dotgraph fout graph;
      IO.close_stream fout) dotgraphs

(* Dot package *)

let add_handlers = ref []
let find_handlers = ref []

let check_const args =
  if exists_array (o not is_val_const) args then
    pred_unexpected "Underconstrained"

let _ = prepend find_handlers
  (("dotty_graph",[|vt_str;Builtin.vt_bool;vt_dotgraph|]),
   [[|true;true;false|]],
   fun (name,args) ->
     let dgname = val2str args.(0) in
     let directed = Builtin.val2bool args.(1) in
     let dgr =
       try H.find dotgraph_hash dgname
       with Not_found ->
         let dgr = V.append dotgraphs (new_dotgraph dgname directed) in
           H.add dotgraph_hash dgname dgr; dgr in
     [name,[|args.(0);args.(1);dgval2sval dgr|]])

let _ = prepend add_handlers
  (("dotty_node",[|vt_dotgraph;vt_any|]),
   fun (name,args) ->
     check_const args;
     let dg = sval2dgval args.(0) in
     let graph = V.get dotgraphs dg in
       if not (H.mem graph.dg_nodes args.(1)) then
         H.add graph.dg_nodes args.(1) (ref []))

let _ = prepend add_handlers
  (("dotty_edge",[|vt_dotgraph;vt_any;vt_any;vt_any|]),
   fun (name,args) ->
     check_const args;
     let dg = sval2dgval args.(0) in
     let graph = V.get dotgraphs dg in
       if not (H.mem graph.dg_edges args.(1)) then begin
         if not (H.mem graph.dg_nodes args.(2)) then
           H.add graph.dg_nodes args.(2) (ref []);
         if not (H.mem graph.dg_nodes args.(3)) then
           H.add graph.dg_nodes args.(3) (ref []);
         H.add graph.dg_edges args.(1) (args.(2),args.(3),ref [])
       end)

let _ = prepend add_handlers
  (("dotty_attr",[|vt_dotgraph;vt_any;vt_str;vt_str|]),
   fun (name,args) ->
     check_const args;
     let dg = sval2dgval args.(0) in
     let attr = val2str args.(2) in
     let vs = val2str args.(3) in
     let graph = V.get dotgraphs dg in
     let attrs =
       if H.mem graph.dg_nodes args.(1) then
         H.find graph.dg_nodes args.(1)
       else if H.mem graph.dg_edges args.(1) then
         thd3 (H.find graph.dg_edges args.(1))
       else if args.(1) = str2val "graph" then graph.dg_graphattrs
       else if args.(1) = str2val "node" then graph.dg_nodeattrs
       else if args.(1) = str2val "edge" then graph.dg_edgeattrs
       else pred_unexpected "Can't find name" in
     if List.for_all (o ((<>)attr) fst) !attrs then
       attrs := (attr,vs)::!attrs)

let _ = register_package { empty_package with
  p_name = "dotty";
  p_writeout = write_dotgraphs;
  p_reset = (fun () -> V.clear dotgraphs; H.clear dotgraph_hash);
  p_add_handlers = !add_handlers;
  p_find_handlers = !find_handlers;
  p_abstypes = [vt_dotgraph];
}
