/*
 * Authors:
 *   Brian Hackett  <bhackett@stanford.edu>
 *   Peter Hawkins  <hawkinsp@cs.stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* Parser for specification files */

%{

open Spec
open Spectypes

module A = Array

let goal_id = ref 0

let new_goal gbody = 
  let g ={ 
    id = !goal_id;
    sfa = cur_sfa ();
    body = gbody; 
    menv = empty_menv
  } in
  incr goal_id;
  g

(* Take the conjunction of two AST goals, simplifying the result if possible *)
let conjoin g0 g1 = new_goal (match (g0.body, g1.body) with 
        (Ag_conjunction c0, Ag_conjunction c1) -> Ag_conjunction (c0 @ c1)
      | (Ag_conjunction c0, _) -> Ag_conjunction (c0 @ [g1])
      | (_, Ag_conjunction c1) -> Ag_conjunction (g0 :: c1)
      | (_, _) -> Ag_conjunction [g0; g1])
      
let disjoin g0 g1 = new_goal (match (g0.body, g1.body) with 
        (Ag_disjunction c0, Ag_disjunction c1) -> Ag_disjunction (c0 @ c1)
      | (Ag_disjunction c0, _) -> Ag_disjunction (c0 @ [g1])
      | (_, Ag_disjunction c1) -> Ag_disjunction (g0 :: c1)
      | (_, _) -> Ag_disjunction [g0; g1])

let unify_goal a0 a1 =
   let unify_pred = rep2pred ("equal", [|a0;a1|]) in
   new_goal (Ag_find {s_session = rep2predo None; s_pred = unify_pred})

%}

%token EOF
%token QUERY COLON EQUAL NOTEQUAL BNFDEF VBAR PERIOD ARROW SEMICOLON COMMA
%token FORALL QUESTION PLUS TURNSTILE TILDE
%token LPAREN RPAREN LBRACE RBRACE LBRACKET RBRACKET
%token IMPORT USING PREDICATE SESSION ANALYZE TYPE IN OUT CONTAINING
%token <Big_int.big_int> INT
%token <float> FLOAT
%token <string> STRING SYMBOL HSYMBOL DSYMBOL VAR ANYVAR

%right SEMICOLON
%right COMMA

%right TILDE

%start sentences
%type <Spectypes.ast_rule list> sentences

%%

sentences   :  EOF                { [] }
            |  rule sentences     { $1::$2 }
            |  query sentences    { $1::$2 }
            |  import sentences   { $1 @ $2 }
            |  define sentences   { $2 }
            |  analyze sentences  { $2 }
            |  using sentences    { $2 }

import      :  IMPORT STRING PERIOD
               {
                 try !parse_file $2 [Filename.dirname !curfile]
                 with File_not_found file -> 
                    parse_error ("File not found: " ^ file)
               }

define      :  PREDICATE mtpred PERIOD
               {
                 let pred = (fst $2,A.map snd (snd $2)) in
                 let mode = A.map fst (snd $2) in
                   add_predicate (cur_sfa()) pred; add_mode (fst $2) mode; []
               }
            |  SESSION tpred PERIOD  { add_session (cur_sfa ()) ($2,None); []}
            |  SESSION tpred CONTAINING LBRACKET session_contents RBRACKET 
                       PERIOD        { add_session (cur_sfa()) ($2,Some $5); []}
            |  TYPE typhead PERIOD   { [] }
            |  TYPE sumtype PERIOD   { add_sum_values $2; [] }
            |  TYPE aliastype PERIOD { add_alias_type $2; [] }

session_contents : /* empty */                           { [] }
                 | session_contents_items                { $1 }

session_contents_items : SYMBOL                               { [$1]   }
                       | SYMBOL COMMA session_contents_items  { $1::$3 }

rule        :  goal PERIOD                  { Ar_flat $1 }
            |  SYMBOL COLON spred TURNSTILE goal PERIOD  
                { Ar_horn_clause (Some $1, $3, $5) }
            |  SYMBOL COLON spred TURNSTILE PERIOD  
                { Ar_horn_clause (Some $1, $3, new_goal (Ag_conjunction [])) }
            |  spred TURNSTILE goal PERIOD  
                { Ar_horn_clause (None, $1,$3) }
            |  spred TURNSTILE PERIOD  
                { Ar_horn_clause (None, $1, new_goal (Ag_conjunction [])) }


analyze     :  ANALYZE pred PERIOD  { add_analyze $2 }

query       :  QUERY pred PERIOD    { Ar_query ($2, cur_sfa ()) }
using       :  USING SYMBOL PERIOD  { add_package $2 }

spred       :  pred { {s_session = rep2predo None; s_pred = $1} }
            |  pred ARROW pred 
                { {s_session = rep2predo (Some $1); s_pred = $3} }

goal        : spred                     { new_goal (Ag_find $1) }
            | PLUS spred                { new_goal (Ag_add   $2) }
            | LPAREN goal RPAREN        { $2 }  
            | goal COMMA goal           { conjoin $1 $3 }
            | goal SEMICOLON goal       { disjoin $1 $3 }
            | TILDE goal                { new_goal (Ag_negation $2) }
            | QUESTION pred             { new_goal (Ag_wait $2) }
            | FORALL goal COLON pred    { new_goal (Ag_collection ($2, $4)) }
            | arg EQUAL arg             { unify_goal $1 $3 }
            | arg NOTEQUAL arg          { new_goal 
                    (Ag_negation (unify_goal $1 $3)) }

predsym  :     SYMBOL   { $1 }
            |  HSYMBOL  { $1 }
            |  DSYMBOL  { $1 }

pred        :  predsym LPAREN RPAREN            { rep2pred ($1,[||]) }
            |  predsym LPAREN arglist RPAREN    { rep2pred ($1,A.of_list $3) }
tpred       :  predsym LPAREN RPAREN            { ($1,[||]) }
            |  predsym LPAREN targlist RPAREN   { ($1,A.of_list $3) }
mtpred      :  predsym LPAREN RPAREN            { ($1,[||]) }
            |  predsym LPAREN mtarglist RPAREN  { ($1,A.of_list $3) }

arglist     :  arg                    { [$1] }
            |  arg COMMA arglist      { $1::$3 }
targlist    :  targ                   { [$1] }
            |  targ COMMA targlist    { $1::$3 }
mtarglist   :  mtarg                  { [$1] }
            |  mtarg COMMA mtarglist  { $1::$3 }

arg         :  ANYVAR        { any_val }
            |  VAR           { var2val $1 }
            |  STRING        { str2val $1 }
            |  INT           { int2val $1 }
            |  FLOAT         { flt2val $1 }
            |  SYMBOL        { check_sum_value ($1,[||]); sum2val ($1,[||]) }
            |  SYMBOL tuple  { check_sum_value ($1,$2); sum2val ($1,$2) }
            |  LBRACKET RBRACKET                  { list2val None      [] }
            |  LBRACKET arglist RBRACKET          { list2val None      $2 }
            |  LBRACKET arglist VBAR arg RBRACKET { list2val (Some $4) $2 }

targ        :  typ            { check_type $1; $1 }
            |  VAR COLON typ  { check_type $3; $3 }
            |  VAR            { vt_var $1 }
            |  VAR COLON VAR  { vt_var $3 }
typhead     :  typ            { add_type (cur_sfa ()) $1; $1 }
typ         :  SYMBOL         { make_type $1 [||] }
            |  SYMBOL polytyp { make_type $1 $2 }
polytyp     :  LBRACKET targlist RBRACKET  { A.of_list $2 }

mtarg       :  targ      { (false,$1) }
            |  OUT targ  { (false,$2) }
            |  IN targ   { (true,$2) }

tuple       :  LBRACE arglist RBRACE   { A.of_list $2 }
ttuple      :  LBRACE targlist RBRACE  { A.of_list $2 }

aliastype   :  typhead EQUAL typ       { check_type $3; ($1,$3) }
sumtype     :  typhead BNFDEF sumlist  { ($1,$3) }

sumlist     :  sumval               { [($1,cur_sfa ())] }
            |  sumval VBAR sumlist  { ($1,cur_sfa ())::$3 }

sumval      :  SYMBOL         { ($1,[||]) }
            |  SYMBOL ttuple  { ($1,$2) }
