(*
 * Authors:
 *   Peter Hawkins  <hawkinsp@cs.stanford.edu>
 *   Brian Hackett  <bhackett@stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(* Type and mode checking for CLP rules *)

open Spec
open Clpautil
open Flags

module A = Array
module H = Hashtbl
module S = String
module IO = Streamio

module VS = Set.Make(OrderedString)

type ast_spred = {
  s_session : Spec.t_pred_option;
  s_pred    : Spec.t_pred;
}

(* a mode environment, indicating the state of all variables wrt a goal *)
type mode_env = {
  (* variables used by this goal, and which were bound either elsewhere
   * within this goal, or by an earlier goal *)
  used    : VS.t;

  (* variables that are *always* bound by this goal *)
  bound   : VS.t;

  (* variables that are either (a) sometimes but not always bound by this goal,
   * or (b) are used (implicitly) existentially quantified within this goal,
   * such as in a negation or collection, and thus can't be reused later *)
  noreuse : VS.t;
}

type ast_goal = {
  id : int;
  sfa : sfa;
  body : ast_goal_body;
  mutable menv : mode_env;
}

and ast_goal_body =
    Ag_conjunction of ast_goal list
  | Ag_disjunction of ast_goal list
  | Ag_find        of ast_spred
  | Ag_wait        of Spec.t_pred 
  | Ag_add         of ast_spred
  | Ag_negation    of ast_goal
  | Ag_collection  of ast_goal * Spec.t_pred

type ast_rule = 
    Ar_flat        of ast_goal
  | Ar_horn_clause of string option * ast_spred * ast_goal
  | Ar_query       of Spec.t_pred * Spec.sfa

type symbol_kind =
    Sk_predicate
  | Sk_session
  | Sk_collector
  | Sk_type
  | Sk_sum_value
  | Sk_analyze

let debug_parse_warnings = register_debug "parse_warnings"
  "Warn instead of fail on non-fatal parse-time errors"
let rule_warning sfa msg =
  (if debug_parse_warnings() then rule_warning else rule_error) sfa msg

(* opened packages *)
let packages : (string,unit) H.t = H.create 10

type t_symbol = {
    s_name : string;
    s_sfa  : sfa;              (* Location of the declaration of this symbol *)
    s_kind : symbol_kind;
    s_args : t_val_type array; (* Argument types, if any *)
}

(* types of arguments to all known symbols *)
let symbols : (string, t_symbol) H.t = H.create 100

(* all defined types (sumtypes or type aliases). no types can be redefined,
 * even to the same contents *)
let types : (string,unit) H.t = H.create 10

(* all known predicate modes. each predicate can have multiple modes *)
let modes : (string,t_pred_mode) H.t = H.create 10

(* all add/find package handlers *)
let add_handlers : (string,unit) H.t = H.create 10
let find_handlers : (string,unit) H.t = H.create 10

(* all sumvalue*sumtype pairs, indexed by the name of the sumtype value *)
let sumvalues : (string,t_pred_type * t_val_type) H.t = H.create 10

(* all type aliases, indexed by the name of the alias. get the arguments
 * to the alias type via symbols *)
let aliases : (string,t_val_type) H.t = H.create 10

(* for each session, an optional hashtable containing a list of predicate
 * names that may be present in a session. *)
let sessions : (string, (string,unit) H.t) H.t = H.create 10

(* session names that must be used with an external session *)
let sess_handlers = [merge_pred_name;clear_pred_name;process_dep_name]


(* all preds which have been used with the '?' operation *)
let magic_preds : (string,unit) H.t = H.create 10

(* analysis directives *)
let analysis_session = ref ""
let analysis_name = ref ""
let analysis_topdown = ref None  (* Some topdown if set *)
let analysis_eager = ref false
let analysis_priority = ref None (* Some prio if set *)

let reset_type_checking() =
  H.clear packages;
  H.clear symbols;
  H.clear types;
  H.clear modes;
  H.clear sessions;
  H.clear add_handlers;
  H.clear find_handlers;
  H.clear sumvalues;
  H.clear aliases;
  H.clear magic_preds;
  analysis_session := "";
  analysis_name := "";
  analysis_topdown := None;
  analysis_eager := false;
  analysis_priority := None

let sfa_invalid = ("Invalid location. This should not occur.", 0)
let sfa_builtin = ("(builtin)", 0)
let sfa_analysis = ("(analysis)", 0)

let add_symbol sfa kind (name,vals) =
  try let osym = H.find symbols name in
    if kind <> osym.s_kind then
      rule_error sfa ("Redefinition with different kind: " ^ name)
    else if vals <> osym.s_args then
      rule_error sfa ("Redefinition with different args: " ^ name)
    else ()
  with Not_found -> H.add symbols name {s_name = name; s_sfa = sfa; 
                                        s_kind = kind; s_args = vals}

let get_symbol sfa kind name =
  try let osym = H.find symbols name in
    (* special case for Sk_session, which can also be used as Sk_predicate *)
    if kind <> osym.s_kind && (kind <> Sk_predicate || 
       osym.s_kind <> Sk_session) then
      rule_error sfa ("Bad symbol kind: " ^ name)
    else osym
  with Not_found -> rule_error sfa ("Unknown symbol: " ^ name)

let add_predicate sfa pred = add_symbol sfa Sk_predicate pred
let add_session sfa (sess,co) = 
  let sname = fst sess in
  add_symbol sfa Sk_session sess ;
  match co with
      None -> ()
    | Some contents -> 
        let h = try H.find sessions sname with Not_found -> H.create 10 in
        H.add h sname (); (* Every session implicitly contains a predicate
                             with the same name as the session *)
        List.iter (fun p -> H.add h p ()) contents;
        H.replace sessions sname h
  
let add_collector coll = add_symbol sfa_builtin Sk_collector coll
let add_type sfa typ = add_symbol sfa Sk_type (break_type typ)

let define_type name =
  if H.mem types name then
    parse_error ("Type redefinition: " ^ name)
  else H.add types name ()

(* only for user .clp predicates, which can have only one mode *)
let add_mode name mode =
  try let old = H.find modes name in
    if old <> mode then
      parse_error ("Redefinition with different mode: " ^ name)
  with Not_found -> H.add modes name mode

let add_alias_type (atyp,ntyp) =
  let (name,_) = break_type atyp in
    define_type name;
    H.add aliases name ntyp

let add_sum_values (name,preds) =
  define_type (fst (break_type name));
  List.iter (fun (pred,sfa) ->
    add_symbol sfa Sk_sum_value pred;
    H.add sumvalues (fst pred) (pred,name)) preds

let add_package name =
  if not (H.mem packages name) then begin
    H.add packages name ();
    try let pack = get_package name in
      List.iter (fun ((n,a),_) ->
          add_predicate sfa_builtin (n,a); H.add add_handlers n ())
        pack.p_add_handlers;
      List.iter (fun ((n,a),ms,_) ->
          add_predicate sfa_builtin (n,a); H.add find_handlers n ();
          List.iter (H.add modes n) ms)
        pack.p_find_handlers;
      List.iter (fun ((n,a),m,_) -> add_collector (n,a); H.add modes n m)
        pack.p_collect_handlers;
      List.iter (add_predicate sfa_builtin) pack.p_predicates;
      List.iter (add_session sfa_builtin) pack.p_sessions;
      List.iter (fun (t,vals) -> add_type sfa_builtin t; 
        add_sum_values (t, List.map (fun v -> (v,sfa_builtin)) vals))
        pack.p_sumtypes;
      List.iter (fun t -> add_type sfa_builtin t; 
        define_type (fst (break_type t))) pack.p_abstypes
    with Not_found ->
      parse_error ("Unknown package: " ^ name)
  end

let _ = add_symbol sfa_builtin Sk_analyze ("session_name",[|vt_str|])
let _ = add_symbol sfa_builtin Sk_analyze ("analysis_name",[|vt_str|])
let _ = add_symbol sfa_builtin Sk_analyze ("topdown",[||])
let _ = add_symbol sfa_builtin Sk_analyze ("bottomup",[||])
let _ = add_symbol sfa_builtin Sk_analyze ("eager",[||])
let _ = add_symbol sfa_builtin Sk_analyze ("priority",[|vt_int|])

let add_analyze pred =
  let (name,args) = pred2rep pred in
  try match name with
      "session_name" ->
        let sname = val2str args.(0) in
          if !analysis_session <> "" then
            parse_error "Analysis session set multiple times";
          ignore (get_symbol (cur_sfa ()) Sk_session sname);
          analysis_session := sname
    | "analysis_name" ->
        let aname = val2str args.(0) in
          if !analysis_name <> "" then
            parse_error "Analysis name set multiple times";
          analysis_name := aname
    | "topdown" ->
        if !analysis_topdown <> None then
          parse_error "Analysis topdown/bottomup set multiple times";
        analysis_topdown := Some true
    | "bottomup" ->
        if !analysis_topdown <> None then
          parse_error "Analysis topdown/bottomup set multiple times";
        analysis_topdown := Some false
    | "eager" ->
        if !analysis_eager <> false then
          parse_error "Analysis eager set multiple times";
        analysis_eager := true
    | "priority" ->
        if !analysis_priority <> None then
          parse_error "Analysis priority set multiple times";
        analysis_priority := Some (val2sint args.(0))
    | _ ->
        parse_error ("Unknown symbol: " ^ name)
  with Unexpected _ ->
    parse_error ("Bad argument type for " ^ (full_string_of_pred pred))

let set_analysis_session sess =
  if !analysis_session <> "" then
    parse_error "Analysis session set multiple times";
  ignore (get_symbol (cur_sfa ()) Sk_session sess);
  analysis_session := sess

let set_analysis_name name =
  if !analysis_name <> "" then
    parse_error "Analysis name set multiple times";
  analysis_name := name

let set_analysis_topdown topdown =
  if !analysis_topdown <> None then
    parse_error "Analysis topdown/bottomup set multiple times";
  analysis_topdown := Some topdown

let rec remove_aliases typ =
  try let (name,args) = break_type typ in
    try let (nname,nargs) = break_type (H.find aliases name) in
      match make_type_vmap args (get_symbol (cur_sfa ()) Sk_type name).s_args 
        with
          Some tvm ->
            remove_aliases (make_type nname (A.map (vmap_type tvm) nargs))
        | None -> parse_error "Alias type mismatch"
    with Not_found -> make_type name (A.map remove_aliases args)
  with Unexpected _ -> typ

(* Check a symbol with arguments for common errors:
 * - Missing or wrong kind of prototype.
 * - Incorrect number of arguments.
 **)
let check_symbol kind sfa (name,argnum,str) =
  let pargs = (get_symbol sfa kind name).s_args in
    match is_pred_vargs (name,pargs) with
        Some (required,_) ->
          if argnum < (A.length required) then 
            rule_error sfa "Missing required arguments for varargs"
      | None ->
          if argnum <> (A.length pargs) then 
            rule_error sfa ("Bad argument count: " ^ str)

let filt_pred2sym p =
  let (name,args) = pred2rep p in
    (name,A.length args,full_string_of_pred p)
let check_predicate sfa = o (check_symbol Sk_predicate sfa) filt_pred2sym
let check_session sfa = o (check_symbol Sk_session sfa) filt_pred2sym
let check_spred sfa spo pred = 
  check_predicate sfa pred ;
  match predo2rep spo with
      None -> ()
    | Some s -> 
        check_session sfa s ;
        let sname = fst (pred2rep s) in
        let pname = fst (pred2rep pred) in
        if not (List.mem pname sess_handlers) then
        (try let scontents = H.find sessions sname in
          if not (H.mem scontents pname) then
            rule_error sfa ("Predicate " ^ pname ^ " is not a valid member of "
            ^ "session " ^ sname)
         with Not_found -> ()
        )

let check_collector sfa = o (check_symbol Sk_collector sfa) filt_pred2sym

let filt_sumv2sym (name,args) =
  (name,A.length args,string_of_val !print_level (sum2val (name,args)))
let check_sum_value v = check_symbol Sk_sum_value (Spec.cur_sfa ()) 
    (filt_sumv2sym v)

let filt_typ2sym typ =
  let (n,a) = break_type typ in (n,A.length a,string_of_type typ)
let check_type v = check_symbol Sk_type (Spec.cur_sfa ()) (filt_typ2sym v)

let get_all_modes (name,args) =
  match H.find_all modes name with
      [] when name = order_pred_name -> [A.make (A.length args) true]
    | [] -> [A.make (A.length args) false]
    | modes ->
        if List.exists (fun m -> A.length m <> A.length args) modes then
          failwith ("Bad mode length for " ^ name)
        else modes
let get_mode p = List.hd (get_all_modes p)

let all_out_mode (name,args) = forall_array ((=) false) (get_mode (name,args))

let magic_name name = "mgc$" ^ name
let magic_revert name =
  if String.length name < 4 then None
  else if String.sub name 0 4 <> "mgc$" then None
  else Some (String.sub name 4 (String.length name - 4))

let magic_pred p =
  let (name,args) = pred2rep p in
    (* convert all 'out' args to any_val *)
    let mode = get_mode (name,args) in
    let nargs = A.mapi (fun i v -> if mode.(i) then v else any_val) args in
      rep2pred (magic_name name,nargs)

let set_magic p =
  let (name,_) = pred2rep p in
    add_predicate sfa_builtin (magic_name name,(get_symbol (cur_sfa ()) 
        Sk_predicate name).s_args);
    H.replace magic_preds name ()
let get_magic p =
  H.mem magic_preds (fst (pred2rep p))


(* Well-formedness checking *)

(* Check an AST rule for well-formedness:
 * - symbol does not exist
 * - incorrect number of arguments
 * Also registers magic symbols.
 *)
let ast_check_rule rule =
  let rec ast_check_goal goal =
    match goal.body with
        Ag_conjunction goals -> List.iter ast_check_goal goals
      | Ag_disjunction goals -> List.iter ast_check_goal goals
      | Ag_find  sp   -> check_spred goal.sfa sp.s_session sp.s_pred
      | Ag_wait  pred -> check_predicate goal.sfa pred; set_magic pred
      | Ag_add   sp   -> check_spred goal.sfa sp.s_session sp.s_pred
      | Ag_negation goal -> ast_check_goal goal
      | Ag_collection (cgoal, cpred) -> 
          ast_check_goal cgoal; 
          check_collector goal.sfa cpred in
  match rule with 
      Ar_flat goal ->  ast_check_goal goal
    | Ar_horn_clause (_, head, body) -> 
        check_spred body.sfa head.s_session head.s_pred;
        let (name, args) = pred2rep head.s_pred in
        if not (all_out_mode (name,args)) then begin
          set_magic head.s_pred;
          if (predo2rep head.s_session) <> None then
            rule_error body.sfa
              "Rule head with external session has 'in' arguments"
        end;
        ast_check_goal body
    | Ar_query (pred,sfa) -> check_predicate sfa pred

let session_check () =
  H.iter (fun s preds -> 
    let ssym = get_symbol sfa_invalid Sk_session s in
    H.iter (fun pred _ -> ignore (get_symbol ssym.s_sfa Sk_predicate pred)) 
           preds
  ) sessions

(* get the main predicate information for use in checking *)
let pred_info rp =
  let (name,args) = pred2rep rp in
    (name,args,full_string_of_pred rp)

(* Type Checking *)

(* map variables to their types *)
type t_vt_vmap = (t_val * t_val_type) list
exception Type_error of string

(* Type check an AST rule *)
let type_check rule =
  let renames : int ref = ref 0 in
  let next_rename() =
    let nstr = (string_of_int !renames) ^ "%" in
      incr renames; rename_type nstr in
  (* skolemize a type, removing all type variables. leave vt_any alone as
   * this will only be used for special handler predicates like print() *)
  let rec skolem_type typ =
    try let x = rev_vt_var typ in
      make_type (x ^ "$") [||]
    with Unexpected _ -> try let (n,args) = break_type typ in
      make_type n (A.map skolem_type args)
    with Unexpected _ -> typ in
  (* unify types et and vt (type of v), updating tvm *)
  let type_unify v where tvm et vt =
    match add_type_vmap tvm vt et with
        Some tvm -> tvm
      | None ->
          raise (Type_error (Printf.sprintf
            "Bad type for %s at %s: expected %s, got %s"
            (full_string_of_val v) where
            (string_of_type (vmap_type tvm et))
            (string_of_type (vmap_type tvm vt)))) in
  (* unify the type of val v with typ, updating the variable type unifier vm
   * (var->typ) and type variable unifier tvm (typvar->typ). *)
  let rec val_type_unify where (vm,tvm) v typ =
    if v = any_val then (vm,tvm)
    else try let name = val2var v in
      try (vm,type_unify v where tvm typ (snd (List.find (o ((=) v) fst) vm)))
      with Not_found ->
        let var_typ = vt_var name in
        let tvm = type_unify v where tvm typ var_typ in
          ((v,var_typ)::vm,tvm)
    with Unexpected _ -> try ignore (val2str v);
      (vm,type_unify v where tvm typ vt_str)
    with Unexpected _ -> try ignore (val2int v);
      (vm,type_unify v where tvm typ vt_int)
    with Unexpected _ -> try ignore (val2flt v);
      (vm,type_unify v where tvm typ vt_flt)
    with Unexpected _ -> try let (name,args) = val2sum v in
      let ((name,targs),sumtyp) = H.find sumvalues name in
      let rename = next_rename() in
      let (vm,tvm) = val_type_pred_unify rename (full_string_of_val v)
                       (vm,tvm) name args targs in
        (vm,type_unify v where tvm typ (rename sumtyp))
    with Unexpected _ -> failwith "Bad value type"
  and val_type_pred_unify rename where (vm,tvm) name args targs =
    let targs = A.map (o rename remove_aliases) targs in
    let array_unify = fold_array2 (val_type_unify where) in
      match is_pred_vargs (name,targs) with
          Some (reqargs,otyp) ->
            let reqlen = A.length reqargs in
            let optlen = A.length args - reqlen in
            let sargs = A.sub args 0 reqlen in
            let oargs = A.sub args reqlen optlen in
            let (vm,tvm) = array_unify (vm,tvm) sargs reqargs in
              array_unify (vm,tvm) oargs (A.create optlen otyp)
        | None ->
            array_unify (vm,tvm) args targs in
  let type_check_pred (vm, tvm) pred kind is_add sfa =
    let rename = next_rename() in
    let (pname,pargs,pstr) = pred_info pred in
    let ptargs =
      if is_add then
        A.map (o skolem_type rename) (get_symbol sfa kind pname).s_args
      else (get_symbol sfa kind pname).s_args in
    try val_type_pred_unify rename pstr (vm,tvm) pname pargs ptargs
    with Type_error err -> rule_error sfa err in
  let type_check_spred (vm, tvm) sp is_add sfa =
    let (vm,tvm) = match predo2rep sp.s_session with
        Some sp ->
          let (name,args,str) = pred_info sp in
          let targs = (get_symbol sfa Sk_session name).s_args in
          let rename = next_rename() in
            (try val_type_pred_unify rename str (vm,tvm) name args targs
             with Type_error err -> rule_error sfa err)
      | _ -> (vm,tvm) in
      type_check_pred (vm,tvm) sp.s_pred Sk_predicate is_add sfa in
  (* walk through a goal, maintaining var->type and tvar->type maps.
   * for add operations the types must be at least as general as the prototype,
   * for find operations the types must be unifiable with the prototype *)
  let rec type_check_goal tenv goal =
    match goal.body with
        Ag_conjunction goals -> List.fold_left type_check_goal tenv goals
      | Ag_disjunction goals -> List.fold_left type_check_goal tenv goals
      | Ag_find spred -> type_check_spred tenv spred false goal.sfa
      | Ag_wait  pred ->  type_check_pred tenv pred Sk_predicate false goal.sfa
      | Ag_add   spred -> type_check_spred tenv spred true goal.sfa
      | Ag_negation goal -> type_check_goal tenv goal
      | Ag_collection (cgoal, cpred) ->
          let te1 = type_check_goal tenv cgoal in
          type_check_pred te1 cpred Sk_collector false goal.sfa in
  let empty_tenv = ([], type_vmap_empty) in
  match rule with 
      Ar_flat goal             -> 
        ignore (type_check_goal empty_tenv goal)
    | Ar_horn_clause (_, head, body) -> 
        let tenv = type_check_goal empty_tenv body in
        ignore (type_check_spred tenv head true body.sfa)
    | Ar_query (pred, sfa) -> 
        ignore (type_check_pred empty_tenv pred Sk_predicate false sfa)

(* Mode Checking *)

let empty_menv = {
  bound = VS.empty; 
  used = VS.empty; 
  noreuse = VS.empty;
}

let union_menv m0 m1 = {
  bound = VS.union m0.bound m1.bound;
  used = VS.union m0.used m1.used; 
  noreuse = VS.union m0.noreuse m1.noreuse
}

exception Mode_error of string
let mode_error s = raise (Mode_error s)

(* check each non-ground value in v, marking as consumed any ground vars.
 * any_fn will be called for wildcards, and var_fn will be called for
 * non-ground vars.
 * m is the environment for the goal currently being checked,
 * mprev indicates any variables bound or noreuse due to previous goals *)
let rec check_ground any_fn var_fn mprev m loc v =
  if v = any_val then any_fn mprev m loc
  else try let x = val2var v in
    if VS.mem x m.bound || VS.mem x mprev.bound then 
      { m with used = VS.add x m.used }
    else var_fn mprev m x loc
  with Unexpected _ ->
    (try let (_,args) = val2sum v in
       A.fold_left (fun m av ->
         check_ground any_fn var_fn mprev m (full_string_of_val v) av) m args
     with Unexpected _ -> m)

(* different ways we can check wildcards/variables in a value *)
let bad_any _ _ l = mode_error ("Free wildcard \'_\' at " ^ l)
let bad_var _ _ x l = mode_error ("Free variable \'" ^ x ^ "\' at " ^ l)
let ignore_any _ m l = m
let add_var mprev m x l =
  if VS.mem x mprev.noreuse then
    mode_error ("Can't reuse variable \'" ^ x ^ "\' at " ^ l)
  else { m with bound = VS.add x m.bound }

(* most common ways of checking wildcards/variables *)
let must_ground = check_ground bad_any bad_var
let add_ground = check_ground ignore_any add_var

(* check that a pred mode is matched by a name/args for a given operation *)
let check_find_mode mprev m (name,args,pstr) mode =
  fold_array2 (fun m mv av ->
    (if mv then must_ground else add_ground) mprev m pstr av) m mode args
let check_wait_mode mprev m (name,args,pstr) mode =
  fold_array2 (fun m mv av ->
    if mv then add_ground mprev m pstr av else m) m mode args
let check_add_mode mprev m (name,args,pstr) mode =
  fold_array2 (fun m mv av ->
    (if mv then add_ground else must_ground) mprev m pstr av) m mode args

(* check that a find operation on a pred is matched by at least one mode
 * for that pred *)
let check_all_find_modes mprev m (name,args,pstr) =
  let rec check_list mprev m modes = match modes with 
      [] -> failwith "Missing mode"
    | [mode] -> check_find_mode mprev m (name,args,pstr) mode
    | mode::xs -> try check_find_mode mprev m (name,args,pstr) mode
                  with Mode_error _ -> check_list mprev m xs in
  check_list mprev m (get_all_modes (name,args))

let bad_pred name msg = mode_error ("Predicate " ^ name ^ msg)

(* check an external session access through pred 'name' for mode errors *)
let check_session_mode name mprev spo = match predo2rep spo with
    Some sp ->
      if (H.find symbols name).s_kind = Sk_session then
        bad_pred name " cannot be used with an external session";
      if (H.mem add_handlers name || H.mem find_handlers name) then
        if List.for_all ((<>)name) sess_handlers then
          bad_pred name " cannot be used with an external session" ;
      let (_,sargs,spstr) = pred_info sp in
        A.fold_left (fun m sav ->
          must_ground mprev m spstr sav) empty_menv sargs
  | None ->
      if List.exists ((=)name) sess_handlers then
        bad_pred name " must be used with an external session";
      empty_menv

(* report an error if a set is non-empty *)
let check_empty_set sfa msg =
  VS.iter (fun v -> rule_warning sfa ("Variable " ^ v ^ msg))

(* Mode check a rule by doing a forward scan over the goals keeping track
 * of how each variable has been used so far.
 * Also checks for:
 * - adding preds with find handlers
 * - finding preds with add handlers
 * - variables appearing only once
 *)
let mode_check rule =
  (* Mode check a goal, given a mode environment which provides a list
   * of bound and non-reusable variables. The result of mode checking is stored
   * in the menv field of the goal. *)
  let rec mode_check_goal mprev goal = try
    match goal.body with 
      Ag_conjunction subgoals -> 
        goal.menv <- List.fold_left (fun m conjunct ->
            mode_check_goal (union_menv mprev m) conjunct ;
            union_menv m conjunct.menv) empty_menv subgoals
    | Ag_disjunction subgoals -> 
        List.iter (mode_check_goal mprev) subgoals;
        let boundlist   = List.map (fun g -> g.menv.bound)   subgoals in
        let usedlist    = List.map (fun g -> g.menv.used)    subgoals in
        let noreuselist = List.map (fun g -> g.menv.noreuse) subgoals in
        let anybound = List.fold_left VS.union VS.empty boundlist in
        let allbound = List.fold_left VS.inter anybound boundlist in
        let used     = List.fold_left VS.union VS.empty usedlist in
        let noreuse  = List.fold_left VS.union VS.empty noreuselist in
          goal.menv <- {
            bound = allbound;
            used = used;
            noreuse = VS.union noreuse (VS.diff anybound allbound);
          }
    | Ag_find spred ->
        let (name,args,pstr) = pred_info spred.s_pred in
        if H.mem add_handlers name then
          bad_pred name " cannot be used in find operations";
        let msp = check_session_mode name mprev spred.s_session in 
          goal.menv <- check_all_find_modes mprev msp (name,args,pstr)
    | Ag_wait pred ->
        let (name,args,pstr) = pred_info pred in
        if H.mem add_handlers name then
          bad_pred name " cannot be used in find operations";
        let mode = get_mode (name,args) in
          goal.menv <- check_wait_mode mprev empty_menv (name,args,pstr) mode
    | Ag_add spred ->
        let (name,args,pstr) = pred_info spred.s_pred in
        if H.mem find_handlers name then
          bad_pred name " cannot be used in add operations";
        if (H.find symbols name).s_kind = Sk_session then
          bad_pred name " cannot be used in add operations";
        let msp = check_session_mode name mprev spred.s_session in 
        let mode = get_mode (name,args) in
        let ma = check_add_mode mprev msp (name,args,pstr) mode in
          check_empty_set goal.sfa " is bound but not used within add"
            (VS.diff (VS.diff ma.bound msp.bound) ma.used);
          goal.menv <-  {
            bound = msp.bound;
            used = ma.used;
            noreuse = VS.union ma.noreuse (VS.diff ma.bound msp.bound);
          }
    | Ag_negation subgoal ->
        mode_check_goal mprev subgoal;
        check_empty_set subgoal.sfa " is bound but not used by negation"
          (VS.diff subgoal.menv.bound subgoal.menv.used);
        goal.menv <- {
          bound = VS.empty;
          used = subgoal.menv.used;
          noreuse = VS.union subgoal.menv.bound subgoal.menv.noreuse;
        }
    | Ag_collection (lambda, cp) ->
        mode_check_goal mprev lambda;
        let ml = union_menv mprev lambda.menv in
        let (cname,cargs,cpstr) = pred_info cp in
        let mode = get_mode (cname,cargs) in
        let mcoll = check_find_mode ml empty_menv (cname,cargs,cpstr) mode in
        check_empty_set lambda.sfa " is bound but not used by collection"
          (VS.diff lambda.menv.bound (VS.union lambda.menv.used mcoll.used));
        goal.menv <- {
          bound = mcoll.bound;
          used = VS.union lambda.menv.used mcoll.used;
          noreuse = VS.union lambda.menv.bound lambda.menv.noreuse;
        }
    with Mode_error err -> rule_error goal.sfa err in
  (* Mode check a rule *)
  let (sfa,m) = match rule with 
      Ar_flat goal ->
        (* check last predicate is always an add *)
        let rec check_last_add goal =
          match goal.body with
              Ag_add _ -> ()
            | Ag_conjunction goals ->
                check_last_add (List.hd (List.rev goals))
            | Ag_disjunction goals ->
                List.iter check_last_add goals
            | _ ->
                rule_warning goal.sfa "Last predicate in goal is not an add" in
        check_last_add goal;
        mode_check_goal empty_menv goal;
        (goal.sfa,goal.menv)
    | Ar_horn_clause (_, head,body) -> (try
        let (name,args,hstr) = pred_info head.s_pred in
        let mode = get_mode (name,args) in
        let mqhead =
          check_wait_mode empty_menv empty_menv (name,args,hstr) mode in
        mode_check_goal mqhead body;
        let munion = union_menv mqhead body.menv in
        let munion = union_menv munion
          (check_session_mode name munion head.s_session) in
        let munion = check_add_mode empty_menv munion (name,args,hstr) mode in
          (body.sfa,munion)
       with Mode_error err -> rule_error body.sfa err)
    | Ar_query (p,sfa) -> (sfa,empty_menv) in
  check_empty_set sfa " is bound but not used in rule"
    (VS.diff m.bound m.used)

(* Perform static checking on rules *)
let check_rules rules =
  session_check ();
  List.iter (fun r -> ast_check_rule r; type_check r; mode_check r) rules;
  if !analysis_session = "" then rule_error sfa_analysis 
    "Missing session to iterate over, must add 'analyze session_name(...)'";

(* Parsing hooks *)
exception File_not_found of string
let parse_file = ref (fun _ _ -> [])
