(*
 * Authors:
 *   Peter Hawkins  <hawkinsp@cs.stanford.edu>
 *   Brian Hackett  <bhackett@stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(* reset state after checking an entire spec file *)
val reset_type_checking : unit -> unit

(* commit user-defined symbols *)
val add_predicate : Spec.sfa -> Spec.t_pred_type -> unit
val add_session : Spec.sfa -> Spec.t_pred_type * string list option -> unit
val add_type : Spec.sfa -> Spec.t_val_type -> unit
val add_alias_type : Spec.t_val_type * Spec.t_val_type -> unit
val add_sum_values : Spec.t_val_type * (Spec.t_pred_type * Spec.sfa) list 
                     -> unit

val add_mode : string -> Spec.t_pred_mode -> unit
val get_mode : Spec.t_pred_rep -> Spec.t_pred_mode

(* add a package or analysis directive *)
val add_package : string -> unit
val add_analyze : Spec.t_pred -> unit

(* check symbols for well-formedness:
 * - symbol doesn't exist
 * - incorrect number of arguments
 **)
val check_sum_value : string * Spec.t_val array -> unit
val check_type : Spec.t_val_type -> unit

module VS : Set.S with type elt = string

type ast_spred = {
    s_session : Spec.t_pred_option ;
    s_pred    : Spec.t_pred
}

(* mode environment, internal to spectypes.ml *)
type mode_env = {
  used : VS.t;
  bound : VS.t;
  noreuse : VS.t;
}

val empty_menv : mode_env

type ast_goal = {
    id : int;
    sfa : Spec.sfa;
    body : ast_goal_body;
    mutable menv : mode_env
}

and ast_goal_body =
    Ag_conjunction of ast_goal list
  | Ag_disjunction of ast_goal list
  | Ag_find        of ast_spred
  | Ag_wait        of Spec.t_pred 
  | Ag_add         of ast_spred
  | Ag_negation    of ast_goal
  | Ag_collection  of ast_goal * Spec.t_pred

type ast_rule = 
    Ar_flat        of ast_goal
  | Ar_horn_clause of string option * ast_spred * ast_goal 
  | Ar_query       of Spec.t_pred * Spec.sfa

(* Hook for spec parser *)
exception File_not_found of string
val parse_file : (string -> string list -> ast_rule list) ref

val get_magic : Spec.t_pred -> bool
val magic_pred : Spec.t_pred -> Spec.t_pred 
val magic_name : string -> string

val packages : (string, unit) Hashtbl.t
val sessions : (string, (string,unit) Hashtbl.t) Hashtbl.t

val analysis_session : string ref
val analysis_name: string ref
val analysis_topdown : bool option ref
val analysis_eager : bool ref
val analysis_priority : int option ref

val find_handlers : (string,unit) Hashtbl.t

val check_rules : ast_rule list -> unit

(* Things needed by the Coq translation *)
type symbol_kind =
    Sk_predicate
  | Sk_session
  | Sk_collector
  | Sk_type
  | Sk_sum_value
  | Sk_analyze

type t_symbol = {
    s_name : string;
    s_sfa  : Spec.sfa;         (* Location of the declaration of this symbol *)
    s_kind : symbol_kind;
    s_args : Spec.t_val_type array; (* Argument types, if any *)
}

(* types of arguments to all known symbols *)
val symbols : (string, t_symbol) Hashtbl.t 
val find_handlers : (string, unit) Hashtbl.t

