(*
 * Authors:
 *   Peter Hawkins  <hawkinsp@cs.stanford.edu>
 *
 * Copyright (c) 2005-2006,
 *   The Board of Trustees of The Leland Stanford Junior University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(* Interface to the STP theorem prover *)

open Clpautil
open Flags
open Spec

module H = Hashtbl
module V = Vector
module IO = Streamio

open Biteval


let reset_fns : (unit -> unit) list ref = ref []
let reset() = List.iter (fun f -> f()) !reset_fns

(* Satisfiability *)

(* Satisfying assignments constructed from SAT *)
type bassign = int
let tbl_assigns : (t_val,bool) H.t V.t = V.create 10

(* Per-bval solver results and solve times (in seconds) *)
let tbl_results : (bval,(bassign option)*float) H.t = H.create 10

let tbl_var2bvar : (t_val, Stp.bexpr) H.t = H.create 10

let _ = prepend reset_fns (fun () ->
  V.clear tbl_assigns;
  H.clear tbl_results;
  H.clear tbl_var2bvar)

let debug_stp_solve = register_debug "stp_solve"
  "Print result of each STP query"

(* Get the STP variable corresponding to an abstract leaf *)
let get_bvar x =
  try H.find tbl_var2bvar x
  with Not_found ->
    let zx = Stp.b_newvar () in
    H.add tbl_var2bvar x zx; zx

(* Convert a bval to a STP bexpr *)
let rec bval_to_stp_bexpr (bv:bval) =
  match get_bval bv with
      Bb_const  c   -> if c then Stp.b_true else Stp.b_false
    | Bb_var    v   -> get_bvar v
    | Bb_unop  (Bo_eq,v) -> bval_to_stp_bexpr v
    | Bb_unop  (Bo_not,v) -> Stp.b_not (bval_to_stp_bexpr v)
    | Bb_binop (Bo_and,vs) -> List.fold_left 
        (fun a v -> Stp.b_and (bval_to_stp_bexpr v) a) Stp.b_true vs
    | Bb_binop (Bo_or,vs) -> List.fold_left 
        (fun a v -> Stp.b_or (bval_to_stp_bexpr v) a) Stp.b_false vs
        
let get_assign () =
  let assign = H.create 10 in
  H.iter (fun x bx -> match Stp.vc_get_bval bx with
        Stp.False -> H.add assign x false
      | Stp.True  -> H.add assign x false
      | _ -> ()) tbl_var2bvar;
  assign

let stp_solve bv =
  let start = Sys.time() in
  let bv = simplify bv in
    try fst (H.find tbl_results bv)
    with Not_found ->
      if debug_stp_solve() then
        IO.printf "stp: %s\n" (string_of_bval !print_level bv);
      let be = bval_to_stp_bexpr bv in
      let valid = Stp.vc_query (Stp.b_not be) in
      let res = if valid then None else Some (V.append tbl_assigns 
        (get_assign ())) in
      if debug_stp_solve () then
        IO.printf "result: %s\n" (if res <> None then "SAT" else "UNSAT");
      let elapsed = Sys.time() -. start in
        H.add tbl_results bv (res,elapsed); res
let stp_solve = make_track_1 "stp_solve" stp_solve

let apply_assignment av v =
  let assign = V.get tbl_assigns av in
  let cache = H.create 10 in
  let rec evaluate v = match get_bval v with
      Bb_const b -> b
    | Bb_var x -> (try H.find assign x with Not_found -> false)
    | Bb_unop(Bo_eq,v) -> evaluate v
    | Bb_unop(Bo_not,v) -> not (evaluate v)
    | Bb_binop(op,vals) ->
        try H.find cache v
        with Not_found ->
          let combfn = if op = Bo_and then (&&) else (||) in
          let res = list_fold_right (o combfn evaluate)
                      (List.tl vals) (evaluate (List.hd vals)) in
            H.add cache v res; res in
  evaluate v

(* Spec integration *)

(* The type of STP bitvectors *)
let vt_stp_bvec ntyp = make_type "stp_bvec" [|ntyp|]

(* The type of comparisons of STP bitvectors *)
let vt_stp_bvcmp ntyp = make_type "stp_bvcmp" [|ntyp|]

(* The type of STP formula leaf values. A leaf value is either a comparison
 * over STP bit vectors, or an extra stp_abit *)
let vt_stp_leaf ntyp atyp = make_type "stp_leaf" [|ntyp;atyp|]
let sum_stp_leaf = (vt_stp_leaf vt_xvar vt_yvar, [
  ("stp_leaf_cmp", [|vt_stp_bvcmp vt_xvar|]);
  ("stp_leaf_abit", [|vt_yvar|]);
])

let vt_bassign typ = make_type "bval_asn" [|typ|]
let vk_bassign = register_val_kind {
  vki_tostring = (fun _ a -> "ASN#" ^ (string_of_int a));
  vki_compare = compare;
  vki_dump = (fun _ _ -> failwith "Cannot dump bassign");
  vki_load = (fun _ -> failwith "Cannot load bassign");
}

let val2bassign = val2cst vk_bassign
let bassign2val = cst2val vk_bassign

(* Packages *)

let add_handlers = ref []
let find_handlers = ref []

let vt_xvar = vt_var "X"
let vt_xbval = vt_bval vt_xvar
let vt_xbassign = vt_bassign vt_xvar

let _ = prepend find_handlers
  (("#sat",[|vt_bval vt_any|]),[[|true|]],
   fun (name,args) ->
     if stp_solve (val2bval args.(0)) <> None then [name,args] else [])

let _ = prepend find_handlers
  (("#satasn",[|vt_xbval;vt_xbassign|]),[[|true;false|]],
   fun (name,args) ->
     let res = stp_solve (val2bval args.(0)) in
     match res with
         None -> []
       | Some ar -> 
         [name,[|args.(0);bassign2val ar|]])

let _ = prepend find_handlers
  (("#asng",[|vt_xbassign;vt_xbval|]),[[|true;true|]],
   fun (name,args) ->
     let asn = val2bassign args.(0) in
     if apply_assignment asn (val2bval args.(1)) then [name,args] else [])

let stats_stp_counts = register_stats "stp_counts"
  "Summarize SAT solver problems and times"

let write_stp_counts() =
  if stats_stp_counts() then begin
    (* keep track of (total queries, total time, maximum time) *)
    let (q,t,m) = H.fold (fun _ (_,time) (q,t,m) ->
                            (q+1,t+.time,max m time)) tbl_results (0,0.,0.) in
      IO.printf "SAT Solving Stats:\n";
      IO.printf "  Count: %d\n" q;
      if q <> 0 then begin
        let avg = t /. (float_of_int q) in
          IO.printf " Total:   %.3f\n" t;
          IO.printf " Average: %.3f\n" avg;
          IO.printf " Max:     %.3f\n" m
      end
  end

let _ = register_package { empty_package with
  p_name = "stp";
  p_writeout = write_stp_counts;
  p_reset = reset;
  p_add_handlers = !add_handlers;
  p_find_handlers = !find_handlers;
  p_abstypes = [vt_xbassign];
}

